/* Error and debug handling.
   Copyright (c) 2000, 2001 Idaya Ltd.
   Written by Nick Burrett <nick@dsvr.net>

   This file is part of the Virtual Server Administrator (FreeVSD)

   FreeVSD is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2, or (at your option)
   any later version.

   FreeVSD is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with FreeVSD; see the file COPYING.  If not, write to
   the Free Software Foundation, Inc., 59 Temple Place - Suite 330,
   Boston, MA 02111-1307, USA.  */

#include <ctype.h>
#include <stdio.h>
#include <errno.h>
#include <stdarg.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <syslog.h>
#include "libvsd.h"

void vsd_log_error (int severity, const char *fmt, ...)
{
  va_list ap;
  int log;

  va_start (ap, fmt);

  switch (severity)
    {
    case VSD_LOG_PROTOCOL: log = LOG_INFO; break;
    case VSD_LOG_DEBUG: log = LOG_DEBUG; break;
    case VSD_LOG_INFO: log = LOG_INFO; break;
    default: log = LOG_ERR; break;
    }

  vsyslog (log, fmt, ap);
  /* vfprintf (stderr, fmt, ap); */

  va_end (ap);
}

/* Static buffer for storing a VSD error message.  */
static char *vsd_errmsg = NULL;

/* Length of that static buffer.  */
static int vsd_errmsg_len = 0;

/* Set `vsd_errmsg' to an error specified by the printf-like format
   string `msg'.  `error_number' is currently unused but is reserved
   for future use.

   Always return 1 for failure.  */
int vsd_log_set_error (int error_number, const char *msg, ...)
{
  int len;
  va_list ap;

  if (vsd_errmsg == NULL)
    {
      vsd_errmsg_len = 128;
      vsd_errmsg = (char *) malloc (vsd_errmsg_len + 1);
    }

  va_start (ap, msg);

  /* Safeguard against a buffer overflow.  */
  len = vsnprintf (vsd_errmsg, vsd_errmsg_len, msg, ap);
  if (len >= vsd_errmsg_len)
    {
      /* Adjust our buffer to cope with the new error string length.  */
      vsd_errmsg_len = len + 32;
      vsd_errmsg = (char *) realloc (vsd_errmsg, vsd_errmsg_len + 1);
    }
  vsnprintf (vsd_errmsg, vsd_errmsg_len, msg, ap);
  va_end (ap);

  return 1;
}

/* Return a pointer to the error stored in vsd_errmsg.  Useful for
   functions to finally report the error message back to the client
   application.

   The `level' is currently unused but reserved for future use.  */
const char *vsd_log_get_error (int level)
{
  return vsd_errmsg;
}
