/* openssl.c - openssl handling.
 */

/* 
 * VSD - Virtual Server Daemon
 * Copyright (c) 2000 - 2001 Idaya Ltd.
 *
 * VSD is free software; you can redistribute it and/or 
 * modify it under the terms of the GNU General Public License 
 * as published by the Free Software Foundation; either 
 * version 2 of the License, or (at your option) any later version.
 *
 * VSD is distributed in the hope that it will be useful, 
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU 
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License 
 * along with VSD - see the COPYING file;  if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330, 
 * Boston, MA 02111-1307, USA.  
 */

/*
 * See the AUTHORS file for a list of contributors to VSD. 
 * See the ChangeLog files for a list of changes. 
 */

#include "config.h"

#include <unistd.h>
#include <time.h>

#ifdef WITH_OPENSSL
#include <openssl/conf.h>
#include <openssl/txt_db.h>
#include <openssl/asn1.h>
#endif

#include "libvsd.h"

#ifdef WITH_OPENSSL
#define VSD_OPENSSL_PATH_LENGTH_MAX 256

#define VSD_OK  0
#define VSD_ERR -1

/* Shamelessly ripped from the OpenSSL ca application */
#define DB_type         0
#define DB_exp_date     1
#define DB_rev_date     2
#define DB_serial       3       /* index - unique */
#define DB_file         4       
#define DB_name         5       /* index - unique for active */
#define DB_NUMBER       6

#define DB_TYPE_REV	'R'
#define DB_TYPE_EXP	'E'
#define DB_TYPE_VAL	'V'

static int check_time_format(char *str)
	{
	ASN1_UTCTIME tm;

	tm.data=(unsigned char *)str;
	tm.length=strlen(str);
	tm.type=V_ASN1_UTCTIME;
	return(ASN1_UTCTIME_check(&tm));
	}

static unsigned long index_serial_hash(char **a)
	{
	char *n;

	n=a[DB_serial];
	while (*n == '0') n++;
	return(lh_strhash(n));
	}

static int index_serial_cmp(char **a, char **b)
	{
	char *aa,*bb;

	for (aa=a[DB_serial]; *aa == '0'; aa++);
	for (bb=b[DB_serial]; *bb == '0'; bb++);
	return(strcmp(aa,bb));
	}

static unsigned long index_name_hash(char **a)
	{ return(lh_strhash(a[DB_name])); }

static int index_name_qual(char **a)
	{ return(a[0][0] == 'V'); }

static int index_name_cmp(char **a, char **b)
	{ return(strcmp(a[DB_name],
	     b[DB_name])); }

/* End of OpenSSL ca code */

/* FIXME - This is here because it is commented out of the OpenSSL library! */
time_t ASN1_UTCTIME_get(const ASN1_UTCTIME *s)
{
  struct tm tm;
  int offset;

  memset(&tm,'\0',sizeof tm);

#define g2(p) (((p)[0]-'0')*10+(p)[1]-'0')
  tm.tm_year=g2(s->data);
  if(tm.tm_year < 50)
    tm.tm_year+=100;
  tm.tm_mon=g2(s->data+2)-1;
  tm.tm_mday=g2(s->data+4);
  tm.tm_hour=g2(s->data+6);
  tm.tm_min=g2(s->data+8);
  tm.tm_sec=g2(s->data+10);  
  if(s->data[12] == 'Z')
    offset=0;
        else
	  {
	    offset=g2(s->data+13)*60+g2(s->data+15);
	    if(s->data[12] == '-')
	      offset= -offset;
	  }
#undef g2

  return mktime(&tm)-offset*60; /* FIXME: mktime assumes the current timezone
				 * instead of UTC, and unless we rewrite OpenSSL
				 * in Lisp we cannot locally change the timezone
				 * without possibly interfering with other parts
				 * of the program. timegm, which uses UTC, is
				 * non-standard.
				 * Also time_t is inappropriat for general
				 * UTC times because it may a 32 bit type. */
}



/********************************************************************************
 * Description: Initialise OpenSSL configuration.
 *
 *     Returns: LHASH * to OpenSSL configuration on success.
 *              NULL on error.
 *******************************************************************************/
LHASH * 
vsd_openssl_conf_init () 
{
  char fname[VSD_OPENSSL_PATH_LENGTH_MAX];
  char *pch = NULL;
  char *section = NULL;
  long errorline = -1;
  LHASH *conf;

  /* Obtain configuration location via environment variables? */
  memset (fname, 0, sizeof (fname));
  if (!pch)
    {
      pch = getenv ("VSD_OPENSSL_CONF");
      if (!pch)
	pch = getenv ("OPENSSL_CONF");
      if (!pch)
	pch = getenv ("SSLEAY_CONF");
    }  

  if (pch)
    {
      strncpy (fname, pch, sizeof (fname) - 1);
      fname [sizeof(fname) - 1] = '\0';
    }
  else if (access ("/etc/vsd/openssl.cnf", R_OK | X_OK) == 0)
    strcpy (fname, "/etc/vsd/openssl.cnf");
  else if (access ("/usr/local/etc/vsd/openssl.cnf", R_OK | X_OK) == 0)
    strcpy (fname, "/usr/local/etc/vsd/openssl.cnf");
  else
    {
      strncpy (fname, X509_get_default_cert_area(), sizeof(fname)-13);
      strcat(fname,"/openssl.cnf");
    }

  vsd_log_error (VSD_LOG_DEBUG, "using OpenSSL configuration from %s", fname);

  if ((conf = CONF_load (NULL, fname, &errorline)) == NULL) 
    {
      if (errorline <= 0)
	vsd_log_error (VSD_LOG_ERROR, "error loading OpenSSL configuration file %s", 
		       fname);
      else
	vsd_log_error (VSD_LOG_ERROR, "error on line %ld of OpenSSL configuration file %s", 
		       errorline, fname);
      return NULL;
    }
  
  return conf;
}

/********************************************************************************
 * Description: Lookup value from OpenSSL configuration hash.
 *
 *     Returns: char * to value on success.
 *              NULL on error.
 *******************************************************************************/
char *
vsd_openssl_lookup (LHASH **conf, 
		    char  *section, 
		    char  *key)
{
  char *value;
  
  if ((value = CONF_get_string (*conf, section, key)) == NULL)
    vsd_log_error (VSD_LOG_ERROR, "OpenSSL variable lookup failed for %s::%s", 
		   section, key);
  
  return value;
}

/********************************************************************************
 * Description: Initialise and lookup value from OpenSSL configuration hash.
 *
 *     Returns: char * to value on success.
 *              NULL on error.
 *******************************************************************************/
char *
vsd_openssl_ca_lookup (LHASH **conf,
		       char  *key)
{
  char *section, *value;

  if (!(*conf) && (*conf = vsd_openssl_conf_init()) == NULL)
    return NULL;
 
  if ((section = vsd_openssl_lookup (conf, "ca", "default_ca")) == NULL)
    return NULL;
  
  if ((value = vsd_openssl_lookup (conf, section, key)) == NULL)
    return NULL;
  
  return value;
}

/********************************************************************************
 * Description: Load and establish indexes for VSD CA database.
 *
 *     Returns: TXT_DB * to VSD CA database on success.
 *              NULL on error.
 *******************************************************************************/
TXT_DB *
vsd_openssl_ca_load_database (LHASH **conf)
{
  char *database;
  TXT_DB *txt_db = NULL;
  BIO *in = NULL;

  if ((database = vsd_openssl_ca_lookup (conf, "database")) == NULL) 
    return NULL;
  
  if ((in = BIO_new(BIO_s_file())) == NULL)
    {
      vsd_log_error (VSD_LOG_ERROR, "error setting 'BIO_new' in");
      return NULL;
    }
  
  if (BIO_read_filename (in, database) <= 0)
    {
      vsd_log_error (VSD_LOG_ERROR, "error opening VSD CA database %s", 
		     database);
      return NULL;
    }
    
  if ((txt_db = TXT_DB_read (in, DB_NUMBER)) == NULL)
    {
      vsd_log_error (VSD_LOG_ERROR, "error loading VSD CA database %s", 
		     database);
      return NULL;;
    }
  
  if (!TXT_DB_create_index (txt_db, DB_serial, NULL, 
			    index_serial_hash, index_serial_cmp))
    {
      vsd_log_error (VSD_LOG_ERROR, 
		     "error creating serial number index:(%ld,%ld,%ld)",
		     txt_db->error, txt_db->arg1, txt_db->arg2);
      return NULL;
    }
  
  if (!TXT_DB_create_index (txt_db, DB_name, index_name_qual, 
			    index_name_hash, index_name_cmp))
    {
      vsd_log_error (VSD_LOG_ERROR, 
		     "error creating name index:(%ld,%ld,%ld)",
		     txt_db->error, txt_db->arg1, txt_db->arg2);
      return NULL;
    }

  return txt_db;
}

/********************************************************************************
 * Description: Verify VSD CA database.
 *
 *     Returns: char * to description of errors.
 *              NULL on error.
 *******************************************************************************/
char *
vsd_openssl_ca_verify_database (LHASH **conf)
{
  char *errors = NULL;
  char *p, **pp;
  TXT_DB *txt_db = NULL;
  int i, j;

  if ((txt_db = vsd_openssl_ca_load_database (conf)) == NULL)
    return NULL;

  for (i = 0; i < sk_num (txt_db->data); i++)
    { 
       pp = (char **) sk_value (txt_db->data, i);
      if ((pp[DB_type][0] != DB_TYPE_REV) 
	  && (pp[DB_rev_date][0] != '\0'))
	{
	  vsd_log_error (VSD_LOG_ERROR, 
			 "entry %d: not revoked yet, but has a revocation date.", i+1);
	}
      if ((pp[DB_type][0] == DB_TYPE_REV) 
	  && !check_time_format (pp[DB_rev_date]))
	{
	  vsd_log_error (VSD_LOG_ERROR, "entry %d: invalid revocation date.", i+1);
	}
      if (!check_time_format (pp[DB_exp_date]))
	{
	  vsd_log_error (VSD_LOG_ERROR, "entry %d: invalid expiry date.", i+1);
	}
      p = pp[DB_serial];
      j = strlen(p);
      if ((j&1) || (j < 2))
	{
	  vsd_log_error (VSD_LOG_ERROR, "entry %d: bad serial number length (%d).", i+1, j);
	}
      while (*p)
	{
	  if (!(((*p >= '0') && (*p <= '9')) ||
		((*p >= 'A') && (*p <= 'F')) ||
		((*p >= 'a') && (*p <= 'f')))  )
	    {
	      vsd_log_error (VSD_LOG_ERROR, "entry %d: bad serial number characters," 
			     "char pos %ld, char is '%c'\n",i+1,(long)(p-pp[DB_serial]), *p);
	    }
	  p++;
	}
    }

  vsd_log_error (VSD_LOG_INFO, "%d VSD CA entries checked.", txt_db->data->num);
  
  return NULL;
}

/********************************************************************************
 * Description: Determine serial number of requested VSD CA database entry.
 *
 *     Returns: char * to serialnum on success.
 *              NULL on error.
 *******************************************************************************/
char *
vsd_openssl_ca_serial_from_name (LHASH      **conf,
				 const char *cn)
{
  char *database, *serialnum, *subject, *buf, *pch;
  TXT_DB *txt_db = NULL;
  char *row[DB_NUMBER], **rrow;
  struct vsd_file_line *fdatabase, *line;

  if ((txt_db = vsd_openssl_ca_load_database (conf)) == NULL)
    return NULL;

  /* FIXME - Do this properly, using X509 calls on root certificate... */
  {
    if ((database = (char *) vsd_openssl_ca_lookup (conf, "database")) == NULL) 
      return;
    
    if (vsd_load_file (&fdatabase, "/", database))
      return;
    
    if ((buf = (char *) strdup (strchr (fdatabase->line, '/'))) == NULL)
      return;
    
    for (pch = &buf [strlen (buf) - 1]; *pch ;pch--)
      {
	if (*pch == '/')
	  break;
      }
    *pch = '\0';
    if ((subject  = (char *)malloc (strlen (buf) + 4 + strlen (cn) + 1)) == NULL)
      return;
    
    sprintf (subject, "%s/CN=%s", buf, cn);
    
    vsd_free_file (fdatabase);
    free (buf);
  }

  row [DB_name] = subject;
  if ((rrow = TXT_DB_get_by_index (txt_db, DB_name, row)) == NULL)
    {
      vsd_log_error (VSD_LOG_ERROR, "no active VSD CA entry for %s", cn);
      return NULL;
    }
  
  if ((serialnum = rrow[DB_serial]) == NULL)
    {
      vsd_log_error (VSD_LOG_ERROR, "no active serial number for %s in VSD CA", cn);
      return NULL;
    }
  
  free (subject);
  
  return serialnum;
}

/********************************************************************************
 * Description: Path to a VSD CA issued certificate.
 *
 *     Returns: char * to crt path on success.
 *              NULL on error.
 *******************************************************************************/
char *
vsd_openssl_ca_vsdcrt (LHASH      **conf,
		       const char *cn)
{
  char *certs;
  static char vsdcrt[VSD_OPENSSL_PATH_LENGTH_MAX];

  if ((certs = vsd_openssl_ca_lookup (conf, "certs")) == NULL)
    return NULL;
  
  sprintf (vsdcrt, "%s/%s.crt", certs, cn);
  
  return vsdcrt;
}

/********************************************************************************
 * Description: Path to a VSD CA issued key.
 *
 *     Returns: char * to keys path on success.
 *              NULL on error.
 *******************************************************************************/
char *
vsd_openssl_ca_vsdkey (LHASH      **conf,
		       const char *cn)
{
  char *keys;
  static char vsdkey[VSD_OPENSSL_PATH_LENGTH_MAX];
  
  if ((keys = vsd_openssl_ca_lookup (conf, "keys")) == NULL)
    if ((keys = vsd_openssl_ca_lookup (conf, "certs")) == NULL)
      return NULL;
  
  sprintf (vsdkey, "%s/%s.key", keys, cn);
  
  return vsdkey;
}

/********************************************************************************
 * Description: Path to a VSD CA issued crl.
 *
 *     Returns: char * to crl path on success.
 *              NULL on error.
 *******************************************************************************/
char *
vsd_openssl_ca_vsdcrl (LHASH **conf)
{
  char *crls, *crl, *pch;
  static char vsdcrl[VSD_OPENSSL_PATH_LENGTH_MAX];

  if ((crls = vsd_openssl_ca_lookup (conf, "crls")) == NULL)
    if ((crls = vsd_openssl_ca_lookup (conf, "certs")) == NULL)
      return NULL;

  if ((crl = vsd_openssl_ca_lookup (conf, "crl")) == NULL)
    return NULL;
  
  for (pch = &crl [strlen (crl) - 1]; *pch ;pch--)
    {
      if (*pch == '/' || *pch == '=' || isspace (*pch))
	break;
    }
  sprintf (vsdcrl, "%s/%s", crls, ++pch);

  return vsdcrl;
}

/********************************************************************************
 * Description: Path to a VSD CA issued root certificate.
 *
 *     Returns: char * to cacert path on success.
 *              NULL on error.
 *******************************************************************************/
char *
vsd_openssl_ca_vsdcertificate (LHASH **conf)
{
  char *cacerts, *certificate, *pch;
  static char vsdcertificate[VSD_OPENSSL_PATH_LENGTH_MAX];
  
  if ((cacerts = vsd_openssl_ca_lookup (conf, "cacerts")) == NULL)
    if ((cacerts = vsd_openssl_ca_lookup (conf, "certs")) == NULL)
      return NULL;

  if ((certificate = vsd_openssl_ca_lookup (conf, "certificate")) == NULL)
    return NULL;

  for (pch = &certificate [strlen (certificate) - 1]; *pch ;pch--)
    {
      if (*pch == '/' || *pch == '=' || isspace (*pch))
	break;
    }
  sprintf (vsdcertificate, "%s/%s", cacerts, ++pch);
  
  return vsdcertificate;
}

/********************************************************************************
 * Description: Path to a VSD CA generated certificate.
 *
 *     Returns: char * to crt path on success.
 *              NULL on error.
 *******************************************************************************/
char *
vsd_openssl_ca_crt (LHASH      **conf,
	            const char *cn)
{
  char *serialnum;
  char *new_certs_dir;
  static char crt[VSD_OPENSSL_PATH_LENGTH_MAX];
  
  if ((serialnum = vsd_openssl_ca_serial_from_name (conf, cn)) == NULL)
    return NULL;  

  if ((new_certs_dir = vsd_openssl_ca_lookup (conf, "new_certs_dir")) == NULL)
    return NULL;
  
  sprintf (crt, "%s/%s.pem", new_certs_dir, serialnum);
  
  return crt;
}

/********************************************************************************
 * Description: Path to a VSD CA generated key.
 *
 *     Returns: char * to keys path on success.
 *              NULL on error.
 *******************************************************************************/
char *
vsd_openssl_ca_key (LHASH      **conf,
		    const char *cn)
{
  char *serialnum;
  char *private_key, *private, *pch;
  static char key[VSD_OPENSSL_PATH_LENGTH_MAX];
  
  if ((serialnum = vsd_openssl_ca_serial_from_name (conf, cn)) == NULL)
    return NULL;  

  if ((private_key = vsd_openssl_ca_lookup (conf, "private_key")) == NULL)
    return NULL;
    
  private = (char *)strdup (private_key);
  for (pch = &private [strlen (private) - 1]; *pch ;pch--)
    {
      if (*pch == '/' || *pch == '=' || isspace (*pch))
	break;
    }
  *pch = '\0';

  sprintf (key, "%s/%s.pem", private, serialnum);

  free (private);

  return key;
}

/********************************************************************************
 * Description: Return OpenSSL CA database records.
 *
 *     Returns: char * pointer to buffer of formatted database records.
 *              NULL on error.
 *******************************************************************************/
char *
vsd_openssl_ca_db_records (LHASH **conf,
			   int   offset,
			   int   num)
{
  TXT_DB *txt_db = NULL;
  char **pp, *p;
  ASN1_UTCTIME asn1_tm;
  time_t rev_t;
  time_t exp_t;
  int i, max, buf_sz;
  char *exp_date, *rev_date;
  char tmp[1024];
  static char *buf;

  if ((txt_db = vsd_openssl_ca_load_database (conf)) == NULL)
    return NULL;

  if (offset > sk_num (txt_db->data))
    return NULL;

  if (num == 0 || num > 250)
    num = 250;
  
  max = offset + num;
  if (max > sk_num (txt_db->data))
    max = sk_num (txt_db->data);

  buf_sz = 1;
  buf = NULL;
  for (i = offset; i < max; i++)
    {
      pp=(char **)sk_value(txt_db->data,i);

      asn1_tm.data=(unsigned char *)pp[DB_exp_date];
      asn1_tm.length=strlen(pp[DB_exp_date]);
      asn1_tm.type=V_ASN1_UTCTIME;
      exp_t = ASN1_UTCTIME_get (&asn1_tm);
      exp_date = (char *) strdup (ctime (&exp_t));
      exp_date[strlen (exp_date) - 1] = '\0';

      if (pp[DB_rev_date][0] != '\0')
	{
	  asn1_tm.data=(unsigned char *)pp[DB_rev_date];
	  asn1_tm.length=strlen(pp[DB_rev_date]);
	  asn1_tm.type=V_ASN1_UTCTIME;
	  rev_t = ASN1_UTCTIME_get (&asn1_tm);
	  rev_date = (char *) strdup (ctime (&rev_t));
	  rev_date[strlen (rev_date) - 1] = '\0';
	}
      else
	rev_date = NULL;

      p = (char *) strstr (pp[DB_name], "CN=");
      p += 3;

      sprintf (tmp, "%s,%s,%s,%s,%c\n",
	       pp[DB_serial],
	       exp_date,
	       rev_date,
	       p,
	       pp[DB_type][0]);

      buf_sz = buf_sz + strlen (tmp);
      p = buf;
      if ((buf = (char *) realloc (buf, buf_sz)) == NULL)
	return p;

      if (p)
	strcat (buf, tmp);
      else
	strcpy (buf, tmp);

      free (exp_date);
      free (rev_date);
    }			     

  return buf;
}

#endif /* WITH_OPENSSL */








