/* pam_vsd module.
   Copyright (c) 2001 Designer Servers Ltd.  All rights reserved.
   Written by Nick Burrett <nick@dsvr.net>  */

#include <features.h>
#include <stdio.h>
#include <unistd.h>
#include <string.h>
#include <syslog.h>
#include <stdarg.h>
#include <sys/types.h>
#include <pwd.h>
#include <grp.h>

#include "../libvsd/libvsd.h"

/*
 * here, we make a definition for the externally accessible function
 * in this file (this definition is required for static a module
 * but strongly encouraged generally) it is used to instruct the
 * modules include file to define the function prototypes.
 */

#define PAM_SM_AUTH

#include <security/pam_modules.h>

/* variables */
static char check_priv[BUFSIZ];

/* some syslogging */
static void
_pam_log (int err, const char *format, ...)
{
  va_list args;

  va_start (args, format);
  openlog ("PAM-vsd", LOG_CONS | LOG_PID, LOG_AUTH);
  vsyslog (err, format, args);
  va_end (args);
  closelog ();
}

/* argument parsing */
#define PAM_PRIV_ARG  0x1
#define PAM_DEBUG_ARG 0x2
#define PAM_USER_ADMIN 0x4

static int
_pam_parse (int argc, const char **argv)
{
  int ctrl;

  /* step through arguments */
  for (ctrl = 0; argc-- > 0; ++argv)
    {
      /* generic options */
      if (! strcmp (*argv, "debug"))
	ctrl |= PAM_DEBUG_ARG;
      else if (! strcmp (*argv, "admin"))
	ctrl |= PAM_USER_ADMIN;
      else if (! strncmp (*argv, "priv=", 5))
	strcpy (check_priv, *argv + 5);
      else
	_pam_log (LOG_ERR, "pam_parse: unknown option; %s", *argv);
    }

  return ctrl;
}


/* --- authentication management functions (only) --- */
PAM_EXTERN int
pam_sm_authenticate (pam_handle_t * pamh, int flags, int argc,
		     const char **argv)
{
  int ctrl;
  const char *username;
  struct passwd *pwd;
  int retval = PAM_SUCCESS;

  /* Init the optional group */
  memset (check_priv, 0, sizeof (check_priv));

  ctrl = _pam_parse (argc, argv);
  retval = pam_get_user (pamh, &username, NULL);
  if (retval != PAM_SUCCESS || ! username)
    {
      if (ctrl & PAM_DEBUG_ARG)
	_pam_log (LOG_DEBUG, "can not get the username");
      return PAM_SERVICE_ERR;
    }

  /* su to a uid 0 account ? */
  pwd = getpwnam (username);
  if (!pwd)
    {
      if (ctrl & PAM_DEBUG_ARG)
	_pam_log (LOG_NOTICE, "unknown user %s", username);
      return PAM_USER_UNKNOWN;
    }

  /* Check `username' has privilege `check_priv'.  Set PAM_PERM_DENIED
     if it doesn't.  */
  if (* check_priv && vsd_priv_access (username, check_priv))
    {
      _pam_log (LOG_NOTICE, "user %s does not have %s privilege",
		username, check_priv);
      retval = PAM_PERM_DENIED;
    }

  /* If we didn't fail at the last hurdle, and we're checking whether
     the user was admin or not, set PAM_PERM_DENIED if it isn't.  */
  if ((ctrl & PAM_USER_ADMIN) && strcmp (username, "admin") != 0
      && retval == PAM_SUCCESS)
    {
      retval = PAM_PERM_DENIED;
      _pam_log (LOG_NOTICE, "%s is not the admin user", username);
    }

  return retval;
}

PAM_EXTERN int
pam_sm_setcred (pam_handle_t * pamh, int flags, int argc, const char **argv)
{
  return PAM_SUCCESS;
}


#ifdef PAM_STATIC
/* static module data */
struct pam_module _pam_vsd_modstruct = {
  "pam_vsd",
  pam_sm_authenticate,
  pam_sm_setcred,
  NULL,
  NULL,
  NULL,
  NULL,
};
#endif
