/* Apache configuration file manipulation.
   Copyright (c) 1999, 2000, 2001 Idaya Ltd.
   Contributed by Nick Burrett <nick@dsvr.net>

   This file is part of the Virtual Server Administrator (FreeVSD)

   FreeVSD is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2, or (at your option)
   any later version.

   FreeVSD is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with FreeVSD; see the file COPYING.  If not, write to
   the Free Software Foundation, Inc., 59 Temple Place - Suite 330,
   Boston, MA 02111-1307, USA.  */

#include "config.h"

#include <ctype.h>
#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <errno.h>
#include <unistd.h>
#include <dirent.h>
#include <sys/stat.h>

#include "vsd.h"
#include "libvsd.h"

/* Should be moved to config.h and configure.in.  */
#define HTTPD_CONF "/etc/httpd/conf/httpd.conf"

#define HTPASSWD_DIR "/etc/vsd/htpasswd"

static char **scan_virtual_hosts (struct vsd_file_line *head, int *total);
static int delete_decl (struct vsd_file_line *head, const char *virtual_host,
			const char *server_name, const char *token,
			const char *token_param);
static int add_decl (struct connection *vc, struct vsd_file_line *head,
		     const char *data, const char *token);
static struct vsd_file_line *find_virtual_host (struct vsd_file_line *head,
						const char *server,
						const char *servername);
static int check_vh_config (struct connection *vc, struct vsd_file_line *line,
			    struct vsd_file_line *last);
static struct vsd_file_line *find_definition (struct vsd_file_line *head,
					      const char *virtualhost,
					      const char *servername,
					      const char *token,
					      const char *token_param);

/* Generic httpd.conf processing functions.  */

/* Look in `line' for a token that contains the characters in `directive',
   and ends with the `terminator' character.  If `prefix' is not \0 then
   the line must start with that characters.  If found, return a malloc'd
   string consisting of the characters between the `directive' and the
   `terminator'.  Return NULL if not found.  */
static char *scan_directive (const char *line, char prefix,
			     const char *directive, char terminator)
{
  int count;
  const char *p;
  char *buffer;

  line = vsd_file_skip_spaces (line);

  if (prefix && *line++ != prefix)
    return NULL;

  if (strncmp (line, directive, strlen (directive)) != 0)
    return NULL;

  /* Skip past directive.  */
  line += strlen (directive);
  line = vsd_file_skip_spaces (line);

  /* Determine length of directive argument.  */
  for (count = 0, p = line; *p && *p != terminator; count++, p++)
    ;

  if (p == line)
    /* Nothing after the directive.  */
    return NULL;

  /* Strip trailing spaces.  */
  while (*p == '\0' || *p == terminator || isspace (*p))
    p--, count--;

  buffer = (char *) malloc (count + 2);
  strncpy (buffer, line, count + 1);
  buffer[count + 1] = '\0';

  return buffer;
} 

/* Delete an Apache style declaration.  With the following example:

     <VirtualHost foobar.com>
     ServerName www.foobar.com
     <Directory "/www/docs/foobar">
     Allow from all
     ...
     </Directory>
     DocumentRoot /www/docs
     </VirtualHost>

   If `token' and `token_param' are both NULL, then if `virtual_host'
   and `servername' are set to foobar.com and www.foobar.com respectively,
   then the whole VirtualHost declaration will be deleted.

   If `token' is set to Directory and `token_param' set to /www/docs/foobar
   then the <Directory>...</Directory> block will be deleted instead.

   Return 0 on success or -1 on error.  */
static int delete_decl (struct vsd_file_line *head,
			const char *virtual_host, const char *servername,
			const char *token, const char *token_param) 
{
  struct vsd_file_line *host_start, *host_end, *link;

  /* Point `host_start' to the start of the VirtualHost declaration.  */
  host_start = find_definition (head, virtual_host, servername,
				token, token_param);
  if (host_start == NULL)
    return -1;

  /* No token was given, so assume we are deleting a complete VirtualHost
     declaration.  */
  if (token == NULL)
    token = "VirtualHost";

  /* Point `host_end' to the end of the declaration.  */
  for (host_end = host_start; host_end; host_end = host_end->next)
    {
      const char *p = vsd_file_skip_spaces (host_end->line);
      if (p[0] == '<' && p[1] == '/'
	  && strncmp (p + 2, token, strlen (token)) == 0
	  && p[strlen (token) + 2] == '>')
	break;
    }

  if (host_end == NULL)
    {
      /* Closing </VirtualHost> couldn't be found.  */
      errno = ESRCH;
      return -1;
    }

  /* Unlink the declaration from the configuration file.  */
  link = host_start->prev;
  link->next = host_end->next;

  /* Remove next pointer and free the allocated memory.  */
  host_end->next = NULL;
  vsd_free_file (host_start);

  return 0;
}

/* Add a virtual host declaration at the next line after `head'.
   Perform some rudimentary syntax checking on the entry - people don't
   seem to be able to type correctly.

   Return 0 on success, -1 on error.  */
static int add_decl (struct connection *vc,
		     struct vsd_file_line *head,
		     const char *data, const char *token)
{
  struct vsd_file_line *first = head, *line;
  char *p, *ibuf;
  int c, c1;

  /* Read incoming data, a line at a time.  */
  while ((ibuf = io_getline ('\n', &data)))
    {
      p = ibuf;
      /* Skip whitespace and ignore blank lines.  */
      while (*p && isspace (*p))
	p++;

      if (*p)
	{
	  /* Clean up by removing trailing spaces.  */
	  char *q = p + strlen (p) - 1;
	  while (q > p && isspace (*q))
	    q--;
	  q[1] = '\0';

	  if (vsd_file_insert_line (head, p))
	    {
	      io_buffer_store (vc, "-ERR Out of memory");
	      return -1;
	    }
	  head = head->next;
	}
      free (ibuf);
    }

  /* Perform some simple checks on the data just received.  Most critically
     we need to ensure <VirtualHost> and </VirtualHost> are the first and
     last lines and that they are spelt correctly because the scanning
     code elsewhere in this source will otherwise not find them.  */
  first = first->next;

  /* Report an error if there are multiple VirtualHost decls.  */
  c = c1 = 0;
  for (line = first; line != head; line = line->next)
    {
      p = strstr (line->line, token);
      if (p && p[-1] == '<')  /* Check for <VirtualHost */
	c++;
      if (p && p[-1] == '/')  /* Check for /VirtualHost */
	c1++;
    }
  if (c > 1 || c1 > 1)
    {
      io_buffer_store (vc, "-ERR add_decl: Cannot store multiple declarations");
      return -1;
    }

  /* The first line should solely contain `<VirtualHost xxx>'.  */
  if (first->line[0] != '<'
      || strncmp (first->line + 1, token, strlen (token)) != 0
      || first->line[strlen (token) + 1] != ' ')
    {
      io_buffer_store (vc, "-ERR The first line must be `<%s ...>'", token);
      return -1;
    }
  if (first->line[strlen (first->line) - 1] != '>')
    {
      io_buffer_store (vc, "-ERR <%s line has missing `>' or contains junk",
		       token);
      return -1;
    }

  if (head->line[0] != '<' || head->line[1] != '/'
      || strncmp (head->line, token, strlen (token)) != 0
      || head->line[strlen (token) + 2] != '>')
    {
      /* No </VirtualHost>. A common problem is that the browser has
	 appended it to the end of the previous line.  */
      p = strstr (head->line, token);
      if (p && p[-1] == '/' && p[-2] == '<')
	{
	  char *temp = (char *) malloc (strlen (token) + 4);
	  sprintf (temp, "</%s>", token);
	  p[-2] = '\0';
	  vsd_file_insert_line (head, temp);
	  free (temp);
	}
      else
	{
	  io_buffer_store (vc, "-ERR The last line must be `</%s>'", token);
	  return -1;
	}
    }

  return check_vh_config (vc, first, head);
}


/* Apache VirtualHost hacking functions.  */

/* Fill `host_list' with a vector of virtual host entries.  */
static char **scan_virtual_hosts (struct vsd_file_line *head, int *total)
{
  struct vsd_file_line *entry;
  int x;
  char *vhost, *sname, **host_list;

  /* Count the number of <VirtualHost ...> entries that exist in the
     file.  */
  *total = 0;
  entry = head;
  while (entry)
    {
      if (strncmp (vsd_file_skip_spaces (entry->line),
		   "<VirtualHost", 12) == 0)
	*total = *total + 1;
      entry = entry->next;
    }

  if (*total == 0)
    return NULL;

  host_list = (char **) malloc (((*total) + 1) * sizeof (char *));
  x = 0;
  entry = head;
  while (entry)
    {
      /* Look for line containing <VirtualHost ...>.  */
      if ((vhost = scan_directive (entry->line, '<', "VirtualHost", '>')))
	{
	  int found = 0;
	  while (entry)
	    {
	      /* Within the VirtualHost decl. look for a ServerName decl.
	         Both the VirtualHost header and the ServerName decl are
	         used together to uniquify a VirtualHost decl.  */
	      if ((sname = scan_directive (entry->line, '\0',
					   "ServerName", '\n')))
		{
		  host_list[x] = (char *) malloc (strlen (vhost)
						  + strlen (sname) + 12);
		  sprintf (host_list[x], "vh=\"%s\" sn=\"%s\"", vhost, sname);
		  x++;
		  found = 1;
		  free (sname);
		  break;
		}
	      if (! strncmp (vsd_file_skip_spaces (entry->line),
			     "</VirtualHost>", 14))
		break;

	      entry = entry->next;
	    }
	  if (! found)
	    {
	      /* No ServerName found.  Use `unknown' instead.  */
	      host_list[x] = (char *) malloc (12 + strlen (vhost) + 1);
	      sprintf (host_list[x], "%s;unknown", vhost);
	      x++;
	    }

	  free (vhost);
	}
      entry = entry->next;
    }

  return host_list;
}

/* Extra features. Make sure pathname for DocumentRoot, ErrorLog
   TransferLog and CustomLog exists.

   Return 0 on success, -1 on error.  */
static int check_vh_config (struct connection *vc, struct vsd_file_line *line,
			    struct vsd_file_line *last)
{
  const char *tokens[] = { "DocumentRoot", "ErrorLog", "TransferLog",
			   "CustomLog", NULL };
  int dir;

  for (; line != last; line = line->next)
    {
      struct stat sb;
      char *p;
      const char *status;
      int ret, arglen, len, x;

      for (x = 0; tokens[x] && strncmp (line->line, tokens[x],
					strlen (tokens[x])) != 0; x++)
	;

      if (! tokens[x])
	continue;

      len = strlen (tokens[x]);

      /* Check there is at least one space between token and argument.  */
      if (line->line[len] != ' ' && line->line[len] != '\t')
	{
	  io_buffer_store (vc, "-ERR Need a space between directive and argument");
	  return -1;
	}


      /* Now check the file path exists.  */
      p = line->line + len;
      /* Skip spaces between directive and path.  */
      while (*p && isspace (*p))
	p++;
      dir = (strcmp (tokens[x], "DocumentRoot") == 0) ? 1 : 0;
      status = vsd_check_and_make_path (vc->server_root, p, dir);
      if (status)
	{
	  io_buffer_store (vc, "-ERR %s", status);
	  return -1;
	}
    }
  return 0;
}

/* Return a pointer to the line that starts to define the VirtualHost
   defined by `virtual_host' and `servername'.  If no such VirtualHost
   can be found then return NULL.  */
static struct vsd_file_line *find_virtual_host (struct vsd_file_line *head,
						const char *virtual_host,
						const char *servername)
{
  char *vhost, *sname;

  for (; head; head = head->next)
    if ((vhost = scan_directive (head->line, '<', "VirtualHost", '>')))
      if (strcmp (vhost, virtual_host) == 0)
	{
	  /* Found a matching VirtualHost. Now look for a matching
	     ServerName.  */
	  struct vsd_file_line *x = head->next;
	  int found = 0;

	  free (vhost);
	  for (x = head->next; x && ! found; x = x->next)
	    {
	      if ((sname = scan_directive (x->line, '\0', "ServerName", '\n')))
		{
		  char *p;

		  /* Strip trailing junk from the server name.  */
		  for (p = sname; *p && ! isspace (*p) && *p != '#'; p++)
		    ;
		  *p = '\0';

		  if (strcmp (servername, sname) == 0)
		    {
		      /* We have found a matching VirtualHost decl.  */
		      free (sname);
		      return head;
		    }

		  free (sname);
		}

	      if (! strncmp (vsd_file_skip_spaces (x->line),
			     "</VirtualHost>", 14))
		found = 1;
	    }

	  /* x is either NULL, or host_start points to a VirtualHost
	     with no ServerName.  */
	  if (found && ! strcmp (servername, "unknown"))
	    return head;
	}

  /* Couldn't find VirtualHost.  */
  errno = ENOENT;
  return NULL;
}

/* Given a syntax like:
     <VirtualHost foobar.com>
     ServerName www.foobar.com
     <Directory "/www/docs/foobar">
     Allow from all
     ...
     </Directory>
     DocumentRoot /www/docs
     </VirtualHost>

   This function will search for a VirtualHost declaration defined with
   the name `virtualhost' and the ServerName `servername'.  If successful
   it will return a pointer to the first line of that declaration.  On
   failure it will return NULL.

   If `token' and `token_param' are not NULL then this function will
   return a pointer to the first line of a `token' declaration with
   the argument `token_param'.  In the above example, the Directory
   decl would be returned if `token' was set to "Directory" and `token_param'
   was set to "/www/docs/foobar".  */
static struct vsd_file_line *find_definition (struct vsd_file_line *head,
					      const char *virtualhost,
					      const char *servername,
					      const char *token,
					      const char *token_param)
{
  struct vsd_file_line *p;
  char *arg;
  head = find_virtual_host (head, virtualhost, servername);
  if (head == NULL || token == NULL)
    return head;

  for (; head; head = head->next)
    if ((arg = scan_directive (head->line, '<', token, '>')))
      {
	int ret = strcmp (arg, token_param);
	free (arg);

	if (! ret)
	  return head;
      }

  /* Couldn't find `token'.  */
  errno = ENOENT;
  return NULL;
}


/* Return a list of VirtualHost declarations within httpd.conf.  */
void httpd_enumerate_virtual_hosts (struct connection *vc, int argc,
				    char *argv[])
{
  struct vsd_file_line *httpd_conf;

  if (vsd_load_file (&httpd_conf, vc->server_root, HTTPD_CONF))
    io_buffer_store (vc, "-ERR Cannot load httpd.conf: %m");
  else
    {
      int num_hosts;
      char **host_list = scan_virtual_hosts (httpd_conf, &num_hosts);

      if (num_hosts > 0)
	{
	  int x;
	  for (x = 0; x < num_hosts; x++)
	    {
	      io_buffer_store (vc, "%s\n", host_list[x]);
	      free (host_list[x]);
	    }

	  free (host_list);
	}
      vsd_free_file (httpd_conf);
    }
}

/* We can't cope with every syntax error that the user puts
   into httpd.conf.  Hopefully this will help though.  We save
   the updated httpd.conf as a different file, then run `httpd -t'
   on that file to make sure the syntax is OK.  If it is, then
   re-save over the original httpd.conf.  If it isn't then
   delete the temporary and complain.

   Return 0 on success, anything else on error.  */
static int check_syntax_with_httpd (struct connection *vc,
				    struct vsd_file_line *httpd_conf)
{
  unsigned int t1, t2;
  char template[128], template1[128], *httpd_path;
  int fd, status;

  srand (time (NULL));
  t1 = rand () % 1479349;
  t2 = rand () % 2340985;

  sprintf (template, "-t -D SSL -f /tmp/vsd%06u", t1);
  sprintf (template1, "/tmp/vsd%06u", t2);

  vsd_save_file (httpd_conf, vc->server_root, template + 13);

  /* Look for the httpd binary.  */
  if (access ("/usr/sbin/httpd", R_OK | X_OK) == 0)
    httpd_path = "/usr/sbin/httpd";
  else if (access ("/usr/libexec/httpd", R_OK | X_OK) == 0)
    httpd_path = "/usr/libexec/httpd";
  else if (access ("/bin/httpd", R_OK | X_OK) == 0)
    httpd_path = "/bin/httpd";
  else
    {
      io_buffer_store (vc, "-ERR Cannot find web server binary");
      return 1;
    }

  status = run_command (vc, vc->server_root, "web", httpd_path,
			template, NULL, template1);

  sprintf (template, "%s/tmp/vsd%06u", vc->server_root, t2);
  if (status)
    {
      /* Looks like we found an error.  Send httpd output back
	 to the user.  */
      char line[1024];
      FILE *stream;
      
      stream = fopen (template, "r");
      if (stream != NULL)
	{
	  int p = 0;
	  while (fgets (line, sizeof (line), stream))
	    io_buffer_store (vc, "%s", line), p = 1;
	  fclose (stream);

	  if (p == 0)
	    io_buffer_store (vc, "-ERR Webserver check failed. Perhaps admin user has exceeded it's quota");
	}
    }
  unlink (template);
  sprintf (template1, "%s/tmp/vsd%06u", vc->server_root, t1);
  unlink (template1);

  return status;
}

/* Protocol interface for deleting VirtualHost declarations.  */
void httpd_delete_virtual_host (struct connection *vc, int argc, char *argv[])
{
  struct vsd_file_line *httpd_conf;

  if (vsd_load_file (&httpd_conf, vc->server_root, HTTPD_CONF))
    io_buffer_store (vc, "-ERR Cannot load httpd.conf: %m");
  else
    {
      if (! delete_decl (httpd_conf, argv[0], argv[1], NULL, NULL))
	{
	  int status = check_syntax_with_httpd (vc, httpd_conf);
	  if (! status)
	    /* No errors found.  Save over original httpd.conf.  */
	    vsd_save_file (httpd_conf, vc->server_root, HTTPD_CONF);
	}
      vsd_free_file (httpd_conf);
    }
}

/* Protocol interface for adding VirtualHost declarations.  */
void httpd_add_virtual_host (struct connection *vc, int argc, char *argv[],
			     const char *data)
{
  struct vsd_file_line *httpd_conf;

  if (vsd_load_file (&httpd_conf, vc->server_root, HTTPD_CONF))
    io_buffer_store (vc, "-ERR Cannot load httpd.conf: %m");
  else
    {
      if (! add_decl (vc, vsd_file_tail (httpd_conf), data, "VirtualHost"))
	{
	  int status = check_syntax_with_httpd (vc, httpd_conf);
	  if (! status)
	    /* No errors found.  Save over original httpd.conf.  */
	    vsd_save_file (httpd_conf, vc->server_root, HTTPD_CONF);
	}
      vsd_free_file (httpd_conf);
    }
}

/* Protocol interface for modifying VirtualHost declarations.
   Basically we find and delete the old VirtualHost decl. and add the
   new VirtualHost decl in it's place.  */
void httpd_modify_virtual_host (struct connection *vc, int argc, char *argv[],
				const char *data)
{
  struct vsd_file_line *httpd_conf, *p;

  if (vsd_load_file (&httpd_conf, vc->server_root, HTTPD_CONF))
    io_buffer_store (vc, "-ERR Cannot load httpd.conf: %m");
  else
    {
      p = find_virtual_host (httpd_conf, argv[0], argv[1]);
      if (p == NULL)
	io_buffer_store (vc, "-ERR VirtualHost '%s' not found", argv[0]);
      else
	{
	  p = p->prev;
	  if (! delete_decl (httpd_conf, argv[0], argv[1], NULL, NULL) 
	      && ! add_decl (vc, p, data, "VirtualHost"))
	    {
	      int status = check_syntax_with_httpd (vc, httpd_conf);
	      if (! status)
		/* No errors found.  Save over original httpd.conf.  */
		vsd_save_file (httpd_conf, vc->server_root, HTTPD_CONF);
	    }
	}
      vsd_free_file (httpd_conf);
    }
}

/* Find and return a VirtualHost declaration.  */
void httpd_get_virtual_host (struct connection *vc, int argc, char *argv[])
{
  struct vsd_file_line *httpd_conf, *start, *end;

  if (vsd_load_file (&httpd_conf, vc->server_root, HTTPD_CONF))
    {
      io_buffer_store (vc, "-ERR Cannot load httpd.conf: %m");
      return;
    }

  start = find_virtual_host (httpd_conf, argv[0], argv[1]);
  if (start == NULL)
    io_buffer_store (vc, "-ERR VirtualHost `%s' (%s) not found",
		     argv[0], argv[1]);
  else
    {
      /* Find the end of the VirtualHost declaration.  */
      for (end = start; end; end = end->next)
	if (strncmp (vsd_file_skip_spaces (end->line), "</VirtualHost>",
		     14) == 0)
	  break;

      if (end == NULL)
	io_buffer_store (vc, "-ERR Could not find the end of the VirtualHost declaration");
      else /* Output VirtualHost definition.  */
	for (; start != end->next; start = start->next)
	  io_buffer_store (vc, "%s\n", start->line);
    }
  vsd_free_file (httpd_conf);
}


/* .htpasswd hacking  */

/* Return a list of existing .htpasswd files.  For ease of management,
   we will store all password files in a central location - HTPASSWD_DIR.  */
void apache_htp_grplist (struct connection *vc, int argc, char *argv[])
{
  DIR *dir;
  struct dirent *entry;
  char *path;

  /* APACHE_HTP_GRPLIST */

  /* Scan HTPASSWD_DIR and return a list of filenames.  */
  path = (char *) malloc (strlen (HTPASSWD_DIR)
			  + strlen (vc->server_root) + 3);
  strcpy (path, vc->server_root);
  strcat (path, "/" HTPASSWD_DIR);
  dir = opendir (path);
  if (dir == NULL)
    mkdir (path, 0750);
  else
    {
      while ((entry = readdir (dir)))
	{
	  if (entry->d_name[0] == '.')
	    continue;
	  io_buffer_store (vc, "%s\n", entry->d_name);
	}
      closedir (dir);
    }
  free (path);
}

void apache_htp_grpadd (struct connection *vc, int argc, char *argv[],
			const char *data)
{
  char *path;
  /* APACHE_HTP_GRPADD <group name> {
       user password
       user password
     } 
  */
  if (! isalnum (*argv[0]))
    {
      io_buffer_store (vc, "-ERR %s must start with an alphanumeric character",
		       argv[0]);
      return;
    }
  path = (char *) malloc (strlen (HTPASSWD_DIR) + strlen (argv[0]) + 3);
  if (path == NULL)
    io_buffer_store (vc, "-ERR Out of memory: %m");
  else
    {
      struct vsd_file_line *list;
      char *ibuf, line[512];
	  
      sprintf (path, "%s/%s", HTPASSWD_DIR, argv[0]);
      if (vsd_load_file (&list, vc->server_root, path))
	{
	  /* No file exists.  Create one.  */
	  list = (struct vsd_file_line *) malloc (sizeof (struct vsd_file_line));
	  list->line = strdup ("# Generated by FreeVSD");
	  list->next = NULL;
	}

      /* Add user/password combination.  */
      while ((ibuf = io_getline ('\n', &data)))
	{
	  int x = 0;
	  /* Skip leading spaces.  */
	  while (isspace (*ibuf))
	    ibuf++;
	  /* Skip blank lines.  */
	  if (*ibuf == '\0')
	    continue;
	  
	  while (! isspace (*ibuf) && *ibuf != '\0' && x < 255)
	    line[x++] = *ibuf++;
	  line[x++] = ':';
	  
	  /* Skip over-large entry (i.e. invalid input).  */
	  if (*ibuf == '\0' || x == 255)
	    continue;
	  
	  /* Skip blank spaces.  */
	  while (isspace (*ibuf))
	    ibuf++;
	  
	  /* Invalid entry - missing password parameter.  */
	  if (*ibuf == '\0')
	    continue;
	  
	  while (*ibuf != '\0' && x < 511)
	    line[x++] = *ibuf++;
	  line[x] = '\0';
	  /* Skip over-large entry (i.e. invalid input).  */
	  if (x == 511)
	    continue;
	  
	  /* Insert user:password combination into htpasswd file.  */
	  vsd_file_insert_line (vsd_file_tail (list), line); 	      
	}
      vsd_save_file (list, vc->server_root, path);
      vsd_free_file (list);
    }
  free (path);
}

/* Delete a .htpasswd file.  If a user name is given, then remove that
   user from the .htpasswd file.   */
void apache_htp_grpdel (struct connection *vc, int argc, char *argv[])
{
  char *path;
  /* APACHE_HTP_GRLDEL <group name> [<username>] */

  if (! isalnum (*argv[0]))
    {
      io_buffer_store (vc, "-ERR %s must start with an alphanumeric character",
		       argv[0]);
      return;
    }

  path = (char *) malloc (strlen (vc->server_root) + strlen (HTPASSWD_DIR)
			  + strlen (argv[0]) + 3);
  if (path == NULL)
    io_buffer_store (vc, "-ERR Out of memory: %m");
  else
    {
      if (argc == 1)
	{
	  /* No username supplied.  Just delete the whole file.  */
	  sprintf (path, "%s/%s/%s", vc->server_root, HTPASSWD_DIR, argv[0]);
	  if (unlink (path))
	    io_buffer_store (vc, "-ERR htp_grpdel: Cannot unlink %s: %m",
			     argv[0]);
	}
      else
	{
	  /* A user name supplied, load the file.  */
	  struct vsd_file_line *list, *line;
	  sprintf (path, "%s/%s", HTPASSWD_DIR, argv[0]);
	  if (vsd_load_file (&list, vc->server_root, path))
	    io_buffer_store (vc, "-ERR Cannot load %s: %m", argv[0]);
	  else
	    {
	      /* Find the user name.  */
	      line = vsd_file_find_entry (list, argv[1]);
	      if (! line)
		io_buffer_store (vc, "-ERR User %s did not exist in %s",
				 argv[1], argv[0]);
	      else
		{
		  /* Delete the user from the file.  */
		  list = vsd_file_delete_line (list, line);
		  /* Update the stored copy on disk.  */
		  vsd_save_file (list, vc->server_root, path);
		}
	      vsd_free_file (list);
	    }
	}
      free (path);
    }
}

/* Get a list of users contained within a .htpasswd file.  */
void apache_htp_grpget (struct connection *vc, int argc, char *argv[])
{
  char *path;

  /* APACHE_HTP_GRPGET <group name>  */
  if (! isalnum (*argv[0]))
    {
      io_buffer_store (vc, "-ERR %s must start with an alphanumeric character",
		       argv[0]);
      return;
    }

  path = (char *) malloc (strlen (HTPASSWD_DIR) + strlen (argv[0]) + 3);
  if (path == NULL)
    io_buffer_store (vc, "-ERR Out of memory: %m");
  else
    {
      struct vsd_file_line *list, *line;
      sprintf (path, "%s/%s", HTPASSWD_DIR, argv[0]);
      if (vsd_load_file (&list, vc->server_root, path))
	io_buffer_store (vc, "-ERR Cannot load %s: %m", argv[0]);
      else
	{
	  line = list;
	  while ((line = vsd_file_walk (line)))
	    {
	      char *p, user[256];
	      int x;

	      /* Skip blank lines and comments.  */
	      if (line->line[0] == '\0' || line->line[0] == '#')
		continue;
	      p = line->line;
	      while (isspace (*p))
		p++;
	      if (*p == '#')
		continue;
	      /* Retrieve user name content.  */
	      x = 0;
	      while (*p != ':' && *p != '\0' && x < 255)
		user[x++] = *p++;
	      user[x] = '\0';
	      io_buffer_store (vc, "%s\n", user);
	    }
	  /* Retrieve user list.  */
	  vsd_free_file (list);
	}
      free (path);
    }
}

/* Directory hacking.  */

/* List all <Directory>...</Directory> blocks defined within a VirtualHost
   declaration.  */
void apache_dir_list (struct connection *vc, int argc, char *argv[])
{
  /* APACHE_DIR_LIST <virtual host> <server name>  */
  struct vsd_file_line *httpd_conf, *line;

  if (vsd_load_file (&httpd_conf, vc->server_root, HTTPD_CONF))
    io_buffer_store (vc, "-ERR Cannot load httpd.conf: %m");
  else
    {
      line = find_definition (httpd_conf, argv[0], argv[1], NULL, NULL);
      if (line == NULL)
	io_buffer_store (vc, "-ERR VirtualHost `%s' not found", argv[0]);
      else
	{
	  for (; line; line = line->next)
	    {
	      char *tok;
	      if ((tok = scan_directive (line->line, '<', "Directory", '>')))
		{
		  io_buffer_store (vc, "%s\n", tok);
		  free (tok);
		}
	      if (! strncmp (vsd_file_skip_spaces (line->line),
			     "</VirtualHost>", 14))
		break;
	    }
	}
      vsd_free_file (httpd_conf);
    }
}

/* Add a new Directory decl. to an existing VirtualHost declaration.  */
void apache_dir_add (struct connection *vc, int argc, char *argv[],
		     const char *data)
{
  /* APACHE_DIR_ADD <vhost> <servername> {
        <Directory ...>
	...
	</Directory>
     }
  */
  struct vsd_file_line *httpd_conf, *line;

  if (vsd_load_file (&httpd_conf, vc->server_root, HTTPD_CONF))
    io_buffer_store (vc, "-ERR Cannot load httpd.conf: %m");
  else
    {
      line = find_definition (httpd_conf, argv[0], argv[1], NULL, NULL);
      if (! add_decl (vc, line, data, "Directory"))
	{
	  int status = 0; /*check_syntax_with_httpd (vc, httpd_conf);*/
	  if (! status)
	    /* No errors found.  Save over original httpd.conf.  */
	    vsd_save_file (httpd_conf, vc->server_root, HTTPD_CONF);
	}
      vsd_free_file (httpd_conf);
    }
}

/* Delete a Directory decl. from an existing VirtualHost decl.  */
void apache_dir_del (struct connection *vc, int argc, char *argv[])
{
  /* APACHE_DIR_DEL <vhost> <servername> <directory>  */
  struct vsd_file_line *httpd_conf, *p, *start, *end;

  if (vsd_load_file (&httpd_conf, vc->server_root, HTTPD_CONF))
      io_buffer_store (vc, "-ERR Cannot load httpd.conf: %m");
  else
    {
      if (! delete_decl (httpd_conf, argv[0], argv[1], "Directory", argv[2]))
	{
	  int status = 0; /*check_syntax_with_httpd (vc, httpd_conf);*/
	  if (! status)
	    /* No errors found.  Save over original httpd.conf.  */
	    vsd_save_file (httpd_conf, vc->server_root, HTTPD_CONF);
	}
      vsd_free_file (httpd_conf);
    }
}

/* Get the contents of a Directory decl from an existing VirtualHost decl.  */
void apache_dir_get (struct connection *vc, int argc, char *argv[])
{
  /* APACHE_DIR_GET <vhost> <servername> <directory>  */
  struct vsd_file_line *httpd_conf, *start, *end;

  if (vsd_load_file (&httpd_conf, vc->server_root, HTTPD_CONF))
    {
      io_buffer_store (vc, "-ERR Cannot load httpd.conf: %m");
      return;
    }

  start = find_definition (httpd_conf, argv[0], argv[1], "Directory", argv[2]);
  if (start == NULL)
    io_buffer_store (vc, "-ERR Directory block `%s' not found", argv[2]);
  else
    {
      /* Find the end of the Directory declaration.  */
      for (end = start; end; end = end->next)
	if (strncmp (vsd_file_skip_spaces (end->line),
		     "</Directory>", 14) == 0)
	  break;

      if (end == NULL)
	io_buffer_store (vc, "-ERR Could not find the end of the Directory declaration");
      else /* Output VirtualHost definition.  */
	for (; start != end->next; start = start->next)
	  io_buffer_store (vc, "%s\n", start->line);
    }
  vsd_free_file (httpd_conf);
}
