/* Global declarations.
   This file is part of WebDNS.  */
/*
 * VSD - Virtual Server Daemon
 * Copyright (c) 2000 - 2001 Idaya Ltd.
 *
 * VSD is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * VSD is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with VSD - see the COPYING file;  if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA.
 */

/*
 * See the AUTHORS file for a list of contributors to VSD.
 * See the ChangeLog files for a list of changes.
 */

#ifndef BIND_H
#define BIND_H

#define DB_READ 0
#define DB_WRITE 1
#define DB_WRCREAT 2

/* Domain name database. This contains an entry for every domain on
   the name server.  */
#ifdef HAVE_GDBM_H
#define BIND_DB "/var/named/bind.db"
#elif defined (HAVE_NDBM_H)
#define BIND_DB "/var/named/bind"
#endif

/* Domain exclusion database. If an entry is in this database then the
   user is not permitted from performing add/delete/modify operations
   on the domain.  */
#ifdef HAVE_GDBM_H
#define BIND_EXCL_DB "/var/named/bind-excl.db"
#elif defined (HAVE_NDBM_H)
#define BIND_EXCL_DB "/var/named/bind-excl"
#endif

enum resource_type
{
  rr_a,
  rr_cname,
  rr_gid,
  rr_hinfo,
  rr_mb,
  rr_mg,
  rr_minfo,
  rr_mr,
  rr_mx,
  rr_ns,
  rr_ptr,
  rr_soa,
  rr_uid
};

/* Resource records.  */
struct resource_rec
{
  char *domain; /* Zone name. Can be NULL.  */
  char *ttl; /* TTL for the record.  */
  char *class; /* Class (default IN)  */
  enum resource_type type; /* Record type i.e. NS, MX, CNAME, A etc.  */
  char *server; /* Server name.  */
  union
  {
    struct gid
    {
      int gid;
    } gid;

    struct hinfo
    {
      char *name1;
      char *name2;
    } hinfo;

    struct mx /* Mail eXchange.  */
    {
      int cost;
    } mx;

    struct soa /* Start Of Authority.  */
    {
      char *primary_master_ns; /* Primary master name server.  */
      char *mail_address; /* E-mail address of person in charge of record.  */
      unsigned int serial; /* Serial number.  */
      int refresh; /* Refresh.  */
      int retry; /* Retry.  */
      int expire; /* Expiry.  */
      int min_ttl; /* Minimum Time To Live.  */
    } soa;
  } extra;
};

struct zonefile
{
  int default_ttl;
  int rsvc; /* Number of resource records.  */
  struct resource_rec **rsv; /* There can be many resource records.  */
};

/* Definitions in domain.c  */
extern void list_domains (struct connection *vc, int argc, char *argv[]);
extern void domain_delete (struct connection *vc, int argc, char *argv[]);
extern void domain_get (struct connection *vc, int argc, char *argv[]);
extern void domain_add (struct connection *vc, int argc, char *argv[]);

extern void domain_modify (struct connection *vc, int argc, char *argv[],
			   const char *data);
extern void domain_fix (struct connection *vc, int argc, char *argv[]);
extern void domain_fixall (struct connection *vc, int argc, char *argv[]);
extern void domain_ipmod (struct connection *vc, int argc, char *argv[]);
extern void domain_ipmodall (struct connection *vc, int argc, char *argv[]);
extern void bind_named_gen (struct connection *vc, int argc, char *argv[]);
extern void bind_root_named_gen (struct connection *vc, int argc, char *argv[]);
extern void domain_addvs (struct connection *vc, int argc, char *argv[]);
extern void domain_deletevs (struct connection *vc, int argc, char *argv[]);
extern void domain_restorevs (struct connection *vc, int argc, char *argv[]);
extern void domain_modifyvs (struct connection *vc, int argc, char *argv[]);
extern void bind_disablevs (struct connection *vc, int argc, char *argv[]);
extern void bind_enablevs (struct connection *vc, int argc, char *argv[]);
extern void bind_set_ttl (struct connection *vc, int argc, char *argv[]);
extern void bind_reload (struct connection *vc, int argc, char *argv[]);

/* dbase.c  */
extern void bind_dbase_generate (struct connection *vc);
extern void *bind_dbase_open (int write, const char *db);
extern void bind_dbase_close (void *dbf);
extern int bind_dbase_lookup (void *dbf, const char *domain);
extern void bind_dbase_delete (void *dbf, const char *domain);
extern void bind_dbase_add (void *dbf, const char *domain);
extern void bind_dbase_excl_list (struct connection *vc, int argc,
				    char *argv[]);
extern void bind_dbase_excl_add (struct connection *vc, int argc,
				   char *argv[]);
extern void bind_dbase_excl_delete (struct connection *vc, int argc,
				      char *argv[]);
extern int bind_dbase_check (const char *db, const char *domain);

/* dir.c  */
extern int bind_dir_count_domains (struct connection *vc);
extern char **bind_dir_get_domains (struct connection *vc, int *domainc);
extern char *bind_dir_name (struct connection *vc, const char *domain);
extern const char *bind_dir_named_gen (struct connection *vc, char subdir,
					 int master, const char *ns_ip);
extern const char *bind_dir_domain_add (struct connection *vc,
					  const char *domain);
extern const char *bind_dir_domain_delete (struct connection *vc,
					     const char *domain);

/* named.c */
extern void bind_named_notify_load (struct connection *vc,
				      const char *domain);
extern void bind_named_notify_update (struct connection *vc,
					const char *domain);

/* parse.y */
extern void yyparse_init (FILE *input, struct zonefile *zone);
extern void yyparse_final (void);
extern int yyparse (void);

/* zone.c */
extern void zone_set_origin (const char *new_origin);
extern void zone_set_domain (const char *new_domain);
extern void zone_set_default_ttl (int new_ttl);

extern struct zonefile *zone_read (const char *zone_file);
extern int zone_save (struct zonefile *zone, const char *zone_file,
		      const char *domain_name);

extern int zone_add_resource_record (struct zonefile *zone_file, int pos,
				     enum resource_type type,
				     const char *origin, const char *class,
				     const char *server, ...);
extern int zone_del_resource_record (struct zonefile *zone_file,
				     enum resource_type type,
				     const char *origin, const char *class,
				     const char *server, ...);
extern int zone_find_resource_record (struct zonefile *zone_file,
				     enum resource_type type,
				     const char *origin, const char *class,
				     const char *server, va_list ap);


extern int zone_free_resource_record (struct resource_rec *rec);
extern void zone_free (struct zonefile *zone_file);
extern void zone_update_serial (struct zonefile *zone);
extern unsigned int zone_gen_serial (unsigned int old_serial);

extern const char *zone_type_to_name (enum resource_type type);
extern enum resource_type zone_name_to_type (const char *name);

#endif /* BIND_H */

