/* Directory structure management.

   This file is part of WebDNS. 
 */

/*
 * VSD - Virtual Server Daemon
 * Copyright (c) 2000 - 2001 Idaya Ltd.
 *
 * VSD is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * VSD is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with VSD - see the COPYING file;  if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA.
 */

/*
 * See the AUTHORS file for a list of contributors to VSD.
 * See the ChangeLog files for a list of changes.
 */

/* As the directory structure of WebDNS on the name servers becomes
   more and more complicated it is necessary to seperate directory and
   file management into its own unit.

   /etc/vsd.conf contains a list of all virtual servers that have
   domains on the name server.  The use of vsd.conf then dictates the
   directory structure that we can use because it allows for virtual servers
   to be on different partitions.

   Assuming we have an entry (in /etc/vsd.conf) like:
      mount 0 /var/named
   then name server directories will be created as subdirectories of
   this mount point.

   Due to errors that we discovered with very large named.conf files,
   each virtual server directory is further split into 26 subdirectories
   related to the 26 letters of the English alphabet.  For virtual server
   foo we would therefor have a directory structure that resembles:

     /var/named/vs/foo/f/db.fred.co.uk
     /var/named/vs/foo/f/named.conf
     /var/named/vs/foo/c/db.chuff.co.uk
     /var/named/vs/foo/c/named.conf
     /var/named/vs/foo/named.conf

   The top-level named.conf in each virtual server directory will simply
   reference the named.conf files in each subdirectory of the virtual
   server.  */

#include <ctype.h>
#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <dirent.h>
#include <errno.h>
#include <stdarg.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <unistd.h>

#include "module.h"
#include "vsd.h"
#include "bind.h"
#include "libvsd.h"

/* Return a count of domains that are owned by a virtual server.
   If there is an error, return -1.  */
int bind_dir_count_domains (struct connection *vc)
{
  DIR *dir;
  struct dirent *entry;
  int count = 0;

  dir = opendir (vc->server_root);
  if (dir == NULL)
    return -1;

  while ((entry = readdir (dir)))
    {
      char path[256];
      struct stat sb;

      /* Skip . and .. directories.  */
      if (entry->d_name[0] == '.')
	continue;

      sprintf (path, "%s/%s", vc->server_root, entry->d_name);
      if (! stat (path, &sb) && S_ISDIR (sb.st_mode))
	{
	  struct dirent *entry1;
	  DIR *dirc = opendir (path);
	  if (dirc == NULL)
	    continue;

	  while ((entry1 = readdir (dirc)))
	    {
	      /* Zone files are always of the format `db.<domain-name>'
		 e.g. `db.dsvr.co.uk' and `db.dsvr.net'.
		 Ignore zone files containing a ~ as these are considered
		 backups.  */
	      if (entry1->d_name[0] == 'd' && entry1->d_name[1] == 'b'
		  && entry1->d_name[2] == '.')
		count ++;
	    }
	  closedir (dirc);
	}
    }
  closedir (dir);
  return count;
}

/* Create an vector of domains (zone files) that have been created.  */
char **bind_dir_get_domains (struct connection *vc, int *domainc)
{
  DIR *dir;
  struct dirent *entry;
  int dc;
  char **dv;

  *domainc = -1;
  dir = opendir (vc->server_root);
  if (dir == NULL)
    return NULL;

  dv = NULL;
  dc = 0;
  while ((entry = readdir (dir)))
    {
      char path[256];
      struct stat sb;

      /* Skip . and .. directories.  */
      if (entry->d_name[0] == '.')
	continue;

      sprintf (path, "%s/%s", vc->server_root, entry->d_name);
      if (! stat (path, &sb) && S_ISDIR (sb.st_mode))
	{
	  struct dirent *entry1;
	  DIR *dirc = opendir (path);
	  if (dirc == NULL)
	    continue;

	  while ((entry1 = readdir (dirc)))
	    {
	      /* Zone files are always of the format `db.<domain-name>'
		 e.g. `db.dsvr.co.uk' and `db.dsvr.net'.
		 Ignore zone files containing a ~ as these are considered
		 backups.  */
	      if (entry1->d_name[0] == 'd' && entry1->d_name[1] == 'b'
		  && entry1->d_name[2] == '.')
		{
		  /* Does it have a backup suffix ?  */
		  if (! strchr (entry1->d_name, '~'))
		    dv = vsd_argv_add (dv, &dc, entry1->d_name + 3);
		}
	    }
	  closedir (dirc);
	}
    }
  closedir (dir);
  *domainc = dc;
  if (dc == 0)
    errno = ENOENT;
  return dv;
}

/* Given a domain name, `domain', create a canonicalised path name
   suitable for other functions to operate on the domain.  Return
   a malloc'ed string.  */
char *bind_dir_name (struct connection *vc, const char *domain)
{
  char *path;

  path = (char *) malloc (strlen (vc->server_root) 
			  + strlen (domain) + 7);
  if (path == NULL)
    return NULL;

  sprintf (path, "%s/%c/db.%s", vc->server_root, domain[0], domain);
  return path;
}

/* Generate or re-generate a named.conf file for a virtual server.
   If `subdir' is zero then create the virtual server named.conf
   that includes the named.confs of all subdirectories.  Parameters
   `master' and `pri_ns' are ignored in this case.

   If `subdir' is non-zero (i.e. an alphanumeric character) then create
   the named.conf that references the actual domain name entries.  The
   client must also additionally call dir_named_gen with subdir set to
   zero.  */
const char *bind_dir_named_gen (struct connection *vc, char subdir,
				  int master, const char *ns_ip)
{
  FILE *handle;
  struct stat sb;
  char *path;
  DIR *dir;
  struct dirent *entry;

  path = (char *) malloc (strlen (vc->server_root) + 128);
  if (path == NULL)
    return "cannot allocate temporary buffer";

  if (subdir == 0)
    {
      int has_subdir = 0;

      /* Generate the named.conf that indexes named.confs in a virtual
	 server subdirectory.  */
      sprintf (path, "%s/named.conf", vc->server_root);
      handle = fopen (path, "w");
      dir = opendir (vc->server_root);
      if (dir == NULL)
	return "cannot open virtual server directory";

      /* Read all directory entries.  */
      while ((entry = readdir (dir)))
	{
	  /* Skip . and .. directory entries.  It might confuse things
	     below.  */
	  if (entry->d_name[0] == '.')
	    continue;

	  /* Ensure the entry is a sub-directory.  */
	  sprintf (path, "%s/%s", vc->server_root, entry->d_name);
	  if (! stat (path, &sb) && S_ISDIR (sb.st_mode))
	    {
	      has_subdir = 1;

	      /* Ensure the sub-directory contains a named.conf.  */
	      sprintf (path, "%s/%s/named.conf", vc->server_root,
		       entry->d_name);
	      if (! stat (path, &sb))
		/* Add to named.conf.  */
		fprintf (handle, "include \"%s/%s/named.conf\";\n",
			 vc->server_root, entry->d_name);
	    }
	}

      fclose (handle);
      /* If there were no subdirectories, then delete the directory and
	 the rest of its contents.  */
      if (! has_subdir)
	{
	  rewinddir (dir);
	  while ((entry = readdir (dir)))
	    {
	      sprintf (path, "%s/%s", vc->server_root, entry->d_name);
	      remove (path);
	    }
	  /* Delete the virtual server directory.  */
	  rmdir (vc->server_root);
	}
      closedir (dir);
    }
  else
    {
      int has_domains = 0;

      /* Generate a named.conf for a sub-directory.  */
      sprintf (path, "%s/%c", vc->server_root, subdir);
      dir = opendir (path);
      if (dir == NULL)
	/* We don't report this because of bind_named_gen(). 
	   return "cannot open virtual server sub-directory"; */
	return NULL;
      strcat (path, "/named.conf");
      handle = fopen (path, "w");

      while ((entry = readdir (dir)))
	{
	  /* Zone files are always of the format `db.<domain-name>'
	     e.g. `db.dsvr.co.uk' and `db.dsvr.net'.
	     Ignore zone files containing a ~ as these are considered
	     backups.  */
	  if (entry->d_name[0] == 'd' && entry->d_name[1] == 'b'
	      && entry->d_name[2] == '.')
	    {
	      /* Does it have a backup suffix ?
		 At the moment we only recognise a file containing a `~'.  */
	      if (strchr (entry->d_name, '~'))
		continue;

	      /* Found a domain name.  */
	      has_domains = 1;

	      fprintf (handle, "zone \"%s\" in {\n", entry->d_name + 3);
	      fprintf (handle, "  type %s;\n", (master) ? "master" : "slave");
	      fprintf (handle, "  file \"%s/%c/%s\";\n", vc->server_root,
		       subdir, entry->d_name);
	      if (! master)
		fprintf (handle, "  masters { %s; };\n", ns_ip);
	      fprintf (handle, "};\n");
	    }
	}
      fclose (handle);

      /* If we didn't find any domain names, then delete the directory and
	 the rest of its contents.  */
      if (! has_domains)
	{
	  /* Delete named.conf.  */
	  remove (path);
	  rewinddir (dir);
	  while ((entry = readdir (dir)))
	    {
	      /* Skip . and .. directories.  */
	      if (entry->d_name[0] == '.')
		continue;

	      sprintf (path, "%s/%c/%s", vc->server_root, subdir,
		       entry->d_name);
	      remove (path);
	    }
	  /* Delete the sub-directory.  */
	  sprintf (path, "%s/%c", vc->server_root, subdir);
	  rmdir (path);
	}

      closedir (dir);
    }

  return NULL;
}

/* Add `domain' to a virtual server.  Create the directory structure
   if it doesn't already exist.  Return NULL on successful completion
   or an error string on failure.

   The callee should call dir_named_gen upon successful completion.  */
const char *bind_dir_domain_add (struct connection *vc, const char *domain)
{
  struct stat sb;
  char *path;

  path = (char *) malloc (strlen (vc->server_root)
			  + strlen ("/a/named.conf") + strlen (domain) + 2);
  if (path == NULL)
    return "cannot allocate temporary buffer";

  /* Make sure virtual server directory exists.  */
  if (stat (vc->server_root, &sb))
    if (mkdir (vc->server_root, 0755))
      {
	free (path);
	return "could not create virtual server directory";
      }

  /* The sub-directory that the domain should go in depends on the first
     character of the domain.  */
  sprintf (path, "%s/%c", vc->server_root, domain[0]);

  /* Create subdirectory if it doesn't already exist.  */
  if (stat (path, &sb))
    if (mkdir (path, 0755))
      {
	free (path);
	return "could not create virtual server sub-directory";
      }

  /* Create new domain name file, if it doesn't exist.  If it does exist
     then return an error as it obviously shouldn't exist.  All domains
     start with `db.'.  */
  strcat (path, "/db.");
  strcat (path, domain);
  if (stat (path, &sb))
    {
      /* Create domain.  */
      int fd = open (path, O_CREAT | O_WRONLY | O_TRUNC, 0644);
      if (fd == -1)
	{
	  free (path);
	  return "cannot create domain";
	}
      close (fd);

      /* Domain created.  Now create or re-generate a named.conf for this
	 subdirectory.  */
    }
  else
    {
      free (path);
      return "cannot create domain as it already exists";
    }

  free (path);
  return NULL;
}

/* Delete `domain' from a virtual server.  Don't do any directory
   manipulation - leave this task for bind_dir_named_gen, which must be
   called after this function anyway.  */
const char *bind_dir_domain_delete (struct connection *vc,
				      const char *domain)
{
  struct stat sb;
  char *path;

  path = (char *) malloc (strlen (vc->server_root)
			  + strlen ("/a/named.conf") + strlen (domain) + 2);
  if (path == NULL)
    return "cannot allocate temporary buffer";

  sprintf (path, "%s/%c", vc->server_root, domain[0]);
  if (stat (path, &sb))
    {
      free (path);
      return "virtual server sub-directory does not exist";
    }

  /* Delete domain.  We don't care if it never existed.  */
  sprintf (path, "%s/%c/db.%s", vc->server_root, domain[0], domain);
  if (! stat (path, &sb))
    {
      if (remove (path))
	{
	  free (path);
	  return "cannot delete zone file";
	}

      /* Success.  */
      free (path);
      return NULL;
    }

  free (path);
  return "cannot stat zone file";
}
