/* Domain/zone file manipulation.
   Copyright (c) 1999, 2000, 2001 Idaya Ltd.
   Contributed by Nick Burrett <nick@dsvr.net>

   This file is part of the Virtual Server Administrator (FreeVSD)

   FreeVSD is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2, or (at your option)
   any later version.

   FreeVSD is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with FreeVSD; see the file COPYING.  If not, write to
   the Free Software Foundation, Inc., 59 Temple Place - Suite 330,
   Boston, MA 02111-1307, USA.  */

/* See the AUTHORS file for a list of contributors to VSD.
   See the ChangeLog files for a list of changes.  */

/* Protocol implementation functions.  */

/* #define DEBUG */
#include "config.h"
#include <ctype.h>
#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <dirent.h>
#include <errno.h>
#include <stdarg.h>
#include <sys/stat.h>
#include <netdb.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <fcntl.h>
#include <unistd.h>
#include <sys/socket.h>
#ifdef DEBUG
#include <sys/syslog.h>
#endif

#include "module.h"
#include "vsd.h"
#include "bind.h"
#include "libvsd.h"


extern int vasprintf (char **ptr, const char *template, va_list ap);
static const char *validate_zone_file (struct zonefile *zone);

/* Append a dot to a domain name if one hasn't already been added.  */
static char *add_dot (const char *domain)
{
  char *name, *p;
  int dot, len = strlen (domain);

  dot = (domain[len - 1] == '.') ? 0 : 1;

  name = (char *) malloc (len + 1 + dot);
  for (p = name; *domain;)
    {
      /* Lower-case the domain.  */
      char c = *domain++;
      *p++ = tolower (c);
    }

  if (dot)
    *p++ = '.';
  *p = '\0';
  return name;
}

/* Domain names are case-insensitive.  Force them all to lower-case.
   This will stop stupid people adding multiple versions of the
   same domain name.  */
static char *lowercase (const char *domain)
{
  char *name, *p;
  int dot, len;

  if (domain == NULL)
    return NULL;

  len = strlen (domain);

  name = (char *) malloc (len + 1);
  for (p = name; *domain;)
    {
      char c = *domain ++;
      *p++ = tolower (c);
    }

  *p = '\0';
  return name;
}

/* Find out what type of name server we are.  Return 1 if we are a
   primary name server and 2 if we are a secondary name server.  We
   do this by obtaining the hostname and checking it against those
   that were stored in /etc/vsd.conf.  If no entry exists, then
   return 0.  */
static int what_are_we (struct connection *vc)
{
  char host[256];
  int x;
#if 0
  struct hostent *hent;
  struct in_addr *in;
  char *nsip;
#endif

  if (gethostname (host, sizeof (host) - 1))
    /* Cannot determine.  */
    return 0;

#if 0
  /* Resolve name server to an IP address.  */
  hent = gethostbyname (host);
  if (hent == NULL)
    return 0;

  in = (struct in_addr *) hent->h_addr_list[0];
  nsip = inet_ntoa (*in);

  if (strcmp (nsip, vc->map->global.ns1) == 0)
    return 1;

  if (strcmp (nsip, vc->map->global.ns2) == 0)
    return 2;
#else
  if (strcasecmp (host, vc->map->global.ns1) == 0)
    return 1;
  if (strcasecmp (host, vc->map->global.ns2) == 0)
    return 2;
#endif

  
  return 0;
}

#if 0
/* Check that `hostname' resolves.  Return 0 if the name resolves,
   -1 if there is an error and 1 if the name doesn't resolve.  */
static int host_exists (const char *hostname)
{
  struct hostent *ent;
  extern int h_errno;
  char *temp;
  int temp_len;

  temp_len = strlen (hostname);
  temp = (char *) alloca (temp_len + 1);
  strcpy (temp, hostname);
  /* Remove trailing dot.  */
  if (temp[temp_len - 1] == '.')
    temp[temp_len - 1] = '\0';

  ent = gethostbyname2 (hostname, AF_INET);
  if (ent == NULL)
    {
      if (h_errno == HOST_NOT_FOUND)
	{
	  errno = ENOENT;
	  return -1;
	}

      return 1;
    }

  return 0;
}
#endif

/* Translate a string `(null)' parameter to NULL.  */
static const char *param (const char *arg)
{
  return (strcmp (arg, "(null)") == 0) ? NULL : arg;
}

#if 0
/* See http://www.iana.net/cctld/cctld-whois.htm */
static const char *tlds[] = { "ac", "ad", "ae", "af", "ag", "ai", "al",
			      "am", "an", "ao", "aq", "ar", "as", "at",
			      "au", "aw", "az",
			      "ba", "bb", "bd", "be", "bf", "bg", "bh",
			      "bi", "bj", "bm", "bn", "bo", "br", "bs",
			      "bt", "bv", "bw", "by", "bz",
			      "ca", "cc", "cd", "cf", "cg", "ch", "ci" };
#endif

/* Do some basic syntax checking on a domain name.
   Return NULL if domain name is valid.
   Return a pointer to an error string if invalid.  */
static const char *check_domain_name (const char *name)
{
  int len, x, alpha;
  const char *p;

  if (name == NULL)
    return "invalid argument";

  len = strlen (name);
  /* a) Must start with a alphanumeric character
     b) Never have double periods.
     c) Cannot start/end with a `-' but can contains dashes but not
        consecutively.
     d) Maximum length is 63 characters.  This is difficult to enforce
        as there is no standard database of TLDs.  We workaround this
	by enforcing a 63 character limit between dots.
     e) Can only contain alphanumeric, dashes and dot characters.
     f) Must contain at least one alphabetic character.
  */
  if (! isalnum (name[0]))
    return "domain must start with an alpha-numeric character";

  if (! isalnum (name[len - 1]) && name[len - 1] != '.')
    return "domain must end with an alpha-numeric character or a dot";

  p = name;
  while (*p)
    {
      const char *q;
      for (q = p + 1; *q && *q != '.'; q++)
	;
      if (q - p >= 64)
	return "domain name exceeds maximum length of 63 characters";
      p = q;
    }

  /* if (len > 63)
     return "domain name exceeds maximum length of 63 characters";  */

  /* Check for invalid characters.  */
  alpha = 0;
  for (x = 0; x < len; x++)
    {
      if (! isalnum (name[x]) && name[x] != '-' && name[x] != '.')
	return "domain contains an invalid character";

      if (name[x] == '.' && name[x + 1] == '.')
	return "domain contains `..' sequence"; /* .. */

#if 0
      if (name[x] == '-' && name[x + 1] == '-')
	return "domain contains `--' sequence"; /* -- */
#endif

      if (isalpha (name[x]))
	alpha = 1;
    }

  /* A domain name must consist of characters other than just numbers.  */
  if (! alpha)
    return "domain looks like an IP address";

  /* A valid domain name.  */
  return NULL;
}

/* Do some simple syntax checking on an IP address.
   Return NULL if IP address is valid.
   Return a pointer to an error string if invalid.  */
static const char *check_ip_address (const char *ip)
{ 
  int len, x, y, i1, i2, i3, i4;

  if (ip == NULL)
    return "invalid argument";

  /* IP addresses can only contain numbers and dots.  */
  if (strcspn (ip, "0123456789."))
    return "address contains invalid characters";

  len = strlen (ip);
  /* There should be three dots. */
  for (y = 0, x = 0; x < len; x++)
    if (ip[x] == '.')
      {
	y++;
	if (ip[x + 1] == '.')
	  return "address contains `..' sequence";
      }     

  if (y != 3)
    return "there should be three dots in an IP address";

  /* It should start and end with a digit.  */
  if (! isdigit (ip[0]) && ! isdigit (ip[len]))
    return "address should start and end with a digit";

  /* Numbers should range 0 - 255.  */
  sscanf (ip, "%d.%d.%d.%d", &i1, &i2, &i3, &i4);
  if (i1 < 0 || i1 > 255 || i2 < 0 || i2 > 255
      || i3 < 0 || i3 > 255 || i4 < 0 || i4 > 255)
    return "address has an invalid value";

  return NULL;
}

/* Do some syntax checking on the origin field of a resource record.
   Return NULL if the origin is valid.
   Return a pointer to an error string if invalid.  */
static const char *check_origin (const char *origin, const char *domain)
{
  if (origin == NULL)
    return "invalid argument";

  /* `(null)' is used by the protocol to indicate that the origin
     is a blank field i.e. use the previously defied origin.  We
     accept this.  */ 
  if (strcmp (origin, "(null)") == 0)
    return NULL;

  if (origin[0] == '*')
    {
      /* Accept an origin consisting of a single asterisk - a wildcard
	 domain.  */
      if (origin[1] == '\0')
	return NULL;

      /* Only allow one wildcard per origin.  */
      if (strchr (origin + 1, '*'))
	return "only one asterisk (*) is allowed";

      if (origin[1] != '.')
	return "asterisk (*) must be followed by a dot (.)";

      /* Accept *.<domain>.  */
      if (strcmp (origin + 2, domain) == 0)
	return NULL;
    }

  if (origin[0] == '@')
    {
      /* Accept an origin consisting of a single `@'.  */
      if (origin[1] == '\0')
	return NULL;
      else
	return "the `@' must be freestanding";
    }

  /* By now we should have filtered out the special characters and be
     left with either:
          your-domain.co.uk.
     or   www
  */
  return check_domain_name (origin);
}

void list_domains (struct connection *vc, int argc, char *argv[])
{
  int domainc;
  char **domainv;

  /* BIND_LIST  */
  if (vc->vs->status == S_DISABLED)
    {
      vsd_log_error (VSD_LOG_INFO,
		     "attempted to list domain for disabled vs %s",
		     argv[0], vc->virtual_server);
      io_buffer_store (vc, "-ERR Cannot list domain: account disabled");
      return;
    }

  domainv = bind_dir_get_domains (vc, &domainc);
  if (domainv == NULL && errno == ENOENT)
    {
      /* We're assuming here that if the directory doesn't exist
	 then it must be a new account.  In which case, when the
	 account adds it's first domain, the relevant files will
	 be updated.  */
      return;
    }

  if (domainv == NULL)
    io_buffer_store (vc, "-ERR Couldn't read domain zone files: %m");
  else
    {
      while (--domainc >= 0)
	io_buffer_store (vc, "%s\n", domainv[domainc]);

      vsd_argv_free (domainv, domainc);
    }
}

void domain_delete (struct connection *vc, int argc, char *argv[])
{
  char *lcdomain = lowercase (argv[0]);

  /* BIND_DELETE <domain> */
  if (vc->vs->status == S_DISABLED)
    {
      vsd_log_error (VSD_LOG_INFO,
		     "attempted to delete domain %s from disabled vs %s",
		     lcdomain, vc->virtual_server);
      io_buffer_store (vc, "-ERR Cannot delete domain: account disabled");
      return;
    }

  /* Check exclusion database.  */
  vsd_log_error (VSD_LOG_INFO, "deleting domain %s for vs %s", lcdomain,
		 vc->virtual_server);

  if (bind_dbase_check (BIND_EXCL_DB, lcdomain) == 1)
    io_buffer_store (vc, "-ERR Cannot delete %s: operation not permitted",
		     lcdomain);
  else
    {
      const char *status = bind_dir_domain_delete (vc, lcdomain);
      if (status)
	io_buffer_store (vc, "-ERR Cannot delete %s: %s", lcdomain, status);
      else
	{
	  void *dbf = bind_dbase_open (DB_WRITE, BIND_DB);
	  if (dbf)
	    {
	      bind_dbase_delete (dbf, lcdomain);
	      bind_dbase_close (dbf);
	    }

	  /* Re-generate named.conf index files.  */
	  bind_named_gen (vc, 0, NULL);

	  /* Notify the name server that a domain has been deleted.  */
	  bind_named_notify_load (vc, lcdomain);
	}
    }
}


void domain_get (struct connection *vc, int argc, char *argv[])
{
  char *zonename;
  struct stat sb;
  struct zonefile *zone;
  int x;

  /* BIND_GET <domain name>  */
  if (vc->vs->status == S_DISABLED)
    {
      vsd_log_error (VSD_LOG_INFO,
		     "attempted to get domain %s from disabled vs %s",
		     argv[0], vc->virtual_server);
      io_buffer_store (vc, "-ERR Cannot get domain: account disabled");
      return;
    }

  zonename = bind_dir_name (vc, argv[0]);
  if (stat (zonename, &sb))
    {
      io_buffer_store (vc, "-ERR Zone file '%s' does not exist", argv[0]);
      free (zonename);
      return;
    }

  zone = zone_read (zonename);
  if (zone == NULL)
    {
      free (zonename);
      io_buffer_store (vc, "-ERR Error reading zone for domain %s", argv[0]);
      return;
    }

  /* Send data back to the client in the form:
       <domain> <class> <type> <server> [<type specific parameters>]

       Empty parameters are written as `(null)'.  */
  for (x = 0; x < zone->rsvc; x++)
    {
      struct resource_rec *rec = zone->rsv[x];

      /* Skip deleted records.  */
      if (rec == NULL)
	continue;

      io_buffer_store (vc, "%s %s ",
		       (rec->domain) ? rec->domain : "(null)",
		       (rec->class) ? rec->class : "(null)");
      switch (rec->type)
	{
	case rr_a:
	case rr_cname:
 	case rr_mb:
 	case rr_mg:
 	case rr_minfo:
 	case rr_mr:
 	case rr_ns:
 	case rr_ptr:
	  io_buffer_store (vc, "%s %s\n", zone_type_to_name (rec->type),
			   rec->server);
	  break;
	case rr_gid:
	case rr_uid:
	  io_buffer_store (vc, "%s %d\n", zone_type_to_name (rec->type),
			   rec->extra.gid.gid);
	  break;
	case rr_mx:
	  io_buffer_store (vc, "MX %d %s\n", zone->rsv[x]->extra.mx.cost,
			   zone->rsv[x]->server);
	  break;
	case rr_soa:
	  io_buffer_store (vc, "SOA %s %s %d %d %d %d %d\n",
		  zone->rsv[x]->extra.soa.primary_master_ns,
		  zone->rsv[x]->extra.soa.mail_address,
		  zone->rsv[x]->extra.soa.serial,
		  zone->rsv[x]->extra.soa.refresh,
		  zone->rsv[x]->extra.soa.retry,
		  zone->rsv[x]->extra.soa.expire,
		  zone->rsv[x]->extra.soa.min_ttl);
	  break;
	default:
	  io_buffer_store (vc, "resource rec: %s %s %s\n",
		  zone->rsv[x]->domain, zone->rsv[x]->class,
		  zone->rsv[x]->server);
	  break;
	}
    }

  zone_free (zone);
  free (zonename);
}

/* Add a domain to the primary or secondary nameserver.  This will
   generate a suitable zone file and re-generate named.conf indexes.  */
void domain_add (struct connection *vc, int argc, char *argv[])
{
  const char *status;
  char *lcdomain = NULL, *domain, *ipaddr;
  void *dbf;
  int master, noerror = 0, rootgen = 0;
  struct stat sb;
  struct zonefile *zone;
  char *mail_addr, *ns1, *ns2, *zonename;

  lcdomain = lowercase (argv[0]);
  domain = add_dot (lcdomain);
  ipaddr = argv[1];

  /* BIND_ADD <domain> <ip> [<e-mail addr>] */
  if (vc->vs->status == S_DISABLED)
    {
      vsd_log_error (VSD_LOG_INFO,
		     "attempted to add domain %s to disabled vs %s",
		     lcdomain, vc->virtual_server);
      io_buffer_store (vc, "-ERR Cannot add domain: account disabled");
      goto error;
    }

  if (bind_dir_count_domains (vc) == vc->vs->quota)
    {
      vsd_log_error (VSD_LOG_INFO,
		     "attempted to add domain %s to vs %s but quota exceeded",
		     lcdomain, vc->virtual_server);
      io_buffer_store (vc, "-ERR Cannot add domain: quota exceeded");
      goto error;
    }

  vsd_log_error (VSD_LOG_INFO, "adding domain %s for vs %s", lcdomain,
		 vc->virtual_server);

  /* Stupid people try and add sub-domains as proper zone files.  */
  if (strncmp (domain, "www.", 4) == 0
      || strncmp (domain, "ftp.", 4) == 0
      || strncmp (domain, "smtp.", 5) == 0
      || strncmp (domain, "mail.", 5) == 0)
    {
      io_buffer_store (vc, "-ERR %s is not a valid domain name", lcdomain);
      goto error;
    }

  /* Check domain name and IP address are valid.  */
  status = check_domain_name (domain);
  if (! status)
    status = check_ip_address (ipaddr);
  if (status)
    {
      io_buffer_store (vc, "-ERR Cannot add `%s' to name server: %s",
		       domain, status);
      goto error;
    }

  /* Check exclusion database.  */
  if (bind_dbase_check (BIND_EXCL_DB, lcdomain) == 1)
    {
      io_buffer_store (vc, "-ERR Cannot add %s to nameserver: operation not permitted",
		       lcdomain);
      goto error;
    }
  
  /* Does domain already exist ?  */
  if (bind_dbase_check (BIND_DB, lcdomain) == 1)
    {
      io_buffer_store (vc, "-ERR Cannot add %s to nameserver: domain already exists",
		       lcdomain);
      goto error;
    }

  master = what_are_we (vc);
  if (! master)
    {
      io_buffer_store (vc, "-ERR Cannot determine whether I am primary or secondary");
      goto error;
    }

  /* Add domain to the name server.  */
  
  /* Virtual servers can optionally have their own name server names.
     If they exist then use them, otherwise use the standard name server
     names.  */
  ns1 = add_dot ((*vc->vs->ns1) ? vc->vs->ns1 : vc->map->global.ns1);
  ns2 = add_dot ((*vc->vs->ns2) ? vc->vs->ns2 : vc->map->global.ns2);
  
  if (argc >= 3)
    {
      char *p;
      
      mail_addr = add_dot (argv[2]);
      /* If the user has passed in an ordinary e-mail address
	 then convert it.  */
      for (p = mail_addr; *p; p++)
	if (*p == '@')
	  *p = '.';
    }
  else
    {
      /* If no e-mail address is supplied then default to
	 hostmaster@primary.name.server.  */
      mail_addr = (char *) malloc (strlen (ns1) + 15);
      strcpy (mail_addr, "hostmaster.");
      strcat (mail_addr, ns1);
    }
  
  if (stat (vc->server_root, &sb))
    /* Virtual server directory doesn't exist, so we'll have to
       re-generate the root named.conf.  */
    rootgen = 1;
  
  status = bind_dir_domain_add (vc, lcdomain);
  if (status != NULL)
    {
      io_buffer_store (vc, "-ERR Cannot add domain %s: %s", domain,
		       status);
      free (mail_addr);
      free (ns1);
      free (ns2);
      goto error;
    }
  
  zonename = bind_dir_name (vc, lcdomain);
  zone = (struct zonefile *) malloc (sizeof (struct zonefile));
  memset (zone, 0, sizeof (struct zonefile));
  zone->default_ttl = 86400;
  
  /* Add a $ORIGIN record.  */
  zone_set_origin (domain);
  
  /* Add a SOA record.  */
  zone_add_resource_record (zone, -1, rr_soa, "@", "IN", NULL,
			    ns1,
			    mail_addr,
			    zone_gen_serial (-1),
			    10800 /* refresh */,
			    3600 /* retry */,
			    604800 /* expire */,
			    86400 /* minimum */);
  
  /* Add name server records.  */
  zone_add_resource_record (zone, -1, rr_ns, NULL, "IN", ns1);
  zone_add_resource_record (zone, -1, rr_ns, NULL, "IN", ns2);
  
  /* A MX record.  */
  zone_add_resource_record (zone, -1, rr_mx, NULL, "IN", domain, 5);
  
  /* An Address record.  */
  zone_add_resource_record (zone, -1, rr_a, NULL, "IN", ipaddr);
  
  /* Add CNAME records for www, smtp, mail and ftp.  */
  zone_add_resource_record (zone, -1, rr_a, "www", "IN", ipaddr);
  zone_add_resource_record (zone, -1, rr_cname, "smtp", "IN", "www");
  zone_add_resource_record (zone, -1, rr_cname, "mail", "IN", "www");
  zone_add_resource_record (zone, -1, rr_cname, "ftp", "IN", "www");
  
  zone_save (zone, zonename, domain);
  zone_free (zone);
  free (zonename);
  free (mail_addr);
  free (ns1);
  free (ns2);

  if (rootgen)
    bind_root_named_gen (vc, 0, NULL);

  /* Re-generate named.conf index files.  */
  bind_named_gen (vc, 0, NULL);

  /* Notify the name server that a new domain has been created.  The name
     server will then load it.  */
  bind_named_notify_load (vc, lcdomain);

  dbf = bind_dbase_open (DB_WRITE, BIND_DB);
  if (dbf)
    {
      bind_dbase_add (dbf, lcdomain);
      bind_dbase_close (dbf);
    }

  noerror = 1;

error:
  if (noerror == 0)
    vsd_log_error (VSD_LOG_INFO, "rejecting domain %s for vs %s", argv[0],
		   vc->virtual_server);

  free (domain);
  free (lcdomain);
}

/* Add a resource record to a zone file.  */
static int add_record (struct connection *vc, const char *domain,
		       struct zonefile *zone, int pos,
		       enum resource_type type, char **vec, size_t vecv)
{
  int s = 1;
  const char *st;
  char *lc_origin, *lc_server;

  /* All types have an origin.  */
  lc_origin = lowercase (vec[1]);
  if ((st = check_origin (lc_origin, domain)))
    {
      io_buffer_store (vc, "-ERR Origin error: %s", st);
      free (lc_origin);
      return -1;
    }

  /* Ensure origin is within domain.  */
  if (param (lc_origin) && lc_origin[strlen (lc_origin) - 1] == '.')
    {
      char *p = strstr (lc_origin, domain);

      if (p == NULL)
	{
	  io_buffer_store (vc, "-ERR Origin `%s' is outside domain `%s'",
			   lc_origin, domain);
	  return -1;
	}
      if (strcmp (p, domain) != 0)
	{
	  io_buffer_store (vc, "-ERR Origin `%s' is outside domain `%s': p = '%s'",
			   lc_origin, domain, p);
	  return -1;
	}
    }

  switch (type)
    {
    case rr_a:
      /* add <origin> <class> A <ip-address> */
      if ((st = check_ip_address (param (vec[4]))))
	{
	  io_buffer_store (vc, "-ERR IP address error: %s", st);
	  s = -1;
	}
      else
	{
	  s = zone_add_resource_record (zone, pos, type,
					param (lc_origin) /* origin */,
					param (vec[2]) /* class */,
					param (vec[4]) /* ip address */);
	}
      break;
    case rr_cname:
      /* add <origin> <class> CNAME <canonical domain name> */
      if (lc_origin[0] == '@'
	  || strcmp (lc_origin, "(null)") == 0
	  || strncmp (lc_origin, domain, strlen (lc_origin) - 1) == 0)
	{
	  io_buffer_store (vc, "-ERR CNAME and OTHER data error.  Use address (A) record.");
	  s = 1;
	}
      else if ((st = check_domain_name (param (vec[4]))))
	{
	  io_buffer_store (vc, "-ERR Invalid canonical domain name: %s", st);
	  s = -1;
	}
      else
	{
	  lc_server = lowercase (param (vec[4]));
	  s = zone_add_resource_record (zone, pos, type,
					param (lc_origin) /* origin */,
					param (vec[2]) /* class */,
					lc_server /* domain name */);
	  free (lc_server);
	}
      break;
    case rr_gid:
    case rr_uid:
      /* add <origin> <class> GID <server> <gid> */
      /* add <origin> <class> UID <server> <uid> */
      lc_server = lowercase (param (vec[4]));
      s = zone_add_resource_record (zone, pos, type,
				    param (lc_origin),
				    param (vec[2]),
				    lc_server,
				    atoi (param (vec[5])));
      free (lc_server);
      break;
    case rr_mx:
      {
	int pref = atoi (param (vec[4]));
	lc_server = lowercase (param (vec[5]));

	/* add <origin> <class> MX <preference> <exchange-dname> */
	if ((st = check_domain_name (lc_server)))
	  {
	    io_buffer_store (vc, "-ERR Invalid exchange domain name: %s", st);
	    s = -1;
	  }
	else
	  {
	    if (pref < 0 || pref > 65535)
	      {
		io_buffer_store (vc, "-ERR MX preference is out of range");
		s = -1;
	      }
	    else
	      s = zone_add_resource_record (zone, pos, type,
					    param (lc_origin) /* origin */,
					    param (vec[2]) /* class */,
					    lc_server /* exchange-name */,
					    atoi (param (vec[4])) /* pref */);
	  }
	free (lc_server);
      }
      break;
    case rr_hinfo:
      /* add <origin> <class> HINFO <server> <cpu> <os> */
      lc_server = lowercase (param (vec[4]));
      s = zone_add_resource_record (zone, pos, type,
				    param (lc_origin) /* origin */,
				    param (vec[2]) /* class */,
				    param (vec[4]) /* server */,
				    param (vec[5]) /* cpu */,
				    param (vec[6]) /* os */);
      free (lc_server);
      break;
    case rr_soa:
      /* add <origin> <class> SOA <server> <pri mast ns>
	 <mail addr> <serial> <refresh> <retry> <expire> <min_ttl> */
      lc_server = lowercase (param (vec[4]));
      if ((st = check_domain_name (param (vec[5]))))
	{
	  io_buffer_store (vc, "-ERR Invalid primary master name server domain name: %s",
			   st);
	  s = -1;
	}
      else
	s = zone_add_resource_record (zone, pos, type,
				      param (lc_origin) /* origin */,
				      param (vec[2]) /* class */,
				      param (vec[4]) /* server */,
				      param (vec[5]) /* pri mast ns */,
				      param (vec[6]) /* mail addr */,
				      atoi (param (vec[7])) /* serial */,
				      atoi (param (vec[8])) /* refresh */,
				      atoi (param (vec[9])) /* retry */,
				      atoi (param (vec[10])) /* expire */,
				      atoi (param (vec[11])) /* min_ttl */);
      free (lc_server);
      break;
    default:
      s = zone_add_resource_record (zone, pos, type,
				    param (lc_origin), param (vec[2]),
				    param (vec[4]),
				    (vecv >= 6) ? param (vec[5]) : NULL,
				    (vecv >= 7) ? param (vec[6]) : NULL,
				    (vecv >= 8) ? param (vec[7]) : NULL);
      break;
    }
  free (lc_origin);
  return s;
}

/* Delete a resource record from a zone file.  */
static int delete_record (struct connection *vc, const char *domain,
			  struct zonefile *zone,
			  enum resource_type type, char **vec, size_t vecv)
{
  int pos = -1;
  const char *st;

  /* All types have an origin.  */
  if ((st = check_origin (vec[1], domain)))
    {
      io_buffer_store (vc, "-ERR Origin error: %s", st);
      return -1;
    }

  switch (type)
    {
    case rr_a:
      /* delete <origin> <class> A <ip-address> */
      if ((st = check_ip_address (param (vec[4]))))
	io_buffer_store (vc, "-ERR IP address error: %s", st);
      else
	pos = zone_del_resource_record (zone, type,
					param (vec[1]) /* origin */,
					param (vec[2]) /* class */,
					param (vec[4]) /* ip address */);
      break;
    case rr_cname:
      /* delete <origin> <class> CNAME <canonical domain name> */
      if ((st = check_domain_name (param (vec[4]))))
	io_buffer_store (vc, "-ERR Invalid canonical domain name: %s", st);
      else
	pos = zone_del_resource_record (zone, type,
					param (vec[1]) /* origin */,
					param (vec[2]) /* class */,
					param (vec[4]) /* domain name */);
      break;
    case rr_gid:
    case rr_uid:
      /* delete <origin> <class> GID <server> <gid> */
      /* delete <origin> <class> UID <server> <uid> */
      pos = zone_del_resource_record (zone, type,
				      param (vec[1]) /* origin */,
				      param (vec[2]) /* class */,
				      param (vec[4]) /* server */,
				      atoi (param (vec[5])) /* gid/uid */);
      break;
    case rr_mx:
      {
	char *arg = vec[5];
	int pref = atoi (param (vec[4]));

	/* delete <origin> <class> MX <preference> <exchange-dname> */
	if ((st = check_domain_name (arg)))
	  io_buffer_store (vc, "-ERR Invalid exchange domain name: %s", st);
	else
	  {
	    if (pref < 0 || pref > 65535)
	      io_buffer_store (vc, "-ERR MX preference is out of range");
	    else
	      pos = zone_del_resource_record (zone, type,
					      param (vec[1]) /* origin */,
					      param (vec[2]) /* class */,
					      param (vec[5]) /* exchange-name */,
					      atoi (param (vec[4])) /* pref */);
	  }
      }
      break;
    case rr_hinfo:
      /* add <origin> <class> HINFO <server> <cpu> <os> */
      pos = zone_del_resource_record (zone, type,
				      param (vec[1]) /* origin */,
				      param (vec[2]) /* class */,
				      param (vec[4]) /* server */,
				      param (vec[5]) /* cpu */,
				      param (vec[6]) /* os */);
      break;
    case rr_soa:
      /* add <origin> <class> SOA <server> <pri mast ns>
	 <mail addr> <serial> <refresh> <retry> <expire> <min_ttl> */
      if ((st = check_domain_name (param (vec[5]))))
	io_buffer_store (vc, "-ERR Invalid primary master name server domain name: %s", st);
      else
	pos = zone_del_resource_record (zone, type,
					param (vec[1]) /* origin */,
					param (vec[2]) /* class */,
					param (vec[4]) /* server */,
					param (vec[5]) /* pri mast ns */,
					param (vec[6]) /* mail addr */,
					atoi (param (vec[7])) /* serial */,
					atoi (param (vec[8])) /* refresh */,
					atoi (param (vec[9])) /* retry */,
					atoi (param (vec[10])) /* expire */,
					atoi (param (vec[11])) /* min_ttl */);
      break;
    default:
      pos = zone_del_resource_record (zone, type,
				      param (vec[1]),
				      param (vec[2]),
				      param (vec[4]),
				      (vecv >= 6) ? param (vec[5]) : NULL,
				      (vecv >= 7) ? param (vec[6]) : NULL,
				      (vecv >= 8) ? param (vec[7]) : NULL);
      break;
    }
  return pos;
}

void domain_modify (struct connection *vc, int argc, char *argv[],
		    const char *data)
{
  char *zonename, *domain, *lcdomain;
  struct stat sb;
  struct zonefile *zone;
  int pos = -1, error = 0;
  char *ibuf;
  void *dbf;

  /* BIND_MODIFY <domain-name>  */
  lcdomain = lowercase (argv[0]);
  if (vc->vs->status == S_DISABLED)
    {
      vsd_log_error (VSD_LOG_INFO,
		     "attempted to modify domain %s in disabled vs %s",
		     lcdomain, vc->virtual_server);
      io_buffer_store (vc, "-ERR Cannot modify domain: account disabled");
      return;
    }

  /* Check exclusion database.  */
  dbf = bind_dbase_open (DB_READ, BIND_EXCL_DB);
  if (dbf)
    {
      int result = bind_dbase_lookup (dbf, lcdomain);
      bind_dbase_close (dbf);
      if (result)
	{
	  io_buffer_store (vc, "-ERR Cannot modify %s: operation not permitted",
			   lcdomain);
	  return;
	}
    }

  zonename = bind_dir_name (vc, lcdomain);
  if (stat (zonename, &sb))
    {
      io_buffer_store (vc, "-ERR Zone file %s does not exist", lcdomain);
      free (zonename);
      return;
    }

  zone = zone_read (zonename);
  if (zone == NULL)
    {
      free (zonename);
      io_buffer_store (vc, "-ERR Error reading zone file for domain `%s'",
		       lcdomain);
      return;
    }

  vsd_log_error (VSD_LOG_INFO, "modifying domain %s for vs %s", lcdomain,
		 vc->virtual_server);
  /* Set the default origin to be the domain name as this will help
     users who choose to use an abbreviated form of domain.  */
  domain = add_dot (lcdomain);
  zone_set_origin (domain);

  /* Read changes from the client.  */

  /* Client data comes in the form:
       <origin> <class> <type> <server> [<type specific parameters>]

     If a parameter is `(null)' then it is treated as empty.  */
  while ((ibuf = io_getline ('\n', &data)))
    {
      char **vec = NULL;
      size_t vecc = 0;
      enum resource_type type;
      char *p = ibuf + strlen (ibuf);

      error = 0;
      /* Strip trailing characters.  */
      while ((*p == '\n' || *p == '\r' || *p == '\0')
	     && p > ibuf)
	p--;

      *++p = '\0';
      vec = vsd_argv_parse (' ', ibuf, &vecc);
      if (vec == NULL || *(vec[0]) == '\0')
	error = 1; /* Ignore a blank line.  */

      if (! error && vecc <= 2)
	{
	  io_buffer_store (vc, "-ERR Not enough arguments\n");
	  error = 1;
	}

      if (! error && ((type = zone_name_to_type (vec[3])) == -1))
	{
	  io_buffer_store (vc, "-ERR %s is not a recognised resource record",
			   vec[3]);
	  error = 1;
	}

      if (! error)
	{
	  if (strcasecmp (vec[0], "add") == 0)
	    {
	      if (add_record (vc, domain, zone, -1, type, vec, vecc))
		error = 2; /* Fatal error.  */
	    }
	  else if (strcasecmp (vec[0], "replace") == 0)
	    {
	      if (add_record (vc, domain, zone, pos, type, vec, vecc))
		error = 2; /* Fatal error.  */
	    }
	  else if (strcasecmp (vec[0], "delete") == 0)
	    {
#ifdef DEBUG
	      syslog (LOG_WARNING, "delete: %s %s %s %s %s",
		      vec[1], vec[2], vec[3], vec[4], vec[5]);
#endif
	      pos = delete_record (vc, domain, zone, type, vec, vecc);
	      if (pos == -1)
		error = 2; /* Fatal error.  */
	    }
	  else
	    io_buffer_store (vc, "-ERR Unrecognised operation '%s'", vec[0]);
	}
      
      if (vec)
	free (vec);
      free (ibuf);
    }

  if (error != 2)
    {
      const char *err = validate_zone_file (zone);

      if (err)
	{
	  io_buffer_store (vc, "-ERR %s", err);
	  zone_free (zone);
	  free (domain);
	  free (lcdomain);
	  free (zonename);
	  return;
	}
      else
	{
	  zone_update_serial (zone);
	  zone_save (zone, zonename, domain);
	  /* Notify the name server that a domain has been modified.  */
	  bind_named_notify_update (vc, lcdomain);
	}
    }

  free (domain);
  free (lcdomain);
  free (zonename);
  zone_free (zone);
}

static const char *validate_zone_file (struct zonefile *zone)
{
  int x, y;

  for (x = 0; x < zone->rsvc; x++)
    {
      if (! zone->rsv[x])
	continue;

      /* Check for `MX points to a CNAME'.  */
      if (zone->rsv[x]->type == rr_mx)
	{
	  /* Check if exchange domain name is a CNAME.  */
	  for (y = 0; y < zone->rsvc; y++)
	    if (zone->rsv[y] && zone->rsv[y]->type == rr_cname
		&& strcmp (zone->rsv[x]->server, zone->rsv[y]->domain) == 0)
	      return "MX points to a CNAME";
	}

      /* Check for CNAME errors.  */
      if (zone->rsv[x]->type == rr_cname)
	{
	  for (y = 0 /* x + 1 */; y < zone->rsvc; y++)
	    {
	      if (y == x)
		continue;

	      /* Check if canonical domain name is a CNAME.  */
	      if (zone->rsv[y] && zone->rsv[y]->type == rr_cname
		  && strcmp (zone->rsv[x]->server, zone->rsv[y]->domain) == 0)
		return "CNAME points to a CNAME";

	      /* Check for duplicate CNAME entries.  */
	      if (zone->rsv[y] && zone->rsv[y]->type == rr_cname
		  && strcmp (zone->rsv[x]->domain, zone->rsv[y]->domain) == 0)
		return "Duplicate CNAME found";
	    }
	}
    }

  return NULL;
}

static int fix_zone (struct zonefile *zone)
{
  int x, y, z, found, zone_changed;

  zone_changed = 0;

  for (x = 0; x < zone->rsvc; x++)
    {
      if (zone->rsv[x]->type == rr_mx) /* MX record.  */
	{
	  /* Fix "MX points to a CNAME".

	     First, find an MX record, then check it points to a CNAME.
	     If it does then follow the CNAME chain until we end up at
	     an address record; replace server name in MX record with
	     domain in address record.  */
	  found = 0;

	  /* Check if exchange domain name is a CNAME.  */
	  for (y = 0; y < zone->rsvc && ! found; y++)
	    if (zone->rsv[y]->type == rr_cname
		&& strcmp (zone->rsv[x]->server, zone->rsv[y]->domain) == 0)
	      {
		/* Yes sir, it does.  */
	      next_match:
		for (z = 0; z < zone->rsvc && ! found; z++)
		  if (strcmp (zone->rsv[y]->server, zone->rsv[z]->domain) == 0)
		    {
		      /* Found the next match. */
		      if (zone->rsv[z]->type == rr_cname)
			{
			  /* Ah. Another CNAME, find the next match.  */
			  y = z;
			  goto next_match;
			}
		      else if (zone->rsv[z]->type == rr_a)
			{
			  /* Found an address record. Perfect.  */
			  found = 1;
			  free (zone->rsv[x]->server);
			  zone->rsv[x]->server = strdup (zone->rsv[z]->domain);
			  zone_changed = 1;
			}
		    }
	      }
	}
      else if (zone->rsv[x]->type == rr_soa) /* SOA record.  */
	{
	  /* Fix "SOA expire value is less than refresh + 10 * retry"
	     warning.

	     If the warning is found to be valid for this zone then change
	     the expire value to be refresh + 12 * retry.  */
	  if (zone->rsv[x]->extra.soa.expire <= (zone->rsv[x]->extra.soa.refresh
					   + zone->rsv[x]->extra.soa.retry * 10))
	    {
	      zone->rsv[x]->extra.soa.expire = (zone->rsv[x]->extra.soa.refresh
					  + zone->rsv[x]->extra.soa.retry * 12);
	      zone_changed = 1;
	    }

	  /* Ensure SOA expire value is 1 week or less.  */
	  if (zone->rsv[x]->extra.soa.expire != 604800)
	    {
	      zone->rsv[x]->extra.soa.expire = 604800;
	      zone_changed = 1;
	    }
	  /* Make the SOA retry time 2 hours.  */
	  if (zone->rsv[x]->extra.soa.retry != 7200)
	    {
	      zone->rsv[x]->extra.soa.retry = 7200;
	      zone_changed = 1;
	    }
	  /* Make the SOA refresh 4 hours.  */
	  if (zone->rsv[x]->extra.soa.refresh != 14400)
	    {
	      zone->rsv[x]->extra.soa.refresh = 14400;
	      zone_changed = 1;
	    }
	}

      /* Ensure the default TTL for a zone to 1 day. */
      if (zone->default_ttl != 86400)
	{
	  zone->default_ttl = 86400;
	  zone_changed = 1;
	}
    }

  return zone_changed;
}

/* Fix common errors in zone files.  */
void domain_fix (struct connection *vc, int argc, char *argv[])
{
  char *zonename, *domain;
  struct stat sb;
  struct zonefile *zone;
  int zone_changed;

  /* BIND_FIX <virtual server> <domain> */
  vc->virtual_server = strdup (argv[0]);
  vc->server_root = vsd_map_server_root (vc->map, vc->virtual_server);
  domain = add_dot (argv[1]);
  zonename = bind_dir_name (vc, argv[1]);
  if (stat (zonename, &sb))
    {
      io_buffer_store (vc, "-ERR Zone file %s does not exist", zonename);
      free (zonename);
      return;
    }

  zone = zone_read (zonename);
  if (zone == NULL)
    {
      io_buffer_store (vc, "-ERR Error reading zone");
      free (zonename);
      return;
    }

  /* If the zone file is changed by any of our fixing, then set to 1.  */
  zone_changed = fix_zone (zone);

  /* Only save the zone if it has been changed.  This saves
     unnecessary zone transfers.  */
  if (zone_changed)
    {
      zone_update_serial (zone);
      zone_save (zone, zonename, domain);
    }
  free (zonename);
  free (domain);
  zone_free (zone);
}

void domain_fixall (struct connection *vc, int argc, char *argv[])
{
  int domainc;
  char **domainv, *zonename;
  struct zonefile *zone;
  int zone_changed;

  /* BIND_FIXALL <virtual-serer>  */
  vc->virtual_server = strdup (argv[0]);
  vc->server_root = vsd_map_server_root (vc->map, vc->virtual_server);
  domainv = bind_dir_get_domains (vc, &domainc);
  if (domainv == NULL)
    {
      io_buffer_store (vc, "-ERR Could not get a list of domains for %s: %m",
		       vc->virtual_server);
      return;
    }

  while (--domainc >= 0)
    {
      zonename = bind_dir_name (vc, domainv[domainc]);
      zone = zone_read (zonename);
      if (zone == NULL)
	io_buffer_store (vc, "-ERR Error reading zone %s", domainv[domainc]);
      else
	{
	  /* Set the default origin to be the domain name as this will help
	     users who choose to use an abbreviated form of domain.  */
	  char *dom = add_dot (domainv[domainc]);

	  zone_changed = fix_zone (zone);
	  if (zone_changed)
	    {
	      zone_update_serial (zone);
	      zone_save (zone, zonename, dom);
	    }
	  free (dom);
	  zone_free (zone);
	}
      free (zonename);
    }
}

static void change_ip_addr (struct zonefile *zone,
			    const char *old_ip, const char *new_ip)
{
  int x;

  /* Zone read. Look at each resource record in turn, changing the
     IP address if required.  */
  for (x = 0; x < zone->rsvc; x++)
    {
      if (! zone->rsv[x])
	continue; /* Skip empty/deleted RRs.  */

      if (strcmp (zone->rsv[x]->server, old_ip) == 0)
	{
	  /* We have a match.  */
	  free (zone->rsv[x]->server);
	  zone->rsv[x]->server = strdup (new_ip);
	}
    }
}

/* Change the IP addresses in a zone file from `old-ip' to `new-ip'.  */
void domain_ipmod (struct connection *vc, int argc, char *argv[])
{
  char *zonename, *domain;
  const char *old_ip, *new_ip, *err;
  struct stat sb;
  struct zonefile *zone;

  /* BIND_IPMOD <virtual-server> <domain-name> <old-ip> <new-ip> */
  vc->virtual_server = strdup (argv[0]);
  vc->server_root = vsd_map_server_root (vc->map, vc->virtual_server);
  domain = add_dot (argv[1]);
  old_ip = argv[2];
  new_ip = argv[3];
  err = check_ip_address (old_ip);
  if (! err)
    err = check_ip_address (new_ip);

  if (err)
    {
      io_buffer_store (vc, "-ERR Error in IP addr %s: %s", old_ip, err);
      return;
    }

  zonename = bind_dir_name (vc, argv[1]);
  if (stat (zonename, &sb))
    {
      io_buffer_store (vc, "-ERR Zone file %s does not exist", zonename);
      return;
    }

  zone = zone_read (zonename);
  if (zone == NULL)
    io_buffer_store (vc, "-ERR Error reading zone");
  else
    {
      change_ip_addr (zone, old_ip, new_ip);

      zone_update_serial (zone);
      zone_save (zone, zonename, domain);
      zone_free (zone);
      free (zonename);
    }
  free (domain);
}

/* Change the IP address in all files belonging to `virtual-server'
   from `old-ip' to `new-ip'.  */
void domain_ipmodall (struct connection *vc, int argc, char *argv[])
{
  int domainc;
  char **domainv, *zonename;
  const char *server = argv[0];
  const char *old_ip = argv[1];
  const char *new_ip = argv[2];
  struct zonefile *zone;

  /* BIND_IPMODALL <virtual-server> <old-ip> <new-ip> */
  vc->virtual_server = strdup (server);
  vc->server_root = vsd_map_server_root (vc->map, vc->virtual_server);
  domainv = bind_dir_get_domains (vc, &domainc);
  if (domainv == NULL)
    {
      io_buffer_store (vc, "-ERR Could not get a list of domains for %s: %m",
		       server);
      return;
    }

  while (--domainc >= 0)
    {
      zonename = bind_dir_name (vc, domainv[domainc]);
      zone = zone_read (zonename);
      if (zone == NULL)
	io_buffer_store (vc, "-ERR Error reading zone %s", domainv[domainc]);
      else
	{
	  char *dom = add_dot (domainv[domainc]);
	  change_ip_addr (zone, old_ip, new_ip);

	  zone_update_serial (zone);
	  zone_save (zone, zonename, dom);
	  zone_free (zone);
	  free (dom);
	}
      free (zonename);
    }
}

/* From the zone files in a dns directory, regenerate the
   named.conf for the primary or secondary name server.  */
void bind_named_gen (struct connection *vc, int argc, char *argv[])
{
  int master;
  char subdir, *ns_ip = NULL;
  const char *status;
  struct stat sb;

  /* BIND_NAMEDGEN */
  master = what_are_we (vc);
  if (! master)
    {
      io_buffer_store (vc, "-ERR Badly formed argument");
      return;
    }
  if (master == 2)
    {
      struct hostent *hent;
      struct in_addr *in;
      char *ns1;

      /* Resolve primary name server to an IP address for passing to
	 bind_dir_named_gen.  It is preferable to do it here rather
	 than in bind_dir_named_gen because of the number of times
	 the function will get called.  A lot of time could be consumed
	 waiting for a name server to resolve the domain - especially if
	 the primary is down.  */

      /* Resolve the name server to an IP address.  */
      ns1 = (*vc->vs->ns1) ? vc->vs->ns1 : vc->map->global.ns1;
      hent = gethostbyname (ns1);
      if (hent == NULL)
	{
	  io_buffer_store (vc, "-ERR Cannot resolve primary name server");
	  return;
	}

      in = (struct in_addr *) hent->h_addr_list[0];
      ns_ip = inet_ntoa (*in);

      master = 0;
    }

  status = NULL;
  for (subdir = 'a'; subdir <= 'z' && ! status; subdir ++)
    status = bind_dir_named_gen (vc, subdir, master, ns_ip);
  for (subdir = '0'; subdir <= '9' && ! status; subdir ++)
    status = bind_dir_named_gen (vc, subdir, master, ns_ip);

  if (status)
    io_buffer_store (vc, "-ERR Re-gen of VS named.conf failed: %s",
		     status);
  else
    {
      status = bind_dir_named_gen (vc, 0, master, ns_ip);
      if (status)
	io_buffer_store (vc, "-ERR Re-gen of index named.conf failed: %s",
			 status);
    }  
}

/* From the zone files in a dns directory, regenerate the
   named.conf for the primary or secondary name server.  */
void bind_root_named_gen (struct connection *vc, int argc, char *argv[])
{
  FILE *stream = fopen ("/var/named/named.conf", "w");
  DIR *dir;
  struct dirent *entry;
  struct stat sb;
  char path[256];
  int x;

  if (stream == NULL)
    {
      io_buffer_store (vc, "-ERR Cannot open /var/named/named.conf for writing: %m");
      return;
    }
  for (x = 0; x < vc->map->partitions; x++)
    {
      DIR *dir;
      struct dirent *entry;

      sprintf (path, "%s/vs", vc->map->partition[x].mount);
      dir = opendir (path);
      while ((entry = readdir (dir)))
        {
	  if (entry->d_name[0] == '.')
	    continue;
          sprintf (path, "%s/vs/%s", vc->map->partition[x].mount,
		   entry->d_name);
          if (! stat (path, &sb) && S_ISDIR (sb.st_mode))
	    {
	      sprintf (path, "%s/vs/%s/named.conf",
		       vc->map->partition[x].mount, entry->d_name);
	      if (! stat (path, &sb) && S_ISREG (sb.st_mode))
	        fprintf (stream, "include \"%s/vs/%s/named.conf\";\n",
			 vc->map->partition[x].mount, entry->d_name);
    	    }
	}
      closedir (dir);
    }
  fclose (stream);
}

/* Add a new virtual server.  This registers the virtual server for
   use of the name servers.  */
void domain_addvs (struct connection *vc, int argc, char *argv[])
{
  struct vsd_vs vs;
  const char *ns1, *ns2;

  /* BIND_ADDVS <virtual server> <ip> <max domains> [ns1] [ns2] */

  memset (&vs, 0, sizeof (struct vsd_vs));
  strcpy (vs.name, argv[0]);
  vs.ip = malloc (sizeof (struct in_addr));
  vsd_ip_aton (argv[1], vs.ip);
  vs.startuid = 0;
  vs.users = 0;
  vs.status = S_ACTIVE;
  vs.quota = atoi (argv[2]);
  if (argc >= 4)
    strcpy (vs.ns1, argv[3]);
  if (argc >= 5)
    strcpy (vs.ns2, argv[4]);

  if (vsd_map_add_server (vc->map, &vs))
    io_buffer_store (vc, "-ERR Cannot create account: %s",
		     vsd_log_get_error (0));
  else
    vsd_map_save (vc->map);
}

/* Delete all domains belonging to a virtual server.  */
void domain_deletevs (struct connection *vc, int argc, char *argv[])
{
  struct stat sb;
  const char *path;

  /* BIND_DELETEVS <virtual server> */

  /* Perform a recursive delete on the virtual server directory.  */
  path = vsd_map_server_root (vc->map, argv[0]);
  if (stat ("/var/named/terminated-accounts", &sb))
    mkdir ("/var/named/terminated-accounts", 0755);

  if (path && ! stat (path, &sb))
    {
      char path1[256];

      sprintf (path1, "/var/named/terminated-accounts/%s", argv[0]);
      if (rename (path, path1))
	io_buffer_store (vc, "-ERR Cannot rename vs: %m");
    }

  /* Delete from main named.conf index.  */
  bind_root_named_gen (vc, 0, NULL);
  vsd_map_delete_server (vc->map, argv[0]);
  vsd_map_save (vc->map);
}

/* Restore a previously deleted virtual server.  */
void domain_restorevs (struct connection *vc, int argc, char *argv[])
{
  struct vsd_vs vs;
  const char *ns1, *ns2;

  /* BIND_RESTOREVS <virtual server> <ip> <max domains> [ns1] [ns2] */

  memset (&vs, 0, sizeof (struct vsd_vs));
  strcpy (vs.name, argv[0]);
  vs.ip = malloc (sizeof (struct in_addr));
  vsd_ip_aton (argv[1], vs.ip);
  vs.startuid = 0;
  vs.users = 0;
  vs.status = S_ACTIVE;
  vs.quota = atoi (argv[2]);
  if (argc >= 4)
    strcpy (vs.ns1, argv[3]);
  if (argc >= 5)
    strcpy (vs.ns2, argv[4]);

  if (vsd_map_add_server (vc->map, &vs))
    io_buffer_store (vc, "-ERR Cannot create account: %s",
		     vsd_log_get_error (0));
  else
    {
      const char *path;
      struct stat sb;

      vsd_map_save (vc->map);

      /* Perform a recursive delete on the virtual server directory.  */
      path = vsd_map_server_root (vc->map, argv[0]);
      if (stat ("/var/named/terminated-accounts", &sb))
	io_buffer_store (vc, "-ERR No terminated-accounts directory");
      else
	{
	  char path1[256];
	  
	  sprintf (path1, "/var/named/terminated-accounts/%s", argv[0]);
	  if (rename (path1, path))
	    io_buffer_store (vc, "-ERR Cannot rename vs: %m");
	}

      /* Add to main named.conf index.  */
      bind_root_named_gen (vc, 0, NULL);
    }
}

/* Modify a virtual server definition.  */
void domain_modifyvs (struct connection *vc, int argc, char *argv[])
{
  const char *status;
  char *path, *oldpath;
  int result, ch;
  char *vs = NULL, *new_vs = NULL, *new_ip = NULL;
  char *pri_ns = NULL, *sec_ns = NULL, *quota = NULL;
  char *old_ip = NULL;

  /* BIND_MODIFYVS -v <vs> [-w <vs>] [-i <ip>] [-p <ns>]
		     [-s <ns>] [-q <quota>] */
  /* Get options.  */
  optind = 0;
  while ((ch = getopt (argc + 1, argv - 1, "v:w:i:p:s:q:")) != -1)
    switch (ch)
      {
      case 'v':
	vs = optarg;
	break;
      case 'w':
	new_vs = optarg;
	break;
      case 'i':
	new_ip = optarg;
	break;
      case 'p':
	pri_ns = optarg;
	break;
      case 's':
	sec_ns = optarg;
	break;
      case 'q':
	quota = optarg;
	break;
      default:
	io_buffer_store (vc, "-ERR Bad command argument");
	return;
	break;
      }

  /* Check arguments.  */
  if (! vs)
    {
      io_buffer_store (vc, "-ERR -v argument is missing");
      return;
    }
  if ((status = vsd_valid_vsname (vs)))
    {
      io_buffer_store (vc, "-ERR VS name is invalid: %s", status);
      return;
    }

  if (new_vs && (status = vsd_valid_vsname (vs)))
    {
      io_buffer_store (vc, "-ERR New VS name is invalid: %s", status);
      return;
    }
  if (new_vs && strcmp (vs, new_vs) == 0)
    {
      io_buffer_store (vc, "-ERR VS and new VS are the same");
      return;
    }

  vc->vs = vsd_getmapent (vc->map, vs, NULL);	  
  old_ip = vsd_ip_ntoa (vc->vs->ip);

  if (new_ip && strcmp (old_ip, new_ip) == 0)
    {
      io_buffer_store (vc, "-ERR VS IP and new IP are the same");
      return;
    }
  if (pri_ns && strlen (pri_ns) >= 32)
    {
      io_buffer_store (vc, "-ERR Primary name server hostname is too long");
      return;
    }
  if (sec_ns && strlen (sec_ns) >= 32)
    {
      io_buffer_store (vc, "-ERR Secondary name server hostname is too long");
      return;
    }
  path = vsd_map_server_root (vc->map, vs);
  if (! path)
    {
      io_buffer_store (vc, "-ERR Directory %s does not exist", vs);
      return;
    }

  if (new_vs)
    {
      /* Virtual server directory name is changing.  Rename the directory
	 then update all named.conf files that might reference this
	 directory.  */
      oldpath = strdup (path);
      result = vsd_map_rename_server (vc->map, vs, new_vs);
      if (! result)
	path = vsd_map_server_root (vc->map, new_vs);
      if (result || ! path)
	{
	  io_buffer_store (vc, "-ERR Directory %s already exists", new_vs);
	  free (oldpath);
	  return;
	}
      
      /* Rename virtual server directory.  */
      result = rename (oldpath, path);
      vc->server_root = strdup (path);
      vc->virtual_server = strdup (new_vs);
      free (oldpath);

      if (result)
	io_buffer_store (vc, "-ERR Cannot rename vs: %m");
      else  
	{
	  vsd_map_save (vc->map);

	  vc->vs = vsd_getmapent (vc->map, vc->virtual_server, NULL);	  
	  /* Re-generate named.conf within local directory.  */
	  bind_named_gen (vc, 0, NULL);
	  
	  /* Update main named.conf index.  */
	  bind_root_named_gen (vc, 0, NULL);
	}
      vs = new_vs;
    }
  else
    vc->virtual_server = strdup (vs);

  if (new_ip)
    {
      /* Virtual server IP address is changing.  */
      int domainc;
      char **domainv;

      old_ip = strdup (old_ip);
      /* Change the entry in /etc/vsd.conf.  */
      result = vsd_map_change_server_ip (vc->map, vs, old_ip, new_ip);
      if (! result)
	vsd_map_save (vc->map);
      else if (result == 1)
	{
	  io_buffer_store (vc, "-ERR Cannot change IP: old IP not found");
	  free (old_ip);
	  return;
	}
      else if (result == 2)
	{
	  io_buffer_store (vc, "-ERR Cannot change IP: new IP in use");
	  free (old_ip);
	  return;
	}

      /* Change the IP address of all domains belong to that virtual
	 server.  */
      vc->server_root = vsd_map_server_root (vc->map,
					     vc->virtual_server);
      domainv = bind_dir_get_domains (vc, &domainc);
      if (domainv == NULL)
	{
	  io_buffer_store (vc,
			   "-ERR Cannot change IP: cannot get domain list: %m",
			   argv[1]);
	  free (old_ip);
	  return;
	}

      while (--domainc >= 0)
	{
	  char *zonename = bind_dir_name (vc, domainv[domainc]);
	  struct zonefile *zone = zone_read (zonename);

	  if (zone == NULL)
	    io_buffer_store (vc,
			     "-ERR Cannot change IP: Error reading zone %s",
			     domainv[domainc]);
	  else
	    {
	      char *dom = add_dot (domainv[domainc]);
	      change_ip_addr (zone, old_ip, new_ip);
	      zone_update_serial (zone);
	      zone_save (zone, zonename, dom);
	      zone_free (zone);
	      free (dom);
	      /* Notify the name server that a domain has been
		 modified.  */
	      bind_named_notify_update (vc, domainv[domainc]);
	    }
	  free (domainv[domainc]);
	  free (zonename);
	}
      free (domainv);
      free (old_ip);
    }

  if (pri_ns)
    strcpy (vc->vs->ns1, pri_ns);
  if (sec_ns)
    strcpy (vc->vs->ns2, sec_ns);
  if (quota)
    {
      int q = atoi (quota);
      if (q < 0)
	q = 0;
      vc->vs->quota = q;
    }
  if (pri_ns || sec_ns || quota)
    vsd_map_save (vc->map);
}

/* Disable a previously active account.  This will prevent a virtual
   server user from adding, deleting or modifying domains in the name
   servers.  */
void bind_disablevs (struct connection *vc, int argc, char *argv[])
{
  const char *status = vsd_valid_vsname (argv[0]);
  if (status)
    io_buffer_store (vc, "-ERR VS name is invalid: %s", status);
  else
    {
      if (vsd_map_set_status (vc->map, argv[0], "disabled"))
	io_buffer_store (vc, "-ERR Could not find server %s: %m", argv[0]);
      else
	vsd_map_save (vc->map);
    }
}

/* Activate a previously disabled account.  This will allow a virtual
   server user access to the WebDNS protocol.  */
void bind_enablevs (struct connection *vc, int argc, char *argv[])
{
  const char *status = vsd_valid_vsname (argv[0]);
  if (status)
    io_buffer_store (vc, "-ERR VS name is invalid: %s", status);
  else
    {
      if (vsd_map_set_status (vc->map, argv[0], "active"))
	io_buffer_store (vc, "-ERR Could not find server %s: %m", argv[0]);
      else
	vsd_map_save (vc->map);
    }
}

/* Change the TTL of all domains beloging to a virtual server.  */
void bind_set_ttl (struct connection *vc, int argc, char *argv[])
{
  /* BIND_SETTTL <ttl> */
  char **domainv;
  int domainc;
  int new_ttl = atoi (argv[0]);

  if (new_ttl < 0 || new_ttl > 604800)
    {
      io_buffer_store (vc, "-ERR TTL must be in the range 0 .. 604800 seconds\n");
      return;
    }

  /* Change the IP address of all domains belong to that virtual
     server.  */
  domainv = bind_dir_get_domains (vc, &domainc);
  if (domainv == NULL)
    {
      io_buffer_store (vc,
		       "-ERR Cannot change TTL: cannot get domain list: %m",
		       argv[1]);
      return;
    }
  
  while (--domainc >= 0)
    {
      char *zonename = bind_dir_name (vc, domainv[domainc]);
      struct zonefile *zone = zone_read (zonename);
      
      if (zone == NULL)
	io_buffer_store (vc,
			 "-ERR Cannot change TTL: Error reading zone %s",
			 domainv[domainc]);
      else
	{
	  if (zone->default_ttl != new_ttl)
	    {
	      char *dom = add_dot (domainv[domainc]);

	      vsd_log_error (VSD_LOG_INFO,
			     "ttl change vs %s domain %s ttl %d",
			     vc->virtual_server, domainv[domainc], new_ttl);

	      zone->default_ttl = new_ttl;
	      zone_update_serial (zone);
	      zone_save (zone, zonename, dom);
	      /* Notify the name server that a domain has been
		 modified.  */
	      bind_named_notify_update (vc, domainv[domainc]);
	      free (dom);
	    }
	  zone_free (zone);
	}
      free (domainv[domainc]);
      free (zonename);
    }
  free (domainv);
}

/* Tell the name server to reload all zone files belonging to a
   virtual server.  If an argument is given, then just reload that
   domain.  */
void bind_reload (struct connection *vc, int argc, char *argv[])
{
  /* BIND_RELOAD [<domain>] */
  char **domainv;
  int domainc;
  int new_ttl = atoi (argv[0]);

  /* Get a list of all domains.  */
  domainv = bind_dir_get_domains (vc, &domainc);
  if (domainv == NULL)
    {
      io_buffer_store (vc,
		       "-ERR Cannot reload: cannot get domain list: %m",
		       argv[1]);
      return;
    }
  
  while (--domainc >= 0)
    {
      vsd_log_error (VSD_LOG_INFO,
		     "domain reload vs %s domain %s",
		     vc->virtual_server, domainv[domainc]);
      bind_named_notify_update (vc, domainv[domainc]);
      free (domainv[domainc]);
    }
  free (domainv);
}
