/* Module command interface.
   This file is part of WebDNS.  
 */

/*
 * VSD - Virtual Server Daemon
 * Copyright (c) 2000 - 2001 Idaya Ltd.
 *
 * VSD is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * VSD is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with VSD - see the COPYING file;  if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA.
 */

/*
 * See the AUTHORS file for a list of contributors to VSD.
 * See the ChangeLog files for a list of changes.
 */

#include <stdio.h>
#include <errno.h>
#include <string.h>
#include <stdarg.h>
#include "vsd.h"
#include "module.h"
#include "mod_bind.h"
#include "bind.h"

struct command_tag clist[] =
{
  {
    CMD_BIND_LIST, ADMIN_VS,
    "List domains",
    "BIND_LIST", 0, "",
    "bind_list", 2, "<ns> <vs>"
  },
  {
    CMD_BIND_DELETE, ADMIN_VS,
    "Delete a domain",
    "BIND_DELETE", 1, "<domain name>",
    "bind_delete", 3, "<ns> <vs> <domain>"
  },
  {
    CMD_BIND_GET, ADMIN_VS,
    "Get the details for a domain",
    "BIND_GET", 1, "<domain name>",
    "bind_get", 3, "<ns> <vs> <domain"
  },
  {
    CMD_BIND_ADD, ADMIN_VS,
    "Add a new domain to the name server",
    "BIND_ADD", 2, "<domain-name> <ip> [<e-mail>]",
    "bind_add", 4, "<ns> <vs> <domain> <ip> [<e-mail>]"
  },
  {
    CMD_BIND_NAMEDGEN, ADMIN_VS,
    "Re-generate named.conf",
    "BIND_NAMEDGEN", 0, "",
    "bind_namedgen", 2, "<ns> <vs>"
  },
  {
    CMD_BIND_ROOTNAMEDGEN, ADMIN_HOST,
    "Re-generate main named.conf index",
    "BIND_ROOTNAMEDGEN", 0, "",
    "bind_rootnamedgen", 1, "<ns>"
  },
  {
    CMD_BIND_MODIFY, ADMIN_VS,
    "Modify an existing domain",
    "BIND_MODIFY", 1, "<domain-name>",
    "bind_modify", 4, "<ns> <vs> <domain> <filename>"
  },
  {
    CMD_BIND_IPMOD, ADMIN_VS,
    "Change an IP address for a domain",
    "BIND_IPMOD", 4, "<virtual-server> <domain-name> <old-ip> <new-ip>",
    "bind_ipmod", 5, "<ns> <vs> <domain> <old ip> <new ip>"
  },
  {
    CMD_BIND_IPMODALL, ADMIN_HOST,
    "Change an IP address for all domains held by a VS",
    "BIND_IPMODALL", 3, "<virtual-server> <old-ip> <new-ip>",
    "bind_ipmodall", 4, "<ns> <vs> <oldip> <newip>"
  },
  {
    CMD_BIND_FIX, ADMIN_HOST,
    "Fix common errors in DNS zone files",
    "BIND_FIX", 2, "<virtual-server> <domain>",
    "bind_fix", 3, "<ns> <vs> <domain>"
  },
  {
    CMD_BIND_FIXALL, ADMIN_HOST,
    "Fix common errors in DNS zone files held by a VS",
    "BIND_FIXALL", 1, "<virtual-server>",
    "bind_fixall", 2, "<ns> <vs>"
  },
  {
    CMD_BIND_DBGEN, ADMIN_HOST,
    "(re)generate domain database",
    "BIND_DBGEN", 0, "",
    "bind_dbgen", 1, "<ns>"
  },
  {
    CMD_BIND_DBLIST, ADMIN_HOST,
    "List domains stored in the exclusion database",
    "BIND_DBLIST", 0, "",
    "bind_dblist", 1, "<ns>" },
  {
    CMD_BIND_DBADD, ADMIN_HOST,
    "Add a domain to the exclusion database",
    "BIND_DBADD", 1, "<domain-name>",
    "bind_dbadd", 2, "<ns> <domain>"
  },
  {
    CMD_BIND_DBDELETE, ADMIN_HOST,
    "Delete a domain from the exclusion database",
    "BIND_DBDELETE", 1, "<domain-name>",
    "bind_dbdelete", 2, "<ns> <domain>"
  },
  {
    CMD_BIND_ADDVS, ADMIN_HOST,
    "Allow a vs to add domains to the name server",
    "BIND_ADDVS", 3, "<virtual-server> <ip> <max domains> [<pri ns>] [<sec ns>]",
    "bind_addvs", 4, "<ns> <vs> <ip> <max domains> [<pri ns>] [<sec ns>]"
  },
  {
    CMD_BIND_DELETEVS, ADMIN_HOST,
    "Delete domains belonging to a virtual server",
    "BIND_DELETEVS", 1, "<virtual-server>",
    "bind_deletevs", 2, "<ns> <vs>"
  },
  {
    CMD_BIND_RESTOREVS, ADMIN_HOST,
    "Restore a previously deleted virtual server",
    "BIND_RESTOREVS", 3, "<virtual-server> <ip> <max domains> [<pri ns>] [<sec ns>]",
    "bind_restorevs", 4, "<ns> <vs> <ip> <max domains> [<pri ns>] [<sec ns>]"
  },
  {
    CMD_BIND_MODIFYVS, ADMIN_HOST,
    "Modify a virtual server definition",
    "BIND_MODIFYVS", 1, "-v <vs> [-w <vs>] [-i <ip>] [-p <ns>] [-s <ns>] [-q <quota>]",
    "bind_modifyvs", 2, "<ns> -v <vs> [-w <vs>] [-i <ip>] [-p <ns>] [-s <ns>] [-q <quota>]\n"
    "<ns>       - name server to connect to\n"
    "-v <vs>    - operate on virtual server <vs>\n"
    "-w <vs>    - change virtual server name to <vs>\n"
    "-i <ip>    - Change virtual server IP address to <ip>\n"
    "-p <ns>    - set primary name server for virtual server to <ns>\n"
    "-s <ns>    - set secondary name server for virtual server to <vs>\n"
    "-q <quota> - set maximum number of domains for virtual server to <quota>\n"
  },
  {
    CMD_BIND_SETTTL, ADMIN_VS,
    "Set the TTL for all domains on a virtual server",
    "BIND_SETTTL", 1, "<ttl>",
    "bind_setttl", 3, "<ns> <vs> <new ttl>"
  },
  {
    CMD_BIND_RELOAD, ADMIN_VS,
    "Reload one or all domains for a virtual server",
    "BIND_RELOAD", 0, "[<domain>]",
    "bind_reload", 2, "<ns> <vs> [<domain>]"
  },
  {
    0, 0, NULL, NULL, 0, NULL, NULL, 0, NULL
  }
};

/* Return -1 if `cmd' does not exist.
   Return 1 if there is a syntax error.
   Return 0 if everything is cool.  */
int mod_bind_check_cmd (int level, const char *cmd, int argc, char *argv[],
			  struct command_tag **list)
{
  if (cmd == NULL)
    {
      /* Return a pointer to all commands.  */
      *list = clist;
      return 0;
    }

  /* Find command in clist.  */
  *list = module_lookup_command (clist, level, cmd);
  if (*list == NULL)
    return -1;  /* Command does not exist.  */

  /* Check command syntax.  Pretty basic at the moment.  */
#ifdef VSD_LIB
  if (argc < (*list)->vsdadm_args)
    return 1; /* Syntax error.  */
#else
  if (argc < (*list)->vsd_args)
    return 1; /* Syntax error.  */
#endif

  /* Syntax is ok.  */
  return 0;
}


#ifdef VSD_LIB
char *mod_bind_cmd (void *io, int level, const char *cmd,
		      int argc, char **argv)
{
  struct command_tag *list;
  char *status = NULL;

  list = module_lookup_command (clist, level, cmd);
  if (! list)
    return "notexist";

  vsd_set_level (io, list->level);
  switch (list->token)
    {
    case CMD_BIND_LIST:
      vsd_send_command (io, NULL, "BIND_LIST");
      status = vsd_process_reply (io);
      break;
    case CMD_BIND_DELETE:
      vsd_send_command (io, NULL, "BIND_DELETE %s", argv[2] /* domain */);
      status = vsd_process_reply (io);
      break;
    case CMD_BIND_GET:
      vsd_send_command (io, NULL, "BIND_GET %s", argv[2] /* domain */);
      status = vsd_process_reply (io);
      break;
    case CMD_BIND_ADD:
      if (argv[4])
	vsd_send_command (io, NULL, "BIND_ADD %s %s %s",
			  argv[2] /* domain */,
			  argv[3] /* ip */, argv[4] /* email */);
      else
	vsd_send_command (io, NULL, "BIND_ADD %s %s",
			  argv[2] /* domain */,
			  argv[3] /* ip */);
      status = vsd_process_reply (io);
      break;
    case CMD_BIND_MODIFY:
      {
	char *buffer = vsd_read_file (io, argv[3] /* filename */);

	if (buffer)
	  {
	    vsd_send_command (io, buffer, "BIND_MODIFY %s",
			      argv[2] /* domain */);
	    free (buffer);
	    status = vsd_process_reply (io);
	  }
	else
	  status = strerror (errno);
      }
      break;
    case CMD_BIND_IPMOD:
      vsd_send_command (io, NULL, "BIND_IPMOD %s %s %s %s",
			argv[1] /* vs */, argv[2] /* domain */,
			argv[3] /* old ip */, argv[4] /* new ip */);
      status = vsd_process_reply (io);
      break;
    case CMD_BIND_IPMODALL:
      vsd_send_command (io, NULL, "BIND_IPMODALL %s %s %s",
			argv[1] /* vs */,
			argv[2] /* old ip */, argv[3] /* new ip */);
      status = vsd_process_reply (io);
      break;
    case CMD_BIND_FIX:
      vsd_send_command (io, NULL, "BIND_FIX %s %s",
			argv[1] /* vs */, argv[2] /* domain */);
      status = vsd_process_reply (io);
      break;
    case CMD_BIND_FIXALL:
      vsd_send_command (io, NULL, "BIND_FIXALL %s", argv[1] /* vs */);
      status = vsd_process_reply (io);
      break;
    case CMD_BIND_DBGEN:
      vsd_send_command (io, NULL, "BIND_DBGEN");
      status = vsd_process_reply (io);
      break;
    case CMD_BIND_DBLIST:
      vsd_send_command (io, NULL, "BIND_DBLIST");
      status = vsd_process_reply (io);
      break;
    case CMD_BIND_DBADD:
      vsd_send_command (io, NULL, "BIND_DBADD %s", argv[1] /* domain */);
      status = vsd_process_reply (io);
      break;
    case CMD_BIND_DBDELETE:
      vsd_send_command (io, NULL, "BIND_DBDELETE %s", argv[1] /* dom */);
      status = vsd_process_reply (io);
      break;
    case CMD_BIND_NAMEDGEN:
      vsd_send_command (io, NULL, "BIND_NAMEDGEN");
      status = vsd_process_reply (io);
      break;
    case CMD_BIND_ROOTNAMEDGEN:
      vsd_send_command (io, NULL, "BIND_ROOTNAMEDGEN");
      status = vsd_process_reply (io);
      break;
    case CMD_BIND_ADDVS:
      vsd_send_command (io, NULL, "BIND_ADDVS %s %s %s",
			argv[1] /* vs */, argv[2] /* ip */,
			argv[3] /* max domains */);
      status = vsd_process_reply (io);
      break;
    case CMD_BIND_DELETEVS:
      vsd_send_command (io, NULL, "BIND_DELETEVS %s", argv[1] /* vs */);
      status = vsd_process_reply (io);
      break;
    case CMD_BIND_RESTOREVS:
      vsd_send_command (io, NULL, "BIND_RESTOREVS %s %s %s",
			argv[1] /* vs */, argv[2] /* ip */,
			argv[3] /* max domains */);
      status = vsd_process_reply (io);
      break;
    case CMD_BIND_MODIFYVS:
      vsd_send_vcommand (io, NULL, "BIND_MODIFYVS", argc - 1, argv);
      status = vsd_process_reply (io);
      break;
    case CMD_BIND_SETTTL:
      vsd_send_command (io, NULL, "BIND_SETTTL %s", argv[2] /* TTL */);
      status = vsd_process_reply (io);
      break;
    case CMD_BIND_RELOAD:
      if (argv[2])
	vsd_send_command (io, NULL, "BIND_RELOAD %s",
			  argv[2] /* domain */); 
      else
	vsd_send_command (io, NULL, "BIND_RELOAD");
      status = vsd_process_reply (io);
      break;
    default:
      status = "mod_bind: command not recognised";
      break;
    }

  return status;
}

#else /* ! VSD_LIB */

/* Return 0 if command processed, 1 if not recognised.  */
int mod_bind_cmd (struct connection *vc, int level,
		    const char *command, int argc, char *argv[],
		    const char *data)
{
  int token;
  struct command_tag *tag;

  tag = module_lookup_command (clist, level, command);
  if (tag == NULL)
    return 1;

  token = tag->token;
  switch (token)
    {
    case CMD_BIND_LIST:
      /* List domains that the client has previously configured.  */
      list_domains (vc, argc, argv);
      break;
    case CMD_BIND_DELETE:
      /* Delete a domain.  */
      domain_delete (vc, argc, argv);
      break;
    case CMD_BIND_GET:
      /* Get a domain.  */
      domain_get (vc, argc, argv);
      break;
    case CMD_BIND_ADD:
      /* Add a domain.  */
      domain_add (vc, argc, argv);
      break;
    case CMD_BIND_MODIFY:
      /* Modify a domain.  */
      domain_modify (vc, argc, argv, data);
      break;
    case CMD_BIND_IPMOD:
      /* Change the IP address for a zone.  */
      domain_ipmod (vc, argc, argv);
      break;
    case CMD_BIND_IPMODALL:
      /* Change the IP address for all zones.  */
      domain_ipmodall (vc, argc, argv);
      break;
    case CMD_BIND_FIX:
      /* Fix common errors in DNS zone files.  */
      domain_fix (vc, argc, argv);
      break;
    case CMD_BIND_FIXALL:
      /* Fix common errors in DNS zone files held by a VS.  */
      domain_fixall (vc, argc, argv);
      break;
    case CMD_BIND_DBGEN:
      /* Generate/re-generate bind domain database.  */
      bind_dbase_generate (vc);
      break;
    case CMD_BIND_DBLIST:
      /* List domains in the exclusion database.  */
      bind_dbase_excl_list (vc, argc, argv);
      break;
    case CMD_BIND_DBADD:
      /* Add domains to the exclusion database.  */
      bind_dbase_excl_add (vc, argc, argv);
      break;
    case CMD_BIND_DBDELETE:
      /* Delete domains from the exclusion database.  */
      bind_dbase_excl_delete (vc, argc, argv);
      break;
    case CMD_BIND_NAMEDGEN:
      /* Re-generate named.conf.  */
      bind_named_gen (vc, argc, argv);
      break;
    case CMD_BIND_ROOTNAMEDGEN:
      /* Re-generate named.conf.  */
      bind_root_named_gen (vc, argc, argv);
      break;
    case CMD_BIND_ADDVS:
      /* Add a virtual server.  */
      domain_addvs (vc, argc, argv);
      break;
    case CMD_BIND_DELETEVS:
      /* Delete a virtual server.  */
      domain_deletevs (vc, argc, argv);
      break;
    case CMD_BIND_RESTOREVS:
      /* Restore a previously deleted virtual server.  */
      domain_restorevs (vc, argc, argv);
      break;
    case CMD_BIND_MODIFYVS:
      /* Modify a virtual server definition.  */
      domain_modifyvs (vc, argc, argv);
      break;
    case CMD_BIND_SETTTL:
      /* Rename a virtual server.  */
      bind_set_ttl (vc, argc, argv);
      break;
    case CMD_BIND_RELOAD:
      /* Reload all domains for a virtual server.  */
      bind_reload (vc, argc, argv);
      break;
      break;
    default:
      return 1;
      break;
    }

  return 0;
}
#endif /* VSD_LIB */

module bind_module =
{
  mod_bind_check_cmd,
  mod_bind_cmd
};
