/* ca.c - ca module for openssl ca manipulation.
 */

/* 
 * VSD - Virtual Server Daemon
 * Copyright (c) 2000 - 2001 Idaya Ltd.
 *
 * VSD is free software; you can redistribute it and/or 
 * modify it under the terms of the GNU General Public License 
 * as published by the Free Software Foundation; either 
 * version 2 of the License, or (at your option) any later version.
 *
 * VSD is distributed in the hope that it will be useful, 
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU 
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License 
 * along with VSD - see the COPYING file;  if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330, 
 * Boston, MA 02111-1307, USA.  
 */

/*
 * See the AUTHORS file for a list of contributors to VSD. 
 * See the ChangeLog files for a list of changes. 
 */

#include "config.h"

#ifdef WITH_OPENSSL
#include <openssl/conf.h>
#include <openssl/ssl.h>
#endif

#include "vsd.h"
#include "libvsd.h"
#include "ca.h"

#define VSD_OK  0
#define VSD_ERR -1

/********************************************************************************
 * Description: ca server root path.
 *
 *     Returns: char * pointer to server root path.
 *              NULL on error.
 *******************************************************************************/
static char * 
ca_svr_root (char *root)
{ 
  static char path[256];
  
  sprintf (path, "%s", (root ? root : "/"));
  
  return path;
}

/********************************************************************************
 * Description: VSD CA version information
 *
 *     Returns: void
 *******************************************************************************/
void 
ca_version (struct connection *vc, 
	    const int         argc, 
	    const char        *argv[])
{
  char *version;
  int rc = VSD_OK;

#ifdef WITH_OPENSSL
  LHASH *conf = NULL;
  SSL_METHOD *mth = NULL;
  SSL_CTX *ctx = NULL;
  SSL *ssl = NULL;
#endif

  if (argc != 0)
    {
      rc = VSD_ERR;
      io_buffer_store (vc, "-ERR [ca] Incorrect arguments.\n");
    }

  if (!rc)
    {
#ifndef WITH_OPENSSL
      io_buffer_store (vc, "-ERR [ca] OpenSSL not supported.\n");
#else
      io_buffer_store (vc, "%s\n", SSLeay_version (SSLEAY_VERSION));
      
      if ((mth = SSLv2_method())
	  && (ctx = SSL_CTX_new (mth))
	  && (ssl = SSL_new (ctx)))
	{
	  io_buffer_store (vc, "%s\n", SSL_get_version (ssl));
	}
      else
	{
	  rc = VSD_ERR;
	  io_buffer_store (vc, "-ERR [ca] Unable to obtain SSL version.\n");
	}
#endif
    }
  
#ifdef WITH_OPENSSL
  SSL_free (ssl);
  SSL_CTX_free (ctx);
  CONF_free (conf);
#endif
}

#ifdef WITH_OPENSSL
/********************************************************************************
 * Description: Verify VSD CA database
 *
 *     Returns: void
 *******************************************************************************/
void 
ca_verify (struct connection *vc, 
	   const int         argc, 
	   const char        *argv[])
{
  char *errors;
  LHASH *conf = NULL;
  struct vsd_file_line *fdb, *line;
  int rc = VSD_OK;
  
  if (argc != 0)
    {
      rc = VSD_ERR;
      io_buffer_store (vc, "-ERR [ca] Incorrect arguments.\n");
    }

  if(!rc)
    {
      if ((errors = vsd_openssl_ca_verify_database (&conf)) == NULL)
	io_buffer_store (vc, "OK\n");
      else
	io_buffer_store (vc, errors);
    }

  CONF_free (conf);
}

/********************************************************************************
 * Description: List VSD certificate/key pairs in VSD CA
 *
 *     Returns: void
 *******************************************************************************/
void 
ca_list (struct connection *vc, 
	 const int         argc, 
	 const char        *argv[])
{
  char *buf = NULL;
  LHASH *conf = NULL;
  int rc = VSD_OK;

  if (argc != 2)
    {
      rc = VSD_ERR;
      io_buffer_store (vc, "-ERR [ca] Incorrect arguments.\n");
    }
  
  if (!rc)
    {
      buf = vsd_openssl_ca_db_records (&conf, atoi (argv[0]), atoi (argv[1]));
      io_buffer_store (vc, "%s", buf);
    }

  free (buf);

  CONF_free (conf);
}

/********************************************************************************
 * Description: Add VSD certificate/key pair to VSD CA
 *
 *     Returns: void
 *******************************************************************************/
void 
ca_add (struct connection *vc, 
	const int         argc, 
	const char        *argv[])
{
  char *cmd, tmp[128];
  int x;
  int rc = VSD_OK;

  if (argc != 1)
    {
      rc = VSD_ERR;
      io_buffer_store (vc, "-ERR [ca] Incorrect arguments.\n");
    }

  if (!rc)
    {
      memset (tmp, 0, sizeof(tmp));
#ifdef DEBUG
      srand (time (NULL));
      x = rand () % 1383252;
      sprintf (tmp, "/tmp/vsd%06u", x);
#else
      strcpy (tmp, "/dev/null");
#endif

      if (access ("/usr/sbin/vsd-ca_add", R_OK | X_OK) == 0)
	cmd = "/usr/sbin/vsd-ca_add";
      else if (access ("/usr/local/sbin/vsd-ca_add", R_OK | X_OK) == 0)
	cmd = "/usr/local/sbin/vsd-ca_add";
      else
	{
	  rc = VSD_ERR;
	  io_buffer_store (vc, "-ERR [ca] Cannot find vsd-ca_add.\n");
	}

      if (!rc && (rc = run_command (vc, vc->server_root, "root", cmd, argv[0], NULL, tmp)))
	io_buffer_store (vc, "-ERR [ca] vsd-ca_add failed: %m\n");
    }
}

/********************************************************************************
 * Description: Delete VSD certificate/key pair from VSD CA
 *
 *     Returns: void
 *******************************************************************************/
void 
ca_del (struct connection *vc, 
	const int         argc, 
	const char        *argv[])
{
  char *cmd, tmp[128];
  int x;
  int rc = VSD_OK;
  
  if (argc != 1)
  {
    rc = VSD_ERR;
    io_buffer_store (vc, "-ERR [ca] Incorrect arguments.\n");
  }

  if (!rc)
    {
      memset (tmp, 0, sizeof(tmp));
#ifdef DEBUG
      srand (time (NULL));
      x = rand () % 1383252;
      sprintf (tmp, "/tmp/vsd%06u", x);
#else
      strcpy (tmp, "/dev/null");
#endif
      
      if (access ("/usr/sbin/vsd-ca_del", R_OK | X_OK) == 0)
	cmd = "/usr/sbin/vsd-ca_del";
      else if (access ("/usr/local/sbin/vsd-ca_del", R_OK | X_OK) == 0)
	cmd = "/usr/local/sbin/vsd-ca_del";
      else
	{
	  rc = VSD_ERR;
	  io_buffer_store (vc, "-ERR [ca] Cannot find vsd-ca_del.\n");
	}
    }

  if (!rc && (rc = run_command (vc, vc->server_root, "root", cmd, argv[0], NULL, tmp)))
    io_buffer_store (vc, "-ERR [ca] vsd-ca_del failed: %m\n");
}

/********************************************************************************
 * Description: Get VSD certificate from VSD CA
 *
 *     Returns: void
 *******************************************************************************/
void 
ca_certget (struct connection *vc, 
	    const int         argc, 
	    const char        *argv[])
{
  char *crt;
  LHASH *conf = NULL;
  struct vsd_file_line *fcrt, *line;
  int rc = VSD_OK;

  if (argc != 1)
  {
    rc = VSD_ERR;
    io_buffer_store (vc, "-ERR [ca] Incorrect arguments.\n");
  }

  if (!rc && (crt = (char *) vsd_openssl_ca_crt (&conf, argv[0])) == NULL)
    {
      rc = VSD_ERR;
      io_buffer_store (vc, "-ERR [ca] Certificate not found for %s.\n", argv[0]);
    }
  
  if (!rc && (rc = vsd_load_file (&fcrt, ca_svr_root (vc->server_root), crt)))
    io_buffer_store (vc, "-ERR [ca] Loading %s certificate: %m\n", argv[0]);
  
  for (line = fcrt; !rc && line; line=line->next)
    io_buffer_store (vc, "%s\n", line->line);
  
  vsd_free_file (fcrt);
  CONF_free (conf);
}

/********************************************************************************
 * Description: Put VSD certificate onto VSD server
 *
 *     Returns: void
 *******************************************************************************/
void 
ca_certput (struct connection *vc, 
	    const int         argc, 
	    const char        *argv[],
	    const char        *data)
{
  char *crt;
  LHASH *conf = NULL;
  FILE *pfile;
  char *cmd, args[256], tmp[128];
  int x;
  int rc = VSD_OK;

  if (argc > 2 || argc < 1)
  {
    rc = VSD_ERR;
    io_buffer_store (vc, "-ERR [ca] Incorrect arguments.\n");
  }

  if (!rc && (crt = (char *) vsd_openssl_ca_crt (&conf, argv[0])) == NULL)
    {
      rc = VSD_ERR;
      io_buffer_store (vc, "-ERR [ca] Certificate not found for %s.\n", argv[0]);
    }
  
  if (!rc && (argc == 1))
    {     
      if ((pfile = fopen (crt, "w")) == NULL)
      {
	rc = VSD_ERR;
	io_buffer_store (vc, "-ERR [ca] Cannot write %s certificate.\n", argv[0]);
      }
      else
	{
	  fprintf (pfile, data);
	  fclose (pfile);
	}
    }

  if (!rc && (argc == 2))
    {
      memset (tmp, 0, sizeof(tmp));
      
#ifdef DEBUG
      srand (time (NULL));
      x = rand () % 1383252;
      sprintf (tmp, "/tmp/vsd%06u", x);
#else
      strcpy (tmp, "/dev/null");
#endif
      
      if (access ("/usr/sbin/vsdadm", R_OK | X_OK) == 0)
	cmd = "/usr/sbin/vsdadm";
      else if (access ("/usr/local/sbin/vsdadm", R_OK | X_OK) == 0)
	cmd = "/usr/local/sbin/vsdadm";
      else
	{
	  rc = VSD_ERR;
	  io_buffer_store (vc, "-ERR [ca] Cannot find vsdadm.\n");
	}
      
      sprintf (args, "ca_certput %s %s %s %s", argv[1], argv[1], argv[0], crt);
      if (!rc && (rc = run_command (vc, vc->server_root, "root", cmd, args, NULL, tmp)))
	io_buffer_store (vc, "-ERR [ca] vsdadm failed: %m\n");
    }

  CONF_free (conf);
}

/********************************************************************************
 * Description: Get VSD key from VSD CA
 *
 *     Returns: void
 *******************************************************************************/
void 
ca_keyget (struct connection *vc, 
	   const int         argc, 
	   const char        *argv[])
{
  char *key;
  LHASH *conf = NULL;
  struct vsd_file_line *fkey, *line;
  int rc = VSD_OK;
  
  if (argc != 1)
    {
      rc = VSD_ERR;
      io_buffer_store (vc, "-ERR [ca] Incorrect arguments.\n");
    }
  
  if (!rc && (key = (char *) vsd_openssl_ca_key (&conf, argv[0])) == NULL)
    {
      rc = VSD_ERR;
      io_buffer_store (vc, "-ERR [ca] Key not found for %s.\n", argv[0]);
    }
  
  if (!rc && (rc = vsd_load_file (&fkey, ca_svr_root (vc->server_root), key)))
      io_buffer_store (vc, "-ERR [ca] Loading %s key: %m\n", argv[0]);
  
  for (line = fkey; !rc && line; line=line->next)
    io_buffer_store (vc, "%s\n", line->line);

  vsd_free_file (fkey);
  CONF_free (conf);
}

/********************************************************************************
 * Description: Put VSD key onto VSD server
 *
 *     Returns: void
 *******************************************************************************/
void 
ca_keyput (struct connection *vc, 
	   const int         argc, 
	   const char        *argv[],
	   const char        *data)
{
  char *key;
  LHASH *conf = NULL;
  FILE *pfile;
  char *cmd, args[256], tmp[128];
  int x;
  int rc = VSD_OK;

  if (argc > 2 || argc < 1)
  {
    rc = VSD_ERR;
    io_buffer_store (vc, "-ERR [ca] Incorrect arguments.\n");
  }

  if (!rc && (key = (char *) vsd_openssl_ca_key (&conf, argv[0])) == NULL)
    {
      rc = VSD_ERR;
      io_buffer_store (vc, "-ERR [ca] Key not found for %s.\n", argv[0]);
    }
  
  if (!rc && (argc == 1))
    {     
      if ((pfile = fopen (key, "w")) == NULL)
	{
	  rc = VSD_ERR;
	  io_buffer_store (vc, "-ERR [ca] Cannot write %s key.\n", argv[0]);
	}
      else
	{
	  fprintf (pfile, data);
	  fclose (pfile);
	}
    }

  if (!rc && (argc == 2))
    {
      memset (tmp, 0, sizeof(tmp));
      
#ifdef DEBUG
      srand (time (NULL));
      x = rand () % 1383252;
      sprintf (tmp, "/tmp/vsd%06u", x);
#else
      strcpy (tmp, "/dev/null");
#endif
      
      if (access ("/usr/sbin/vsdadm", R_OK | X_OK) == 0)
	cmd = "/usr/sbin/vsdadm";
      else if (access ("/usr/local/sbin/vsdadm", R_OK | X_OK) == 0)
	cmd = "/usr/local/sbin/vsdadm";
      else
	{
	  rc = VSD_ERR;
	  io_buffer_store (vc, "-ERR [ca] Cannot find vsdadm.\n");
	}
      
      sprintf (args, "ca_keyput %s %s %s %s", argv[1], argv[1], argv[0], key);
      if (!rc && (rc = run_command (vc, vc->server_root, "root", cmd, args, NULL, tmp)))
	io_buffer_store (vc, "-ERR [ca] vsdadm failed: %m\n");
    }

  CONF_free (conf);
}

/********************************************************************************
 * Description: Get VSD crl from VSD server
 *
 *     Returns: void
 *******************************************************************************/
void 
ca_crlget (struct connection *vc, 
	   const int         argc, 
	   const char        *argv[])
{
  char *crl;
  LHASH *conf = NULL;
  struct vsd_file_line *fcrl, *line;
  int rc = VSD_OK;

  if (argc != 0)
    {
      rc = VSD_ERR;
      io_buffer_store (vc, "-ERR [ca] Incorrect arguments.\n");
    }

  if (!rc && (crl = (char *) vsd_openssl_ca_lookup (&conf, "crl")) == NULL)
    {
      rc = VSD_ERR;
      io_buffer_store (vc, "-ERR [ca] Certificate revocation list not found.\n");
    }

  if (!rc && (rc = vsd_load_file (&fcrl, ca_svr_root (vc->server_root), crl)))
      io_buffer_store (vc, "-ERR [ca] Loading VSD CA certificate revocation list: %m\n");
  
  for (line = fcrl; !rc && line; line=line->next)
    io_buffer_store (vc, "%s\n", line->line);

  vsd_free_file (fcrl);
  CONF_free (conf);
}

/********************************************************************************
 * Description: Put VSD crl onto VSD server
 *
 *     Returns: void
 *******************************************************************************/
void 
ca_crlput (struct connection *vc, 
           const int         argc, 
           const char        *argv[],
           const char        *data)
{
  char *crl;
  LHASH *conf = NULL;
  FILE *pfile;
  char *cmd, args[256], tmp[128];
  int x;
  int rc = VSD_OK;

  if (argc > 1)
    {
      rc = VSD_ERR;
      io_buffer_store (vc, "-ERR [ca] Incorrect arguments.\n");
    }

  if (!rc && (crl = (char *) vsd_openssl_ca_lookup (&conf, "crl")) == NULL)
    {
      rc = VSD_ERR;
      io_buffer_store (vc, "-ERR [ca] Certificate revocation list not found.\n");
    }
  
  if (!rc && (argc == 0))
    {     
      if ((pfile = fopen (crl, "w")) == NULL)
	{
	  rc = VSD_ERR;
	  io_buffer_store (vc, "-ERR [ca] Cannot write VSD CA"
			   "certificate revocation list.\n");
	}
      else
	{
	  fprintf (pfile, data);
	  fclose (pfile);
	}
    }

  if (!rc && (argc == 1))
    {
      memset (tmp, 0, sizeof(tmp));
      
#ifdef DEBUG
      srand (time (NULL));
      x = rand () % 1383252;
      sprintf (tmp, "/tmp/vsd%06u", x);
#else
      strcpy (tmp, "/dev/null");
#endif
      
      if (access ("/usr/sbin/vsdadm", R_OK | X_OK) == 0)
	cmd = "/usr/sbin/vsdadm";
      else if (access ("/usr/local/sbin/vsdadm", R_OK | X_OK) == 0)
	cmd = "/usr/local/sbin/vsdadm";
      else
	{
	  rc = VSD_ERR;
	  io_buffer_store (vc, "-ERR [ca] Cannot find vsdadm.\n");
	}
      
      sprintf (args, "ca_crlput %s %s %s", argv[0], argv[0], crl);
      if (!rc && (rc = run_command (vc, vc->server_root, "root", cmd, args, NULL, tmp)))
	io_buffer_store (vc, "-ERR [ca] vsdadm failed: %m\n");
    }

  CONF_free (conf);
}

/********************************************************************************
 * Description: Get VSD CA root certificate
 *
 *     Returns: void
 *******************************************************************************/
void 
ca_cacertget (struct connection *vc, 
	      const int         argc, 
	      const char        *argv[])
{
  char *certificate;
  LHASH *conf= NULL;
  struct vsd_file_line *fcrt, *line;
  int rc = VSD_OK;

  if (argc != 0)
    {
      rc = VSD_ERR;
      io_buffer_store (vc, "-ERR [ca] Incorrect arguments.\n");
    }

  if (!rc && (certificate = (char *) vsd_openssl_ca_lookup (&conf, "certificate")) == NULL)
    {
      rc = VSD_ERR;
      io_buffer_store (vc, "-ERR [ca] VSD CA root certificate not found.\n");
    }
  
  if (!rc && (rc =vsd_load_file (&fcrt, ca_svr_root (vc->server_root), certificate)))
      io_buffer_store (vc, "-ERR [ca] Loading VSD CA root certificate: %m\n");

  for (line = fcrt; !rc && line; line=line->next)
    io_buffer_store (vc, "%s\n", line->line);
  
  vsd_free_file (fcrt);
  CONF_free (conf);
}

/********************************************************************************
 * Description: Put VSD CA root certificate onto VSD server
 *
 *     Returns: void
 *******************************************************************************/
void 
ca_cacertput (struct connection *vc, 
	      const int         argc, 
	      const char        *argv[],
	      const char        *data)
{
  char *certificate;
  LHASH *conf = NULL;
  FILE *pfile;
  char *cmd, args[256], tmp[128];
  int x;
  int rc = VSD_OK;

  if (argc > 1)
    {
      rc = VSD_ERR;
      io_buffer_store (vc, "-ERR [ca] Incorrect arguments.\n");
    }

  if (!rc && (certificate = (char *) vsd_openssl_ca_lookup (&conf, "certificate")) == NULL)
    {
      rc = VSD_ERR;
      io_buffer_store (vc, "-ERR [ca] VSD CA root certificate not found.\n");
    }
  
  if (!rc && (argc == 0))
    {
      if ((pfile = fopen (certificate, "w")) == NULL)
	{
	  rc = VSD_ERR;
	  io_buffer_store (vc, "-ERR [ca] Cannot write VSD CA root certificate.\n");
	}
      else
	{
	  fprintf (pfile, data);
	  fclose (pfile);
	}
    }

  if (!rc && (argc == 1))
    {
      memset (tmp, 0, sizeof(tmp));
      
#ifdef DEBUG
      srand (time (NULL));
      x = rand () % 1383252;
      sprintf (tmp, "/tmp/vsd%06u", x);
#else
      strcpy (tmp, "/dev/null");
#endif
      
      if (access ("/usr/sbin/vsdadm", R_OK | X_OK) == 0)
	cmd = "/usr/sbin/vsdadm";
      else if (access ("/usr/local/sbin/vsdadm", R_OK | X_OK) == 0)
	cmd = "/usr/local/sbin/vsdadm";
      else
	{
	  rc = VSD_ERR;
	  io_buffer_store (vc, "-ERR [ca] Cannot find vsdadm.\n");
	}
      
      sprintf (args, "ca_cacertput %s %s %s", argv[0], argv[0], certificate);
      if (!rc && (rc = run_command (vc, vc->server_root, "root", cmd, args, NULL, tmp)))
	io_buffer_store (vc, "-ERR [ca] vsdadm failed: %m\n");
    }
  
  CONF_free (conf);
}
#endif  /* WITH_OPENSSL */







