/* mod_ca.c - ca module interface.
 */

/* 
 * VSD - Virtual Server Daemon
 * Copyright (c) 2000 - 2001 Idaya Ltd.
 *
 * VSD is free software; you can redistribute it and/or 
 * modify it under the terms of the GNU General Public License 
 * as published by the Free Software Foundation; either 
 * version 2 of the License, or (at your option) any later version.
 *
 * VSD is distributed in the hope that it will be useful, 
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU 
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License 
 * along with VSD - see the COPYING file;  if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330, 
 * Boston, MA 02111-1307, USA.  
 */

/*
 * See the AUTHORS file for a list of contributors to VSD. 
 * See the ChangeLog files for a list of changes. 
 */

#include <stdio.h>

#ifdef WITH_OPENSSL
#include <openssl/conf.h>
#endif

#include "module.h"
#include "config.h"
#include "mod_ca.h"
#include "ca.h"

static struct command_tag clist[] =
{
  {
    CMD_CA_VERSION, ADMIN_HOST,
    "Version information for VSD CA",
    "CA_VERSION", 0, "",
    "ca_version", 1, "<vsdcasvr>"
  },
#ifdef WITH_OPENSSL
  {
    CMD_CA_VERIFY, ADMIN_HOST,
    "Verify VSD CA database",
    "CA_VERIFY", 0, "",
    "ca_verify", 1, "<vsdcasvr>"
  },
  {
    CMD_CA_LIST, ADMIN_HOST,
    "List VSD certificate/key pairs in VSD CA",
    "CA_LIST", 2, "<offset> <records>",
    "ca_list", 3, "<vsdcasvr> <offset> <records>"
  },
  {
    CMD_CA_ADD, ADMIN_HOST,
    "Add VSD certificate/key pair to VSD CA",
    "CA_ADD", 1, "<cn>",
    "ca_add", 2, "<vsdcasvr> <cn>"
  },
  {
    CMD_CA_DEL, ADMIN_HOST,
    "Delete VSD certificate/key pair from VSD CA",
    "CA_DEL", 1, "<cn>",
    "ca_del", 2, "<vsdcasvr> <cn>"
  },
  {
    CMD_CA_CERTGET, ADMIN_HOST,
    "Get VSD certificate from VSD CA",
    "CA_CERTGET", 1, "<cn>",
    "ca_certget", 2, "<vsdcasvr> <cn>"
  },
  {
    CMD_CA_CERTPUT, ADMIN_HOST,
    "Put VSD certificate onto VSD server",
    "CA_CERTPUT", 1, "<cn> <vsdsvr>",
    "ca_certput", 2, "<vsdcasvr> <cn> <vsdsvr>"
  },
  {
    CMD_CA_KEYGET, ADMIN_HOST,
    "Get VSD key from VSD CA",
    "CA_KEYGET", 0, "",
    "ca_keyget", 1, "<vsdcasvr>"
  },
  {
    CMD_CA_KEYPUT, ADMIN_HOST,
    "Put VSD key onto VSD server",
    "CA_KEYPUT", 1, "<cn> <vsdsvr>",
    "ca_keyput", 2, "<vsdcasvr> <cn> <vsdsvr>"
  },
  {
    CMD_CA_CRLGET, ADMIN_HOST,
    "Get VSD crl from VSD CA",
    "CA_CRLGET", 0, "",
    "ca_crlget", 1, "<vsdsvr>"
  },
  {
    CMD_CA_CRLPUT, ADMIN_HOST,
    "Put VSD crl onto VSD server",
    "CA_CRLPUT", 0, "<vsdsvr>",
    "ca_crlput", 1, "<vsdcasvr> <vsdsvr>"
  },
  {
    CMD_CA_CACERTGET, ADMIN_HOST,
    "Get VSD CA root certificate from VSD CA",
    "CA_CACERTGET", 0, "",
    "ca_cacertget", 1, "<vsdcasvr>"
  },
  {
    CMD_CA_CACERTPUT, ADMIN_HOST,
    "Put VSD CA root certificate onto VSD server",
    "CA_CACERTPUT", 0, "<vsdsvr>",
    "ca_cacertput", 1, "<vsdcasvr> <vsdvsr>"
  },
  {
    CMD_CA_CERTADD, ADMIN_HOST,      /* Legacy command */
    "Add VSD certificate/key pair to VSD CA",
    "CA_CERTADD", 1, "<cn>",
    "ca_certadd", 2, "<vsdcasvr> <cn>"
  },
  {
    CMD_CA_CERTDEL, ADMIN_HOST,      /* Legacy command */
    "Delete VSD certificate/key pair from VSD CA",
    "CA_CERTDEL", 1, "<cn>",
    "ca_certdel", 2, "<vsdcasvr> <cn>"
  },
#endif  /* WITH_OPENSSL */
  {
    0, 0, NULL, NULL, 0, NULL, NULL, 0, NULL
  }
};

/* Return -1 if `cmd' does not exist.
   Return 1 if there is a syntax error.
   Return 0 if everything is cool.  */
int mod_ca_check_cmd (int level, const char *cmd, int argc, char *argv[],
		      struct command_tag **list)
{
  if (cmd == NULL)
    {
      /* Return a pointer to all commands.  */
      *list = clist;
      return 0;
    }

  /* Find command in clist.  */
  *list = module_lookup_command (clist, level, cmd);
  if (*list == NULL)
    return -1;  /* Command does not exist.  */

  /* Check command syntax.  Pretty basic at the moment.  */
#ifdef VSD_LIB
  if (argc < (*list)->vsdadm_args)
    return 1; /* Syntax error.  */
#else
  if (argc < (*list)->vsd_args)
    return 1; /* Syntax error.  */
#endif

  /* Syntax is ok.  */
  return 0;
}

#ifdef VSD_LIB
char *mod_ca_cmd (void *io, int level, const char *cmd,
		  const int argc, const char **argv)
{
  struct command_tag *list;
  char *status = NULL;

  list = module_lookup_command (clist, level, cmd);
  if (! list)
    return "notexist";

  vsd_set_level (io, list->level);
  switch (list->token)
    {
    case CMD_CA_VERSION:
      vsd_send_command (io, NULL, 
			"CA_VERSION"); 
      status = vsd_process_reply (io);
      break;
    case CMD_CA_VERIFY:
      vsd_send_command (io, NULL, 
			"CA_VERIFY"); 
      status = vsd_process_reply (io);
      break;
    case CMD_CA_LIST:
      vsd_send_command (io, NULL, 
			"CA_LIST %s %s",
			argv[1],   /* offset */
			argv[2]);  /* records */ 
      status = vsd_process_reply (io);
      break;
    case CMD_CA_ADD:
      vsd_send_command (io, NULL, 
			"CA_ADD %s",
			argv[1]  /* cn */); 
      status = vsd_process_reply (io);
      break;
    case CMD_CA_DEL:
      vsd_send_command (io, NULL, 
			"CA_DEL %s",
			argv[1]  /* cn */); 
      status = vsd_process_reply (io);
      break;
    case CMD_CA_CERTGET:
      vsd_send_command (io, NULL, 
			"CA_CERTGET %s",
			argv[1]  /* cn */); 
      status = vsd_process_reply (io);
      break;
    case CMD_CA_CERTPUT:
      if (argc > 4 && argv[3])
	vsd_send_command_file (io, argv[3], 
			       "CA_CERTPUT %s",
			       argv[1]  /* cn */);
      else
	vsd_send_command (io, NULL, 
			  "CA_CERTPUT %s %s",
			  argv[1],  /* cn */
			  argv[2]   /* vsdsvr */); 
      status = vsd_process_reply (io);
      break;
    case CMD_CA_KEYGET:
      vsd_send_command (io, NULL, 
			"CA_KEYGET %s",
			argv[1]  /* cn */); 
      status = vsd_process_reply (io);
      break;
    case CMD_CA_KEYPUT:
      if (argc > 4 && argv[3])
	vsd_send_command_file (io, argv[3], 
			       "CA_KEYPUT %s",
			       argv[1]  /* cn */); 
      else
	vsd_send_command (io, NULL, 
			  "CA_KEYPUT %s %s",
			  argv[1],  /* cn */
			  argv[2]   /* vsdsvr */); 
      status = vsd_process_reply (io);
      break;
    case CMD_CA_CRLGET:
      vsd_send_command (io, NULL, 
			"CA_CRLGET"); 
      status = vsd_process_reply (io);
      break;
    case CMD_CA_CRLPUT:
      if (argc > 3 && argv[2])
	vsd_send_command_file (io, argv[2], 
			       "CA_CRLPUT");
      else
	vsd_send_command (io, NULL, 
			  "CA_CRLPUT %s",
			  argv[1]  /* vsdsvr */);
      status = vsd_process_reply (io);
      break;
    case CMD_CA_CACERTGET:
      vsd_send_command (io, NULL, 
			"CA_CACERTGET"); 
      status = vsd_process_reply (io);
      break;
    case CMD_CA_CACERTPUT:
      if (argc > 3 && argv[2])
	vsd_send_command_file (io, argv[2], 
			       "CA_CACERTPUT"); 
      else
	vsd_send_command (io, NULL, 
			  "CA_CACERTPUT %s",
			  argv[1]  /* vsdsvr */); 
      status = vsd_process_reply (io);
      break;
    case CMD_CA_CERTADD: /* Deprecated. */
      vsd_send_command (io, NULL, 
			"CA_ADD %s",
			argv[1]  /* cn */); 
      status = vsd_process_reply (io);
      break;
    case CMD_CA_CERTDEL: /* Deprecated. */
      vsd_send_command (io, NULL, 
			"CA_DEL %s",
			argv[1]  /* cn */); 
      status = vsd_process_reply (io);
      break;
    default:
      status = "mod_ca: command not recognised";
      break;
    }
  return status;
}

#else  /* ! VSD_LIB */
/* Return 0 if command processed, 1 if not recognised.  */
int mod_ca_cmd (struct connection *vc,
		int level, const char *command,
		const int argc, const char *argv[], const char *data)
{
  int token;
  struct command_tag *tag;

  tag = module_lookup_command (clist, level, command);
  if (tag == NULL)
    return 1;

  token = tag->token;
  switch (token)
    {
    case CMD_CA_VERSION:
      ca_version (vc, argc, argv);
      break;
#ifdef WITH_OPENSSL
    case CMD_CA_VERIFY:
      ca_verify (vc, argc, argv);
      break;
    case CMD_CA_LIST:
      ca_list (vc, argc, argv);
      break;
    case CMD_CA_ADD:
      ca_add (vc, argc, argv);
      break;
    case CMD_CA_DEL:
      ca_del (vc, argc, argv);
      break;
    case CMD_CA_CERTGET:
      ca_certget (vc, argc, argv);
      break;
    case CMD_CA_CERTPUT:
      ca_certput (vc, argc, argv, data);
      break;
    case CMD_CA_KEYGET:
      ca_keyget (vc, argc, argv);
      break;
    case CMD_CA_KEYPUT:
      ca_keyput (vc, argc, argv, data);
      break;
    case CMD_CA_CRLGET:
      ca_crlget (vc, argc, argv);
      break;
    case CMD_CA_CRLPUT:
      ca_crlput (vc, argc, argv, data);
      break;
    case CMD_CA_CACERTGET:
      ca_cacertget (vc, argc, argv);
      break;
    case CMD_CA_CACERTPUT:
      ca_cacertput (vc, argc, argv, data);
      break;
    case CMD_CA_CERTADD:
      ca_add (vc, argc, argv);
      break;
    case CMD_CA_CERTDEL:
      ca_del (vc, argc, argv);
      break;
#endif  /* WITH_OPENSSL */
    default:
      return 1;
      break;
    }

  return 0;
}
#endif  /* VSD_LIB */

module ca_module =
{
  mod_ca_check_cmd,
  mod_ca_cmd
};




