/* Timezone manipulation for virtual servers
   Copyright (c) 2001 Designer Servers Ltd.  All rights reserved.
   Written by Nick Burrett <nick@dsvr.net>.  */

#include "config.h"
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>
#include <pwd.h>
#include <grp.h>
#include <dirent.h>
#include <unistd.h>
#include <sys/stat.h>

#include "vsd.h"
#include "libvsd.h"
#include "mod_tz.h"

#define LOCALTIME "/etc/localtime"
#define TZROOT "/usr/share/zoneinfo/"
#define TZTAB TZROOT "iso3166.tab"
#define TZZONETAB TZROOT "zone.tab"

/* Return a list of available time zones as country names.  */
void tz_list (struct connection *vc, int argc, char *argv[])
{
  struct vsd_file_line *head, *line;

  /* TZ_LIST */

  if (vsd_load_file (&head, vc->server_root, TZTAB))
    {
      io_buffer_store (vc, "-ERR Cannot load timezone tab file");
      return;
    }
  for (line = head; line; line = line->next)
    {
      char *p = line->line;
      if (*p == '#' || *p == '\0')
	continue;
      /* Skip two letter country code.  */ 
      p += 2;
      /* Skip any white space.  */
      while (*p && isspace (*p))
	p++;
      io_buffer_store (vc, "%s\n", p);
    }
  vsd_free_file (head);

#if 0
  /* TZ_LIST */
  DIR *dir1, *dir2;
  struct dirent *ent1, *ent2;
  struct stat sb;
  char *file1, *file2;

  file1 = (char *) malloc (sizeof (TZDIR) + 32);
  file2 = (char *) malloc (sizeof (TZDIR) + 32 + 32);
  if (! file1 || ! file2)
    {
      io_buffer_store (vc, "-ERR Out of memory");
      return;
    }

  dir1 = opendir (TZDIR);
  if (dir1 == NULL)
    io_buffer_store (vc, "-ERR Cannot get timezone information");
  else
    {
      while ((ent1 = readdir (dir1)))
	{
	  if (ent1->d_name[0] == '.' || ! strcmp (ent1->d_name, "Etc")
	      || ! strcmp (ent1->d_name, "SystemV")
	      || ! strcmp (ent1->d_name, "posix")
	      || ! strcmp (ent1->d_name, "right"))
	    continue;
	  strcpy (file1, TZDIR "/");
	  strcat (file1, ent1->d_name);
	  stat (file1, &sb);
	  if (S_ISDIR (sb.st_mode))
	    {
	      dir2 = opendir (file1);
	      if (dir2 != NULL)
		{
		  while ((ent2 = readdir (dir2)))
		    {
		      if (ent2->d_name[0] == '.')
			continue;
		      /*strcpy (file2, file1);
			strcat (file2, "/");
			strcat (file2, ent2);*/
		      /* Output the region name and the country name.  */
		      io_buffer_store (vc, "%s %s\n", ent1->d_name,
				       ent2->d_name);
		    }
		  closedir (dir2);
		}
	    }
	}
      closedir (dir1);
    }
  free (file1);
  free (file2);
#endif
}

/* Change the time zone for a virtual server to that of the country
   given as an argument.  */
void tz_change (struct connection *vc, int argc, char *argv[])
{
  struct vsd_file_line *head, *line;
  char country_code[4];
  int modified = 0;

  /* TZ_CHANGE <country name>  */
  country_code[0] = '\0';

  /* Lookup the country name and find the ISO 3166 2-character
     country code.  */
  if (vsd_load_file (&head, vc->server_root, TZTAB))
    {
      io_buffer_store (vc, "-ERR Cannot load timezone tab file: %m");
      return;
    }

  for (line = head; line; line = line->next)
    {
      char *p = line->line;
      if (*p == '#' || *p == '\0')
	continue;
      /* Skip two letter country code.  */ 
      p += 2;
      /* Skip any white space.  */
      while (*p && isspace (*p))
	p++;
      if (strcmp (argv[0], p) == 0)
	{
	  country_code[0] = line->line[0];
	  country_code[1] = line->line[1];
	}
    }
  vsd_free_file (head);

  if (country_code[0] == '\0')
    {
      io_buffer_store (vc, "-ERR Couldn't find ISO country code for %s\n",
		       argv[0]);
      return;
    }

  /* Lookup the appropriate timezone data based on the ISO 3166 2-character
     country code.  */
  vsd_load_file (&head, vc->server_root, TZZONETAB);
  for (line = head; line; line = line->next)
    {
      char *p = line->line;
      char *path, *path1;
      if (*p == '#' || *p == '\0')
	continue;

      if (p[0] == country_code[0] && p[1] == country_code[1])
	{
	  char *q;
	  /* Skip country code.  */
	  p += 2;
	  while (*p && isspace (*p))
	    p++;
	  /* Skip coordinates.  */
	  while (*p && ! isspace (*p))
	    p++;
	  while (*p && isspace (*p))
	    p++;
	  /* Found timezone, now get it.  */
	  q = p;
	  while (*q && ! isspace (*q))
	    q++;

	  path = (char *) malloc (4 + strlen (TZROOT) + q - p);
	  strcpy (path, "/" TZROOT);
	  strncat (path, p, q - p);

	  path1 = (char *) malloc (strlen (vc->server_root) + 2
				   + strlen (LOCALTIME));
	  strcpy (path1, vc->server_root);
	  strcat (path1, LOCALTIME);

	  if (! unlink (path1))
	    symlink (path, path1);
	  free (path);
	  free (path1);
	  modified = 1;
	}
    }
  vsd_free_file (head);
  if (! modified)
    io_buffer_store (vc, "-ERR Couldn't modify timezone file");
}
