/* mod_vd.c - managed domain name module interface.
 */

/*
 * VSD - Virtual Server Daemon
 * Copyright (c) 2000 - 2001 Idaya Ltd.
 *
 * VSD is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * VSD is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with VSD - see the COPYING file;  if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA.
 */

/*
 * See the AUTHORS file for a list of contributors to VSD.
 * See the ChangeLog files for a list of changes.
 */

/* DAMION: ONLY REPLACED MOD_VS TO MOD_VD, NO IMPLEMENTATION CHANGES */

#include <stdio.h>
#include "module.h"
#include "mod_vd.h"
#include "vd.h"

static struct command_tag clist[] =
{
  {
    CMD_VD_ADD, ADMIN_VS,
    "Create a new virtual domain account",
    "VD_ADD", 6, "<vs> <domain> <numuid> <max vusers> <max mailbox> <max messages>",
    "vd_add", 7, "<hs> <vs> <domain> <numuid> <max vusers> <max mailbox> <max messages>"
  },
  {
    CMD_VD_DELETE, ADMIN_VS,
    "Delete a virtual server account",
    "VD_DELETE", 2, "<vs> <domain>",
    "vd_delete", 3, "<hs> <vs> <domain>"
  },
  {
    CMD_VD_DETAILS, ADMIN_VD,
    "Return config details for a virtual server",
    "VD_DETAILS", 0, "[<domain name>]",
    "vd_details", 1, "<hs> [<vs> [<domain name>]]"
  },
  {
    CMD_VD_MODIFY, ADMIN_VS,
    "Modify characteristics of a virtual server",
    "VD_MODIFY", 3, "<vs> <old-dn> <new-dn> [<new-users>]",
    "vd_modify", 4, "<hs> <vs> <old-dn> <new-dn> [<new-users>]"
  },
  {
    CMD_VD_ALIASADD, ADMIN_VS,
    "Add a domain alias to an exixting managed domain",
    "VD_ALIASADD", 3, "<vs> <domain> <domain alias>",
    "vd_aliasadd", 4, "<hs> <vs> <domain> <domain alias>"
  },
  {
    CMD_VD_ALIASDEL, ADMIN_VS,
    "Remove a domain alias from a managed domain",
    "VD_ALIASDEL", 3, "<vs> <domain> <domain alias>",
    "vd_aliasdel", 4, "<hs> <vs> <domain> <domain alias>",
  },
  {
    0, 0, NULL, NULL, 0, NULL, NULL, 0, NULL
  }
};

/* Return -1 if `cmd' does not exist.
   Return 1 if there is a syntax error.
   Return 0 if everything is cool.  */
int mod_vd_check_cmd (int level, const char *cmd, int argc, char *argv[],
		      struct command_tag **list)
{
  if (cmd == NULL)
    {
      /* Return a pointer to all commands.  */
      *list = clist;
      return 0;
    }

  /* Find command in clist.  */
  *list = module_lookup_command (clist, level, cmd);
  if (*list == NULL)
    return -1;  /* Command does not exist.  */

  /* Check command syntax.  Pretty basic at the moment.  */
#ifdef VSD_LIB
  if (argc < (*list)->vsdadm_args)
    return 1; /* Syntax error.  */
#else
  if (argc < (*list)->vsd_args)
    return 1; /* Syntax error.  */
#endif

  /* Syntax is ok.  */
  return 0;
}

#ifdef VSD_LIB
char *mod_vd_cmd (void *io, int level, const char *cmd,
		  int argc, char **argv)
{
  struct command_tag *list;
  char *status = NULL;

  list = module_lookup_command (clist, level, cmd);
  if (! list)
    return "notexist";

  vsd_set_level (io, list->level);
  switch (list->token)
    {
    case CMD_VD_ADD:
      vsd_send_command (io, NULL, "VD_ADD %s %s %s %s %s %s",
			argv[1] /* virtual server */, argv[2] /* domain name */,
                        argv[3] /* numuid */, argv[4] /* vusers */, argv[5], /* mbox */
                        argv[6] /* messages */);
      break;
    case CMD_VD_DELETE:
      vsd_send_command (io, NULL, "VD_DELETE %s %s", argv[1] /* virtual server */,
                                                  argv[2] /* domain name */);
      break;
    case CMD_VD_DETAILS:
      if (argv[1] == NULL)
        {
          vsd_send_command (io, NULL, "VD_DETAILS");
        }
      else
        {
          vsd_set_level (io, ADMIN_VS);
          if (argv[2] == NULL)
            vsd_send_command (io, NULL, "VD_DETAILS");
          else
            vsd_send_command (io, NULL, "VD_DETAILS %s", argv[2] /* domain name */);
        }
      break;
    case CMD_VD_MODIFY:
      vsd_send_command (io, NULL, "VD_MODIFY %s %s %s %s",
			argv[1] /* virtual server */, argv[2] /* old-dn */, 
                        argv[3] /* new-dn */, argv[4] /* new-users */);
      break;
    case CMD_VD_ALIASADD:
      vsd_send_command (io, NULL, "VD_ALIASADD %s %s %s", argv[1] /* virtual server */,
                        argv[2] /* domain */, argv[3] /* domain alias */);
      break;
    case CMD_VD_ALIASDEL:
      vsd_send_command (io, NULL, "VD_ALIASDEL %s %s %s", argv[1] /* virtual server */,
                        argv[2] /* domain */, argv[3] /* domain alias */);
      break;
    default:
      return "mod_vd: command not recognised";
      break;
    }

  status = vsd_process_reply (io);
  return status;
}

#else /* ! VSD_LIB */
/* Return 0 if command processed, 1 if not recognised.  */
int mod_vd_cmd (struct connection *vc, int level, const char *command,
		int argc, char *argv[], const char *data)
{
  int token;
  struct command_tag *tag;

  tag = module_lookup_command (clist, level, command);
  if (tag == NULL)
    return 1;

  token = tag->token;
  switch (token)
    {
    case CMD_VD_ADD:
      vd_account_create (vc, argc, argv);
      break;
    case CMD_VD_DELETE:
      vd_account_delete (vc, argc, argv);
      break;
    case CMD_VD_DETAILS:
      vd_details (vc, level, argc, argv);
      break;
    case CMD_VD_MODIFY:
      vd_modify (vc, argc, argv);
      break;
    case CMD_VD_ALIASADD:
      vd_alias_add (vc, argc, argv);
      break;
    case CMD_VD_ALIASDEL:
      vd_alias_del (vc, argc, argv);
      break;
    default:
      return 1;
      break;
    }

  return 0;
}
#endif

module vd_module =
{
  mod_vd_check_cmd,
  mod_vd_cmd
};
