/* vd.c - vd module for manipulation of managed domains.
 */

/*
 * VSD - Virtual Server Daemon
 * Copyright (c) 2000 - 2001 Idaya Ltd.
 *
 * VSD is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * VSD is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with VSD - see the COPYING file;  if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA.
 */

/*
 * See the AUTHORS file for a list of contributors to VSD.
 * See the ChangeLog files for a list of changes.
 */

#include "config.h"
#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <signal.h>
#include <errno.h>
#include <unistd.h>
#include <netdb.h>
#include <netinet/in.h>

#include <sys/stat.h>
#include <sys/types.h>
#include <fcntl.h>

#include <sys/dir.h> 

#include "vsd.h"
#include "vd.h"
#include "module.h"
#include "vsd.h"
#include "../../../libvsd/libvsd.h"

#define VD_DOMAIN_LENGTH_MAX 64
#define VD_USERS_MAX 65000
#define VD_STR_MAX 256
#define VD_TRIES_MAX 3
#define VD_TIME_WAIT 5

static void send_output (struct connection *vc,
			 const char *server_root, const char *file)
{
  char *temp;
  FILE *stream;
  char line[256];

  if (server_root)
    {
      temp = (char *) malloc (strlen (server_root) + strlen (file) + 3);
      strcpy (temp, server_root);
      strcat (temp, file);
    }
  else
    temp = (char *) file;

  stream = fopen (temp, "r");
  if (stream != NULL)
    {
      while (fgets (line, sizeof (line), stream))
	io_buffer_store (vc, "%s", line);

      fclose (stream);
    }
  unlink (temp);

  if (server_root)
    free (temp);
}

/********************************************************************************
 * Description: Returns a pointer to the domain hosting the alias.
 *
 *     Returns: pointer to vsd_vs on successs,
 *              NULL otherwise
 *******************************************************************************/
struct vsd_vs *
vd_getvdalias (const struct vsd_vs_map *map,
               const char *domain_name)
{
  int x, y;

  if (domain_name == NULL || domain_name[0] == '\0')
    {
      errno = EINVAL;
      return NULL;
    }

  for (x = 0; x < map->domains; x++)
    {
      for (y = 0; y < map->domain[x].vdalias_total; y++)
        {
          if (strcmp (domain_name, map->domain[x].vdalias[y]) == 0) 
              return &map->domain[x];
        }
     }
    
  return NULL;
}

/********************************************************************************
 * Description: VSD DN (managed domain name) account creation
 *
 *     Returns: void
 *******************************************************************************/
void 
vd_account_create (struct connection *vc, 
                   int argc, 
                   char *argv[])
{
  FILE *stream;
  int ch, numuid;
  struct vsd_vs vd;
  struct vsd_vs *vs;
  struct connection *vd_config_vc_copy;
  const char *status;

  memset (&vd, 0, sizeof (struct vsd_vs));

  /* VD_ADD <virtual server> <domain name> <numuid> <num addr> <num mbx> <num msgs> */

  /* Get options.  */
  /* Set the name of the domain */
  if (*argv[1] == '\0' || argv[1] == NULL)
    {
      io_buffer_store (vc, "-ERR domain name argument is missing");
      return;
    }
  strcpy (vd.name, argv[1]);

  /* set the virtual server the domain is hosted by */
  if (*argv[0] == '\0' || argv[0] == NULL)
    {
      io_buffer_store (vc, "-ERR virtual server name argument is missing");
      return;
    }
  strcpy (vd.hs, argv[0]);
  
  /* Does the virtual server exist? */
  if ((vs = vsd_getmapent (vc->map, vd.hs, NULL)) == NULL)
    {
      io_buffer_store (vc, "-ERR virtual server does not exist");
      return;
    } 

  /* set the number of users */
  vd.users = atoi (argv[2]);

  /* set the max number of vusers */
  vd.max_vuser = atoi (argv[3]);

  /* set the max number of vmbx */
  vd.max_vmbx = atoi (argv[4]);

  /* set the max number of vmsg */
  vd.max_vmsg = atoi (argv[5]);

  if ((status = vd_valid_domainname (vd.name)))
    {
      io_buffer_store (vc, "-ERR domain name is invalid: %s", status);
      return;
    }

  /* Does managed domain exist ?  */
  if (vsd_getvdent (vc->map, vd.name))
    {
      io_buffer_store (vc, "-ERR managed domain %s already exists", vd.name);
      return;
    }

  /* Does managed domain exists as an alias? */
  if (vd_getvdalias (vc->map, vd.name))
    {
      io_buffer_store (vc, "-ERR managed domain %s already exists as an alias", vd.name);
      return;
    }

  /* Check we can add a virtual domain */
  if ((status = vd_check_domain_limit (vc->map, vs, vsd_map_server_root (vc->map, vs->name))) != NULL )
    {
      io_buffer_store (vc, "-ERR unable to allocate managed domain: %s", status);
      return;
    }

  /* Find a free UID allocation.  */
  if (vd.users <= 0 || vd.users >= VD_USERS_MAX)
    {
      io_buffer_store (vc, "-ERR Number of users (%d) is incorrect", vd.users);
      return;
    }

  /* vd.startuid = vsd_alloc_uid (vc->map, vd.users); */
  vd.startuid = vs->startuid;

  if (vd.startuid == -1)
    {
      io_buffer_store (vc, "-ERR Could not allocate a UID chunk");
      return;
    }

  if (vsd_map_add_vd (vc->map, &vd))
    io_buffer_store (vc, "-ERR Cannot create account: %s",
		     (errno == EMFILE) ? "host server is full"
		     : strerror (errno));
  else
    {
      vd_config_vc_copy = vc;
      if ((status = vd_config_domain (vd_config_vc_copy, &vd)) != NULL)
        {
          io_buffer_store (vc, "-ERR configuring domain: %s: %m", status);
        }
      /* else
        io_buffer_store (vc, "-ERR Domain created but not physically: %s: %m", stream);
        return;
      */

    vsd_map_save (vc->map);
    return;
    }
}

/********************************************************************************
 * Description: VSD DN (managed domain name) account deletion
 *
 *     Returns: void
 *******************************************************************************/
void 
vd_account_delete (struct connection *vc, 
                   int argc, 
                   char *argv[])
{
  FILE *stream;
  struct vsd_vs *vd = NULL;
  struct vsd_vs *vs;
  char line[128];
  const char *status;
  int x;

  /* DN_DELETE <virtual server> <domain name> */

  /* Check that arguments supplied are valid */
  if (*argv[1] == '\0' || argv[1] == NULL)
    {
      io_buffer_store (vc, "-ERR domain name argument is missing");
      return;
    }

  if (*argv[0] == '\0' || argv[0] == NULL)
    {
      io_buffer_store (vc, "-ERR virtual server name argument is missing");
      return;
    }

      
  /* check validity of the domain name */
  if ((status = vd_valid_domainname (argv[1])))
    {
      io_buffer_store (vc, "-ERR managed domain name is invalid: %s", status);
      return;
    }

  vd = vsd_getvdent (vc->map, argv[1]);
  if (vd == NULL)
    {
      io_buffer_store (vc, "-ERR managed domain %s is not known", argv[1]);
      return;
    }

  /* check that the domain is hosted on the specified vs */
  if (strcmp (argv[0], vd->hs) != 0)
    {
      io_buffer_store (vc, "-ERR domain is not hosted on the specified virtual server");
      return;
    }

  /* Check that the vs the domain is hosted on exists */
  if (vs = vsd_getmapent (vc->map, vd->hs, NULL))
    {
      /* Yes it does so schedule the domain to be physically removed */
      /* stream = fopen ("/tmp/vsd.accounts", "a"); */
      /* if (stream != NULL) */
      /*   { */
      /*     fprintf (stream, "domaindelete %s %s %s\n", vd->hs, vd->name, */
      /*                                                vsd_ip_ntoa (vs->ip)); */
      /*    fclose (stream); */
      /*  } */
      /* else */

      if ((status = vd_remove_domain (vc, vd)) != NULL)
        {
          io_buffer_store (vc, "-ERR Domain deleted but not physically: %s: %m", status);
        }

      for (x = 0; x < vd->vdalias_total; x++)
        {
          if ((status = vd_remove_domain_alias (vc, vd, vd->vdalias[x])) != NULL)
            io_buffer_store (vc, "-ERR Domain alias %s deleted but not physically", vd->vdalias[x]);
        }

      vsd_map_delete_vd (vc->map, argv[1]);
      vsd_map_save (vc->map);
    }
  
}

/********************************************************************************
 * Description: VSD VD (managed domain name) subroutine to check the supplied
 *              args are valid and call vd_map_add_vdalias
 *
 *     Returns: void
 *******************************************************************************/
void
vd_alias_add  (struct connection *vc,
                 int argc,
                 char *argv[])
{
  /* VD_ALIASADD <virtual server> <domain> <domain alias> */
  struct vsd_vs *vd;
  struct vsd_vs *vs;
  const char *status;
  int map_add_ret = 0;

  
  /* Check that arguments supplied are valid */
  if (*argv[0] == '\0' || argv[0] == NULL)
    {
      io_buffer_store (vc, "-ERR virtual server name argument is missing");
      return;
    }

  if (*argv[1] == '\0' || argv[1] == NULL)
    {
      io_buffer_store (vc, "-ERR domain name argument is missing");
      return;
    }

  if (*argv[2] == '\0' || argv[2] == NULL)
    {
      io_buffer_store (vc, "-ERR domain alias argument is missing");
      return;
    }

  /* check validity of the domain name */
  if ((status = vd_valid_domainname (argv[1])))
    {
      io_buffer_store (vc, "-ERR managed domain name is invalid: %s", status);
      return;
    }

  /* check the validity of the domain alias */
  if ((status = vd_valid_domainname (argv[2])) != NULL)
    {
      io_buffer_store (vc, "-ERR domain alias name is invalid: %s", status);
      return;
    }

  vd = vsd_getvdent (vc->map, argv[1]);
  if (vd == NULL)
    {
      io_buffer_store (vc, "-ERR managed domain %s is not known", argv[1]);
      return;
    }

  /* check that the domain is hosted on the specified vs */
  if (strcmp (argv[0], vd->hs) != 0)
    {
      io_buffer_store (vc, "-ERR domain is not hosted on the specified virtual server");
      return;
    }

  /* Check that the vs the domain is hosted on exists */
  if ((vs = vsd_getmapent (vc->map, vd->hs, NULL)) == NULL)
    {
      io_buffer_store (vc, "-ERR hosting virtual server does not exist");
      return;
    }

  map_add_ret = vd_map_add_vdalias (vc->map, vd->name, argv[2]);
  if (map_add_ret == 0)
    {
      if (vsd_map_save (vc->map))
        io_buffer_store (vc, "-ERR cannot update vsd.conf: %m");
    }
  else if (map_add_ret == 1)
    {
      io_buffer_store (vc, "-ERR domain name %s does not exist", vd->name);
      return;
    }
  else if (map_add_ret == 2)
    {
      io_buffer_store (vc, "-ERR domain name alias is defined on another virtual server");
      return;
    }
  else if (map_add_ret == 3)
    {
      io_buffer_store (vc, "-ERR domain name alias is defined on this virtual server");
      return;
    }
  else if (map_add_ret == 4)
    {
      io_buffer_store (vc, "-ERR cannot add %s: memory allocation error", argv[2]);
      return;
    }
  else
    {
      io_buffer_store (vc, "-ERR vd_add_alias: unknown error (%d) occurred", map_add_ret); 
      return;
    }

  /* Now physically create the alias: */
  if ((status = vd_config_domain_alias (vc, vd, argv[2])) != NULL)
    {
      io_buffer_store (vc, "-ERR unable to configure domain alias: %s", status);
      return;
    }

}

/********************************************************************************
 * Description: VSD VD (managed domain name) subroutine to check the supplied
 *              args are valid and call vd_map_del_vdalias
 *
 *     Returns: void
 *******************************************************************************/
void
vd_alias_del  (struct connection *vc,
                 int argc,
                 char *argv[])
{
  /* VD_ALIASDEL <virtual server> <domain> <domain alias> */

  struct vsd_vs *vd;
  struct vsd_vs *vs;
  const char *status;
  int map_del_ret = 0;


  /* Check that arguments supplied are valid */
  if (*argv[0] == '\0' || argv[0] == NULL)
    {
      io_buffer_store (vc, "-ERR virtual server name argument is missing");
      return;
    }

  if (*argv[1] == '\0' || argv[1] == NULL)
    {
      io_buffer_store (vc, "-ERR domain name argument is missing");
      return;
    }

  if (*argv[2] == '\0' || argv[2] == NULL)
    {
      io_buffer_store (vc, "-ERR domain alias argument is missing");
      return;
    }

  /* check validity of the domain name */
  if ((status = vd_valid_domainname (argv[1])))
    {
      io_buffer_store (vc, "-ERR managed domain name is invalid: %s", status);
      return;
    }

  /* check the validity of the domain alias */
  if ((status = vd_valid_domainname (argv[2])) != NULL)
    {
      io_buffer_store (vc, "-ERR domain alias name is invalid: %s", status);
      return;
    }

  vd = vsd_getvdent (vc->map, argv[1]);
  if (vd == NULL)
    {
      io_buffer_store (vc, "-ERR managed domain %s is not known", argv[1]);
      return;
    }

  /* check that the domain is hosted on the specified vs */
  if (strcmp (argv[0], vd->hs) != 0)
    {
      io_buffer_store (vc, "-ERR domain is not hosted on the specified virtual server");
      return;
    }

  /* Check that the vs the domain is hosted on exists */
  if ((vs = vsd_getmapent (vc->map, vd->hs, NULL)) == NULL)
    {
      io_buffer_store (vc, "-ERR hosting virtual server does not exist");
      return;
    }

  map_del_ret = vd_map_delete_vdalias (vc->map, vd->name, argv[2]);
  if (map_del_ret == 0)
    {
      if (vsd_map_save (vc->map))
        io_buffer_store (vc, "-ERR cannot update vsd.conf: %m");
    }
  else if (map_del_ret == 1)
    {
      io_buffer_store (vc, "-ERR domain name %s does not exist", vd->name);
      return;
    }
  else if (map_del_ret == 2)
    {
      io_buffer_store (vc, "-ERR domain name alias does not exist");
      return;
    }
  else if (map_del_ret == 3)
    {
      io_buffer_store (vc, "-ERR domain name alias is defined on this virtual server");
      return;
    }
  else if (map_del_ret == 4)
    {
      io_buffer_store (vc, "-ERR cannot add %s: memory allocation error", argv[2]);
      return;
    }
  else
    {
      io_buffer_store (vc, "-ERR vd_del_alias: unknown error (%d) occurred", map_del_ret);
      return;
    }

  /* Now physically remove the alias: */
  if ((status = vd_remove_domain_alias (vc, vd, argv[2])) != NULL)
    {
      io_buffer_store (vc, "-ERR unable to remove the domain alias configuration: %s", status);
      return;
    }
}

/********************************************************************************
 * Description: VSD DN (managed domain name) Return configuration details about 
 *              a managed domain
 *
 *     Returns: void
 *******************************************************************************/
void 
vd_details (struct connection *vc, 
            int level, 
            int argc, 
            char *argv[])
{
  struct vsd_vs *vd, *vs;
  const char *st;
  int x, y;

  /* VD_DETAILS [<domain name>]
     Returns a single line per domain name:
       name=<domain name> uid=<starting uid> users=<No. of users>
       hostingserver=<name of virtual domain hosting the domain>
  */

  /* We need ADMIN_VS to be able to list all DNs, or just one (on particular vs), and ADMIN_VD 
     to just be able to list their own. */

  if ((level == ADMIN_VS && argc == 1) || (level == ADMIN_VD && argc == 0))
    {
      if (argc == 1)
	{
          /* Domain name has been given, so we just want to print out the information */
          /* about that domain. */

	  /* Check validity of domain name.  */
	  if ((st = vd_valid_domainname (argv[0])))
	    {
	      io_buffer_store (vc, "-ERR domain name is invalid: %s", st);
	      return;
	    }

	  vd = vsd_getvdent (vc->map, argv[0]);
	  if (vd == NULL || (strcmp (vd->hs, vc->virtual_server) != 0))
	    {
	      io_buffer_store (vc, "-ERR Managed domain %s is not known", argv[0]);
	      return;
	    }
	}
      else
        {
	  vd = vsd_getvdent (vc->map, vc->virtual_server);
        }

      io_buffer_store (vc, "name=%s uid=%d users=%d maxvuser=%d maxmbx=%d maxmsg=%d hostingserver=%s ",
		       vd->name, vd->startuid, vd->users, vd->max_vuser, 
                       vd->max_vmbx, vd->max_vmsg, vd->hs); 

      for (x = 0; x < vd->vdalias_total; x++)
        {
          io_buffer_store (vc, "alias=%s ", vd->vdalias[x]);
        }
      io_buffer_store (vc, "\n");
    }
  else if (level == ADMIN_VS)
    {
      for (x = 0; x < vc->map->domains; x++)
        {
          vd = &vc->map->domain[x];
          if (vd->name[0] != '\0' && (strcmp (vc->virtual_server, vd->hs) == 0))
            {
              io_buffer_store (vc,
                         "name=%s uid=%d users=%d maxvuser=%d maxmbx=%d maxmsg=%d hostingserver=%s ",
                         vd->name, vd->startuid, vd->users, vd->max_vuser, 
                         vd->max_vmbx, vd->max_vmsg, vd->hs);
              for (y = 0; y < vd->vdalias_total; y++)
                {
                  io_buffer_store (vc, "alias=%s ", vd->vdalias[y]);
                }
              io_buffer_store (vc, "\n");
            }
        }
      }
  else if (level == ADMIN_HOST)
    {
      for (x = 0; x < vc->map->domains; x++)
        {
          vd = &vc->map->domain[x];
          if (vd->name[0] != '\0')
            {
              io_buffer_store (vc,
                         "name=%s uid=%d users=%d maxvuser=%d maxmbx=%d maxmsg=%d hostingserver=%s ",
                         vd->name, vd->startuid, vd->users, vd->max_vuser, 
                         vd->max_vmbx, vd->max_vmsg, vd->hs);
              for (y = 0; y < vd->vdalias_total; y++)
                {
                  io_buffer_store (vc, "alias=%s ", vd->vdalias[y]);
                }
              io_buffer_store (vc, "\n");
            }
        }
    }
  else
    io_buffer_store (vc, "Permission denied");
}

/********************************************************************************
 * Description: VSD DN (managed domain name) modify a vd account
 *
 *     Returns: void
 *******************************************************************************/
void 
vd_modify (struct connection *vc, 
           int argc, 
           char *argv[])
{
  const char *status;
  struct vsd_vs *oldvd, *newvd;
  int newusers, x;

  /* DN_MODIFY <old-vd> <new-vd> <new-users> */

  /* Check arguments.  */
  status = vd_valid_domainname (argv[0]);
  if (status)
    {
      io_buffer_store (vc, "-ERR Old domain name is invalid: %s", status);
      return;
    }

  /* Check that old domain exists */
  oldvd = vsd_getvdent (vc->map, argv[0]);
  if (oldvd == NULL)
    {
      io_buffer_store (vc, "-ERR domain name %s is not known", argv[0]);
      return;
    }

  if (strcmp (argv[0], argv[1]) != 0)
    {
      /* domain name is changing.  */
      status = vd_valid_domainname (argv[1]);
      if (status)
        {
          io_buffer_store (vc, "-ERR New domain name is invalid: %s", status);
          return;
        }


      /* Check new domain name is available if so, rename VSDomain in /etc/vsd.conf */
      newvd = vsd_getvdent (vc->map, argv[1]);
      if (newvd != NULL)
        {
          io_buffer_store (vc, "-ERR domain name %s is already in use", argv[1]);
          return;
        }

      /* Haven't returned yet so must be OK to insert new name */
      /* Find entry with matching name and copy in new one */
      for (x = 0; x < vc->map->domains; x++);
        {
          if (strcmp (oldvd[x].name, argv[0]) == 0)
            {
              strcpy (oldvd[x].name, argv[1]);
            }
         }

      vsd_map_save (vc->map);
    }

  if (argc == 3)
    {
      /* Requested number of users to be changed */

      newusers = (int) argv[2];

      if (newusers <= 0 || newusers >= 65000)
        {
          io_buffer_store (vc, "-ERR Number of users (%d) is incorrect", newusers);
          return;
        }

      /* Find entry with correct name, and set the new number of users */
      for (x = 0; x < vc->map->domains; x++)
        {
          if (strcmp (oldvd[x].name, argv[0]) == 0)
            {
              oldvd[x].users = newusers;
            }
        }
    }
}

/********************************************************************************
 * Description: VSD DN (managed domain name) subroutine to check that the domain 
 *              name supplied is vaild
 *
 *     Returns: NULL on success, otherwise an error string
 *******************************************************************************/
const char *vd_valid_domainname (const char *vd)
{
  if (vd == NULL)
    return "is_domainname_valid: passed NULL argument";

  if (strspn (vd, "abcdefghijklmnopqrstuvwxyz0123456789-.") != strlen (vd))
    return "contains unsuitable characters";

  /* Restrict domain names to VD_DOMAIN_LENGHT_MAX characters.  */
  if (strlen (vd) > VD_DOMAIN_LENGTH_MAX)
    return "name is too long";

  return NULL;
}

/********************************************************************************
 * Description: Add an VD (managed virtual domain) alias to an existing domain 
 *              entry.
 * Returns: 0 on success
 *          1 if `domain_name' does not exist.
 *          2 if `vdalias' is already defined by another virtual server.
 *          3 if `vdalias' is already defined by `virtual_server'.
 *          4 if memory allocation failed. 
 *******************************************************************************/
int 
vd_map_add_vdalias (struct vsd_vs_map *map,
                    const char *domain_name, 
                    const char *vdalias)
{
  struct vsd_vs *vd, *vd1;
  int x;

  vd = vsd_getvdent (map, domain_name);
  if (vd == NULL)
    return 1;

  /* check the name isn't already a domain */
  if (vsd_getvdent (map, vdalias) != NULL)
    return 2;

  /* check the name isn't already an alias */
  vd1 = vd_getvdalias (map, vdalias);
  if (vd1 != NULL && vd != vd1)
    return 2;

  if (vd == vd1)
    return 3;

  x = ++vd->vdalias_total;
  if (x == 1)
    vd->vdalias = malloc (sizeof (char *));
  else
    vd->vdalias = realloc (vd->vdalias, sizeof (char *) * x);
  if (vd->vdalias == NULL)
    return 4;

  vd->vdalias[x - 1] = strdup (vdalias);

  return 0;
}

/********************************************************************************
 * Description: Remove a VD alias from an existing domain entry.
 *
 * Returns: 0 on success
 *          1 if `domain_name' does not exist.
 *          2 if `vdalias' does not exist.
 *          3 if `vdalias' does not belong to `virtual_server'.
 *          4 if memory allocation failed. 
 *******************************************************************************/
int 
vd_map_delete_vdalias (struct vsd_vs_map *map,
                       const char *domain_name, 
                       const char *vdalias)
{
  struct vsd_vs *vd, *vd1;
  char *name;

  vd = vsd_getvdent (map, domain_name);
  if (vd == NULL)
    return 1;

  /* check the name isn't already an alias */
  vd1 = vd_getvdalias (map, vdalias);
  if (vd1 == NULL)
    return 2;

  if (vd != vd1)
    return 3;

  /* Quick case: 1 VD alias.  */
  if (vd->vdalias_total == 1)
    {
      free (vd->vdalias);
      vd->vdalias = NULL;
      vd->vdalias_total = 0;
    }
  else
    {
      int x, y;
      char **vd_aliases;

      /* Re-construct VD alias array without `vdalias'.  */
      vd_aliases = malloc (sizeof (char *) * (vd->vdalias_total - 1));
      if (vd_aliases == NULL)
        return 4;
      for (x = y = 0; x < vd->vdalias_total; x++)
        {
          if (strcmp (vd->vdalias[x], vdalias) != 0)
            vd_aliases[y++] = vd->vdalias[x];
        }
      vd->vdalias_total --;
      free (vd->vdalias);
      vd->vdalias = vd_aliases;
    }

  return 0;
}

/********************************************************************************
 * Description: virtual domain's procmailrc file path.
 *
 *     Returns: char * pointer to procmailrc file path.
 *              NULL on error.
 *******************************************************************************/
static char *
vd_procmailrc_path (const char *root,
                    const char *domain)
{
  static char procmailrc[VD_STR_MAX];

  sprintf (procmailrc, "%s/etc/procmailrcs/virtual/%s", (root ? root : ""), domain);

  return procmailrc;
}

/********************************************************************************
 * Description: virtual domain's procmailrc file path.
 *
 *     Returns: char * pointer to procmailrc file path.
 *              NULL on error.
 *******************************************************************************/
static char *
vd_procmailrc_template_path (const char *root)
{
  static char procmailrc[VD_STR_MAX];

  sprintf (procmailrc, "%s/etc/procmailrcs/", (root ? root : "/"));

  return procmailrc;
}

/********************************************************************************
 * Description: virtual domain's procmailrc template file path.
 *
 *     Returns: char * pointer to procmailrc template file path.
 *              NULL on error.
 *******************************************************************************/
static char *
vd_procmailrc_domain_template_file (const char *root)
{
  static char procmailrc_template[VD_STR_MAX];

  sprintf (procmailrc_template, "%s/pm-vsd-vd.rc", vd_procmailrc_template_path (root));

  return procmailrc_template;
}

/********************************************************************************
 * Description: virtual domain's procmailrc recipe file path.
 *     
 *     Returns: char * pointer to procmailrc domain recipe file path.
 *              NULL on error. 
 *******************************************************************************/
static char *
vd_procmailrc_domain_recipe_file (const char *root,
                                  const char *domain)
{ 
  static char domain_recipe[VD_STR_MAX];
  
  sprintf (domain_recipe, "%s/pm.rc", vd_procmailrc_path (root, domain));
  
  return domain_recipe;
} 

/********************************************************************************
 * Description: virtual domain's admin procmailrc template file path.
 *
 *     Returns: char * pointer to procmailrc template file path.
 *              NULL on error.
 *******************************************************************************/
static char *
vd_procmailrc_admin_template_file (const char *root)
{
  static char admin_template[VD_STR_MAX];

  sprintf (admin_template, "%s/pm-vsd-admin.rc", vd_procmailrc_template_path (root));

  return admin_template;
}

/********************************************************************************
 * Description: virtual domain's admin procmailrc recipe file path.
 *     
 *     Returns: char * pointer to procmailrc admin recipe file path.
 *              NULL on error.
 *******************************************************************************/
static char *
vd_procmailrc_admin_recipe_file (const char *root,
                                 const char *domain)
{ 
  static char admin_recipe[VD_STR_MAX];
  
  sprintf (admin_recipe, "%s/admin.rc", vd_procmailrc_path (root, domain));
  
  return admin_recipe;
}

/********************************************************************************
 * Description: virtual domain's admin directory path.
 *     
 *     Returns: char * pointer to file path.
 *              NULL on error.
 *******************************************************************************/
static char *
vd_procmailrc_admin_path (const char *root,
                          const char *domain)
{
  static char admin_path[VD_STR_MAX];

  sprintf (admin_path, "%s/admin/", vd_procmailrc_path (root, domain));

  return admin_path;
}

/********************************************************************************
 * Description: virtual domain's admin msg directory path.
 *     
 *     Returns: char * pointer to file path. 
 *              NULL on error.
 *******************************************************************************/
static char *
vd_procmailrc_admin_msg_path (const char *root,
                          const char *domain)
{
  static char msg_path[VD_STR_MAX];

  sprintf (msg_path, "%s/msg", vd_procmailrc_admin_path (root, domain));

  return msg_path;
}

/********************************************************************************
 * Description: virtual domain's admin index file path.
 *    
 *     Returns: char * pointer to procmailrc admin index file.
 *              NULL on error.
 *******************************************************************************/
static char *
vd_procmailrc_admin_index_file (const char *root,
                                const char *domain)
{
  static char admin_index[VD_STR_MAX];
 
  sprintf (admin_index, "%s/.index", vd_procmailrc_admin_msg_path (root, domain));
 
  return admin_index;
}

/********************************************************************************
 * Description: virtual domain's passwd file path.
 *
 *     Returns: char * pointer to passwd file path.
 *              NULL on error.
 *******************************************************************************/
static char *
vd_passwd_path (char *root,
                char *domain)
{
  static char path[VD_STR_MAX];

  sprintf (path, "%s/etc/virtual/%s", (root ? root : ""), domain);

  return path;
}

/********************************************************************************
 * Description: virtual domain's passwd file name.
 *
 *     Returns: char * pointer to passwd file name.
 *              NULL on error.
 *******************************************************************************/
static char *
vd_passwd_file (char *root,
                char *domain)
{
  static char file_name[VD_STR_MAX];

  sprintf (file_name, "%s/passwd", vd_passwd_path (root, domain));

  return file_name;
}


/********************************************************************************
 * Description: virtual domain's spool directory path.
 *
 *     Returns: char * pointer to spool directory path.
 *              NULL on error.
 *******************************************************************************/
static char *
vd_spool_path (char *root,
               char *domain)
{
  static char path[VD_STR_MAX];

  sprintf (path, "%s/var/spool/virtual/%s", (root ? root : ""), domain);

  return path;
}

/********************************************************************************
 * Description: Creates a random password string
 *
 *     Returns: char * pointer to the string.
 *              NULL on error.
 *******************************************************************************/
static char *
vd_passwd_string_create (const char *passwd_string)
{
    char *passwd_copy;
    
    passwd_copy = (char *) strdup (passwd_string);
    passwd_copy = (char *) strtok (passwd_copy, ".");
    return vsd_passwd_crypt (passwd_copy);
}

/********************************************************************************
 * Description: Creates a password file populated with user
 *
 *     Returns: char * pointer to the file.
 *              NULL on error.
 *******************************************************************************/
static char *
vd_passwd_file_create (const char *user, 
                       const char *domain_name,
                       const char *path_to_passwd_file)
{
    char *passwd_string;
    char passwd_file[VD_STR_MAX];
    FILE *stream;
    
    /* Obtain the passwd string */
    if ((passwd_string = vd_passwd_string_create (domain_name)) == NULL)
      return "error in creating passwd string";

    /* Set the full path to the passwd file */
    sprintf (passwd_file, "%s/passwd", path_to_passwd_file);

    /* Create the passwd file, open for writing */
    if ((stream = fopen (passwd_file, "w+")) == NULL)
      return "error opening passwd file";

    /* Write the passwd string to the file */
    fprintf (stream, "%s:%s\n", user, passwd_string);

    /* Close the passwd file */
    fclose (stream);
    return NULL;
}


/********************************************************************************
 * Description: Sets up the appropriate args for chown, and passes them to
 *              the run_command executable.
 *
 *     Returns: 0 on success
 *              -1 on error
 *******************************************************************************/
int
vd_chown_run_command (struct connection *vc,
                      const char *hosting_server,
                      char *user,
                      char *file_to_change)
{
  char args[VD_STR_MAX];
  int run_command_ret = 0;

  sprintf (args, "%s %s", user, file_to_change);
  run_command_ret = run_command (vc, vsd_map_server_root (vc->map, hosting_server),
                    "root", "/bin/chown", args, NULL, NULL);

  return run_command_ret;
}

/********************************************************************************
 * Description: Creates the directory specified by second arg. If fisrt arg is
 *              not null, will create it with the specified root dir.
 *
 *     Returns: NULL on success
 *              string on error
 *******************************************************************************/
static char *
vd_mkdir (const char *root_dir,
          const char *dir_to_make)
{
  char full_path[VD_STR_MAX];
  /* *full_path = (char *) malloc (strlen ((root_dir ? root_dir : "")) + strlen (dir_to_make)); */

  sprintf (full_path, "%s/%s", (root_dir ? root_dir : "/"), dir_to_make);
  if ((mkdir (full_path, (mode_t) 0755) == -1) && (errno != EEXIST))
    return strerror (errno);
  else
    return NULL;
}

/********************************************************************************
 * Description: Sets up the appropriate args for makemap, and passes them to 
 *              the run_command executable to regenerate the mailertable.db
 *
 *     Returns: 0 on success
 *              -1 on error
 *******************************************************************************/
int 
vd_refresh_mailertable (struct connection *vc,
                        const char *domain_root_path,
                        const char *database_file, /* mailtertable.db file */
                        const char *input_file /* mailtable file */)
{
  char args[VD_STR_MAX]; 
  int run_command_ret;

  /* args = (char *) malloc (strlen (database_file) + strlen ("hash ") + strlen (input_file) + 4); */
  sprintf (args, "hash %s < %s", database_file, input_file);

  run_command_ret = run_command (vc, domain_root_path, "admin", "/usr/sbin/makemap",
                                 args, NULL, NULL);

  return run_command_ret;
}

/********************************************************************************
 * Description: Returns the path to the mail directory
 *
 *     Returns: string * to mail directory
 *
 *******************************************************************************/
static char *
vd_mailertable_path (const char *root)
{
  static char mail_path[VD_STR_MAX];

  sprintf (mail_path, "%s/etc/mail", (root ? root : "/"));

  return mail_path;
}

/********************************************************************************
 * Description: Returns the path and name of the mailertable file
 *
 *     Returns: string * to mailertable file
 *
 *******************************************************************************/
static char *
vd_mailertable_file (const char *root)
{
  static char file_name[VD_STR_MAX];

  sprintf (file_name, "%s/mailertable", vd_mailertable_path (root));

  return file_name;
}

/********************************************************************************
 * Description: Returns the path and name of the mailertable database file
 *
 *     Returns: string * to mailertable database file
 *
 *******************************************************************************/
static char *
vd_mailertable_database_file (char *root)
{
  static char file_name[VD_STR_MAX];
 
  sprintf (file_name, "%s/mailertable.db", vd_mailertable_path (root));

  return file_name;
}

/********************************************************************************
 * Description: Appends a line to the /etc/mail/mailertable file for the domain
 *              specified on the command line.
 *
 *     Returns: 0 on success
 *              -1 on error
 *******************************************************************************/
int
vd_append_domain_mailertable (const char *domain_root_path,
                              const char *domain_identifier,
                              const char *domain_name)
{

  FILE *stream;

  if ((stream = fopen (vd_mailertable_file (domain_root_path), "a")) == NULL)
    return -1;

  /* <domain>    procmail:/etc/procmailrcs/virtual/<domain>/.pmrc */
  fprintf (stream, "%s    procmail:%s\n", domain_identifier, 
                                          vd_procmailrc_domain_recipe_file (NULL, domain_name)); 

  fclose (stream);
  return 0;
}

/********************************************************************************
 * Description: Removes the line containing <domain_name> from the
 *              /etc/mail/mailertable file.
 *
 *     Returns: 0 on success
 *              -1 on error
 *******************************************************************************/
static char *
vd_remove_domain_mailertable (const char *domain_root_path,
                              const char *domain_name)
{
  FILE *stream_tmp, *stream;
  struct vsd_file_line *line = NULL;
  struct vsd_file_line *mailertable_line = NULL;
  char mailertable_lock[VD_STR_MAX];
  char mailertable_tmp[VD_STR_MAX];
  char mailertable_fullpath[VD_STR_MAX];
  char domain_name_appended_space[VD_STR_MAX];
  int tries = 0;
  int fd;
  int compare;

  /* Create mailertable.tmp to write the new file to */
  sprintf (mailertable_lock, "%s.lock", vd_mailertable_file (domain_root_path));
  while ((fd = open (mailertable_lock, O_WRONLY | O_CREAT | O_EXCL, (mode_t) 0666)) == -1
         && errno == EEXIST)
    {
      /* mailetable.lock exists, so wait a while, and try again */
      if (++tries >= VD_TRIES_MAX)
        return "unable to lock mailertable file";
      sleep (VD_TIME_WAIT);
    }

  /* open mailertable */
  if (vsd_load_file (&mailertable_line, domain_root_path, vd_mailertable_file (NULL)))
    {
      unlink (mailertable_lock);
      return "unable to load the mailertable file";
    }

  /* open temporary mailertable file */
  sprintf (mailertable_tmp, "%s.tmp", vd_mailertable_path (domain_root_path));
  if ((stream_tmp = fopen (mailertable_tmp, "w")) == NULL)
    {
      unlink (mailertable_lock);
      return "unable to create temporary mailertable file";
    }

  /* Set up a string containing the domain name followed by a space, so that */
  /* only the domain name that we are looking for will match. Probably a */
  /* better way of doing it, but my head has stopped working. */
  sprintf (domain_name_appended_space, "%s ", domain_name);

  /* Loop through mailtertable */
  for (line = mailertable_line; line; line = line->next)
    {
      if ((compare = strncmp (domain_name_appended_space, line->line, 
                              strlen (domain_name_appended_space))) != 0)
        fprintf (stream_tmp, "%s\n", line->line);
    }

  /* rename mailertable.tmp to mailertable */
  if (rename (mailertable_tmp, vd_mailertable_file (domain_root_path)) == -1)
    {
      unlink (mailertable_lock);
      return "unable to rename the temporary mailertable file";
    }

  if (unlink (mailertable_lock) == -1)
    return "unable to unlock the mailertable file"; 
  
  return NULL;
}

/********************************************************************************
 * Description: Given the procmailrc template file, generates a domain specific
 *              version and appends the domain name to the DOMAIN variable.
 *
 *     Returns: NULL on success
 *              or a string describing the error
 *******************************************************************************/
static char *
vd_generate_domain_procmailrc_file (const char *domain_name,
                                    const char *virtual_server_root)
{

  FILE *stream;
  struct vsd_file_line *procmailrc_file_line = NULL;
  struct vsd_file_line *line;
  char *contents;

  /* Open the procmailrc template file */
  if (vsd_load_file (&procmailrc_file_line, virtual_server_root, vd_procmailrc_domain_template_file (NULL)))
    return "unable to open the domain recipe template file";

  /* Create the new domain specific procmailrc file */
  if ((stream = fopen (vd_procmailrc_domain_recipe_file (virtual_server_root, domain_name), "w+")) == NULL)
    return "unable to open the domain recipe file";

  fprintf (stream, "DOMAIN=%s\n", domain_name);

  for (line = procmailrc_file_line; line; line = line->next)
    {
      /* DAMION: The parsing of the file was made redundant */
      /* if the line in the template equals DOMAIN=, add the domain name and print */
      /* it to the new file */
      /* if (strcmp (line->line, "DOMAIN=") == 0) */
      /*  fprintf (stream, "DOMAIN=%s\n", domain_name); */
      /* else, just print the line to the new file */
      /* else */
        fprintf (stream, "%s\n", line->line);
    }
  
  vsd_free_file (procmailrc_file_line);
  fclose (stream);

  return NULL;
}

/********************************************************************************
 * Description: Creates a file called admin.rc in the directory given as 
 *              the argument. File format is presumed as:
 *                
 *              :0 c # _vuser admin
 *              {
 * 
 *              }
 *
 *    Returns: NULL on success or a string describing the error
 *******************************************************************************/
static char *
vd_generate_admin_recipe (const char *domain_root_path,
                          const char *domain_name) 
{

  FILE *template_stream, *recipe_stream;
  char *template_file, *recipe_file;
  int ch;

  if ((template_stream = fopen (vd_procmailrc_admin_template_file (domain_root_path), "r")) == NULL)
      return "unable to open the admin recipe template file";

  if ((recipe_stream = fopen (vd_procmailrc_admin_recipe_file (domain_root_path, domain_name), "w")) == NULL)
      return "unable to open the admin recipe file";

  while ((ch = fgetc (template_stream)) != EOF)
   {
      fputc(ch, recipe_stream);
   }
   
   fclose(template_stream);
   
   if (fclose (recipe_stream) != 0)
      return "unable to close the admin recipe file";

   return NULL;
}

/********************************************************************************
 * Description: Carries out the physical configuration steps associated with
 *              creating a managed domain.
 *
 * Returns: NULL on success, or a string describing the error otherwise.
 *
 *******************************************************************************/
static char *
vd_config_domain (struct connection *vc,
                  struct vsd_vs *domain)
{
  const char procmailrc_domain_file[] = "pm.rc";
  const char procmailrc_template_dir[] = "/etc/procmailrcs";
  const char procmailrc_template_file[] = "/etc/procmailrcs/pm-vsd-vd.rc";
  char *status;
  char domain_root_path[VD_STR_MAX];
  char mailertable_file[VD_STR_MAX];
  char chown_file[VD_STR_MAX];
  char passwd_file[VD_STR_MAX];
  char procmailrc_template_full_path[VD_STR_MAX];
  char procmailrc_admin_mesg_dir[VD_STR_MAX];
  char *procmailrc_index_file;
  FILE *procmailrc_index_stream;
  
  /* Set the path to the root dir of the domian */
  strcpy (domain_root_path, vsd_map_server_root (vc->map, domain->hs));

  /* FIXME: Create the /etc/procmailrcs/virtual directory, need to sdort the hard coded path */
  if (vd_mkdir (domain_root_path, "/etc/procmailrcs/virtual") != NULL)
      return "unable to create the procmailrcs/virtual directory"; 

  if (vd_chown_run_command (vc, domain->hs, "admin.admin", "/etc/procmailrcs/virtual"))
      return "unable to chown the procmailrcs/virtual direcotry";

  /* Create /etc/procmailrcs/virtual/<domain> directory */
  if (vd_mkdir (NULL, vd_procmailrc_path (domain_root_path, domain->name)) != NULL)
      return "unable to create procmailrcs domain directory"; 

  /* Ensure permissions are admin.admin */
  if (vd_chown_run_command (vc, domain->hs, "mail.admin", vd_procmailrc_path (NULL,  domain->name)))
      return "unable to chown procmailrc directory";

  /* Generate the procmailrc file for the domain */
  if (vd_generate_domain_procmailrc_file (domain->name, domain_root_path) != NULL)
      return "unable to generate the domain specific procmailrc file";

  /* link all of the rest of the procmailrc files to the domain directory */
  if ((status = vd_dir_files_link (vd_procmailrc_template_path (domain_root_path), 
                vd_procmailrc_path (domain_root_path, domain->name))) != NULL)
      return status;

  /* Set its permissions to mail.admin */
  if (vd_chown_run_command (vc, domain->hs, "mail.admin", vd_procmailrc_domain_recipe_file (NULL, domain->name)))
      return "unable to chown the domain specific procmailrc file";

  /* Create a admin.rc in procmailrcs/virtual/<domain> */
  if ((status = vd_generate_admin_recipe (domain_root_path, domain->name)) != NULL)
      return status;

  /* change permissions to mail.admin */
  if (vd_chown_run_command (vc, domain->hs, "mail.admin", vd_procmailrc_admin_recipe_file (NULL, domain->name)))
      return "unable to chown the domain specific procmailrc admin recipe";

  /* Create procmailrcs/virtual/<domain>/admin directory */
  if ((status = vd_mkdir (vd_procmailrc_path (domain_root_path, domain->name), "admin")) != NULL)
      return "unable to create procmailrcs/virtual/<domain>/admin directory";

  /* change permissions to mail.admin */
  if (vd_chown_run_command (vc, domain->hs, "mail.admin", vd_procmailrc_admin_path (NULL, domain->name)))
      return "unable to chown the domain specific procmailrc admin directory";
  
  /* Create procmailrcs/virtual/<domain>/admin/msg directory */
  if ((status = vd_mkdir (domain_root_path, vd_procmailrc_admin_msg_path (NULL, domain->name))) != NULL)
      return "unable to create procmailrcs/virtual/<domain>/admin/msg directory";

  /* change permissions to mail.admin */
  if (vd_chown_run_command (vc, domain->hs, "mail.admin", vd_procmailrc_admin_msg_path (NULL, domain->name)))
      return "unable to chown the domain specific procmailrc admin/msg directory";
  
  /* Create an empty file called .index in procmailrcs/virtual/<domain>/admin/msg dir */
  if ((procmailrc_index_stream = fopen (vd_procmailrc_admin_index_file (domain_root_path, domain->name), "w")) == NULL)
      return "unable to create an empty .index file for domain";
  fprintf (procmailrc_index_stream, "\n");
  fclose (procmailrc_index_stream);

  /* change permissions to mail.admin */
  if (vd_chown_run_command (vc, domain->hs, "mail.admin", vd_procmailrc_admin_index_file (NULL, domain->name)))
      return "unable to chown the domain specific procmailrc admin/msg/.index file";
  
  /* FIXME: create the /var/spool/virtual directory, need to sort the hard coded path */
  if (vd_mkdir (domain_root_path, "/var/spool/virtual") != NULL)
      return "unable to create the var/spool/virtual directory";
  if (vd_chown_run_command (vc, domain->hs, "admin.admin", "/var/spool/virtual"))
      return "unable to chown var/spool/virtual directory";
 
  /* create the /var/spool/virtual/<domain> directory */
  if (vd_mkdir (NULL, vd_spool_path (domain_root_path, domain->name)) != NULL)
      return "unable to create /var/spool/virtual/<domain> directory";

  /* Ensure permissions are mail.admin */
  if (vd_chown_run_command (vc, domain->hs, "mail.admin",
                            vd_spool_path (NULL, domain->name)))
      return "unable to chown /var/spool/virtual/<domain> directory";

  /* FIXME: create the /etc/virtual directory, need to sort the hardcoded path */
  if (vd_mkdir (domain_root_path, "/etc/virtual") != NULL) 
      return "unable to create the /etc/virtual directory";
  if  (vd_chown_run_command (vc, domain->hs, "admin.admin", "/etc/virtual"))
      return "unable to chown the /etc/virtual directory";

  /* create /etc/virtual/<domain> passwd file, and populate with admin & random password */
  if (vd_mkdir (NULL, vd_passwd_path (domain_root_path, domain->name)) != NULL)
      return "unable to create the domain's passwd directory";

  if (vd_passwd_file_create ("admin", domain->name, vd_passwd_path (domain_root_path, domain->name)))
      return "unable to create the domain's passwd file";
 
  /* Set the permisions for the domain dir & the passwd file */
  if (vd_chown_run_command (vc, domain->hs, "mail.admin", vd_passwd_path (NULL, domain->name)))
      return "unable to chown the /etc/virtual/<domain> directory";

  if (vd_chown_run_command (vc, domain->hs, "mail.admin", vd_passwd_file (NULL, domain->name)))
      return "unable to chown the domain specific passwd file";

  /* Append the domain to the mailertable file */
  if (vd_append_domain_mailertable (domain_root_path, domain->name, domain->name) == -1)
      return "unable to append the domain to the mailertable file";

  /* ensure that the permissions are set to mail.admin */
  if (vd_chown_run_command (vc, domain->hs, "admin.admin", vd_mailertable_file (NULL)))
      return "unable to chown the mailertable file";

  /* & regenerate /etc/mail/mailtertable.db */
  if (vd_refresh_mailertable (vc, domain_root_path, vd_mailertable_database_file (NULL),
                               vd_mailertable_file (NULL)) != 0)
      return "unable to regenerate the mailertable database"; 
  
  return NULL;
}

/********************************************************************************
 * Description: Carries out the removal of the physical configuration associated 
 *              with a managed domain.
 *
 * Returns: NULL on success, or a string describing the error otherwise.
 *
 *******************************************************************************/
static char *
vd_remove_domain  (struct connection *vc,
                   struct vsd_vs *domain)
{
  
  const char procmailrc_domain_file[] = "pm.rc";
  const char procmailrc_template_file[] = "/etc/procmailrcs/pm-vsd-vd.rc";
  char domain_root_path[VD_STR_MAX];
  char *status;

  /* Obtain the root path for the domain */
  strcpy (domain_root_path, vsd_map_server_root (vc->map, domain->hs));

  /* Remove the line/lines containing <domain> from etc/mail/mailertable */
  if ((status = vd_remove_domain_mailertable (domain_root_path, domain->name)) != NULL)
    return status;

  /* ensure that perms are how we found them */
  if (vd_chown_run_command (vc, domain->hs, "mail.admin", vd_mailertable_file (NULL)))
      return "unable to chown the mailertable file";

  /* Regenerate etc/mail/mailtertable.db */
  if (vd_refresh_mailertable (vc, domain_root_path, vd_mailertable_database_file (NULL),
                               vd_mailertable_file (NULL)) != 0)
    return "unable to regenerate the mailertable database";

  /* Remove all files in var/spool/<domain> */
  if ((status = vd_dir_files_unlink (vd_spool_path (domain_root_path, domain->name))) != NULL)
    return status;

  /* Remove the var/spool/virtual/<domain> directory */
  if (rmdir (vd_spool_path (domain_root_path, domain->name)) == -1)
    return "unable to remove the domain's spool directory";

  /* Remove the etc/virtual/<domain>/passwd file */
  if (unlink (vd_passwd_file (domain_root_path, domain->name)) == -1)
    return "unable to remove the domain's passwd file";

  /* Remove the etc/virtual/<domain> directory */
  if (rmdir (vd_passwd_path (domain_root_path, domain->name)) == -1)
    return "unable to remove the domain's passwd directory";

  /* Remove all files under etc/procmailrcs/virtual/<domain>/admin/msg */
  /* Remove etc/procmailrcs/virtual/<domain>/admin/msg directory */
  /* Remove all files under etc/procmailrcs/virtual/<domain>/admin */
  /* Remove the etc/procmailrcs/virtual/<domain>/<procmail rc> files */
  if ((status = vd_dir_files_unlink (vd_procmailrc_path (domain_root_path, domain->name))) != NULL)
    return status; 

  /* Remove the etc/procmailrcs/virtual/<domain> directory */
  if (rmdir (vd_procmailrc_path (domain_root_path, domain->name)) == -1)
    return "unable to remove the domain's procmailrc directory";

  return NULL;
}


/********************************************************************************
 * Description: Carries out the physical configuration steps associated with
 *              creating a managed domain.
 *
 * Returns: NULL on success, or a string describing the error otherwise.
 *
 *******************************************************************************/
static char *
vd_config_domain_alias (struct connection *vc,
                        struct vsd_vs *domain,
                        const char *domain_alias)
{

  const char procmailrc_domain_file[] = "pm.rc";
  char domain_root_path[VD_STR_MAX];

  strcpy (domain_root_path, vsd_map_server_root (vc->map, domain->hs));

  /* append alias to mailertable */
  if (vd_append_domain_mailertable (domain_root_path, domain_alias, domain->name) == -1)
      return "unable to append domain alias to the mailertable file";

  /* ensure that the permissions are set to mail.admin */
  if (vd_chown_run_command (vc, domain->hs, "mail.admin", vd_mailertable_file (NULL)))
      return "unable to chown the mailertable file";

  /* regenerate mailertable */
  if (vd_refresh_mailertable (vc, domain_root_path, vd_mailertable_database_file (NULL),
                               vd_mailertable_file (NULL)) != 0)
      return "unable to regenerate the mailertable database"; 

  return NULL;
}

/********************************************************************************
 * Description: Carries out the removal of the physical configuration associated
 *              with a managed domain.
 *
 * Returns: NULL on success, or a string describing the error otherwise.
 *
 *******************************************************************************/
static char *
vd_remove_domain_alias (struct connection *vc,
                        struct vsd_vs *domain,
                        const char *domain_alias)
{

  const char procmailrc_domain_file[] = "pm.rc";
  char domain_root_path[VD_STR_MAX];
  char *status;
  
  strcpy (domain_root_path, vsd_map_server_root (vc->map, domain->hs));

  /* remove the alias from mailertable */
  if ((status = vd_remove_domain_mailertable (domain_root_path, domain_alias)) != NULL)
    return status;

  /* ensure that permissions are left as we found them: */
  if (vd_chown_run_command (vc, domain->hs, "mail.admin", vd_mailertable_file (NULL)))
      return "unable to chown the mailertable file";

  /* regenerate the mailertable */
  if (vd_refresh_mailertable (vc, domain_root_path, vd_mailertable_database_file (NULL),
                               vd_mailertable_file (NULL)) != 0)

    return "unable to regenerate the mailertable database";

  return NULL;
}

/********************************************************************************
 * Description: Links all files from dir_from to dir_to
 *
 * Returns: NULL on success, or a string describing the error otherwise.
 *
 *******************************************************************************/
static char *
vd_dir_files_link (const char *dir_from, 
                   const char *dir_to)
{
  struct dirent *from_dp;
  struct stat stbuf;
  DIR *from_dfd;
  char link_from[VD_STR_MAX];
  char link_to[VD_STR_MAX];

  if ((from_dfd = opendir (dir_from)) == NULL)
      return "unable to open the directory linking files from";

  while ((from_dp = readdir (from_dfd)) != NULL)
    {
      /* Skip parent & self */
      if ((strcmp (from_dp->d_name, ".") == 0) || (strcmp (from_dp->d_name, "..") == 0))
          continue;
      else if (((strlen (dir_from) + strlen (from_dp->d_name) + 2) > sizeof (link_from)) &&
               ((strlen (dir_to) + strlen (from_dp->d_name) + 2) > sizeof (link_to)))
          return "directory string too long";
      else
        {
          sprintf (link_from, "%s/%s", dir_from, from_dp->d_name);
          sprintf (link_to, "%s/%s", dir_to, from_dp->d_name);
        }

      if ((stat (link_from, &stbuf) == -1) || ((stbuf.st_mode & S_IFMT) == S_IFDIR))
      /* Either failed to stat file, or its a directory so skip */
          continue;

      /* create hardlink to files */
      if ((link (link_from, link_to) != 0) && (errno != EEXIST))
          return "unable to link file";
    }
  return NULL;
}

/********************************************************************************
 * Description: unlinks all files from dir
 *
 * Returns: NULL on success, or a string describing the error otherwise.
 *
 *******************************************************************************/
static char *
vd_dir_files_unlink (const char *dir)
{

  struct dirent *dp;
  struct stat stbuf;
  DIR *dfd;
  char unlink_file[VD_STR_MAX];
  char *status;

  if ((dfd = opendir (dir)) == NULL)
      return "unable to open the directory";

  while ((dp = readdir (dfd)) != NULL)
    {
      /* skip parent and self: */
      if ((strcmp (dp->d_name, ".") == 0) || (strcmp (dp->d_name, "..") == 0))
          continue;
      else if ((strlen (dir) + strlen (dp->d_name) + 2) > sizeof (unlink_file))
          return "unable to unlink, file name too long";
      else 
          sprintf (unlink_file, "%s/%s", dir, dp->d_name);

      if ((stat (unlink_file, &stbuf) == -1) || ((stbuf.st_mode & S_IFMT) == S_IFDIR))
      {
          if ((status = vd_dir_files_unlink (unlink_file)) != NULL)
              return status;
          if (rmdir (unlink_file) == -1)
              return strerror (errno);
          continue;
      }
      if (unlink (unlink_file) != 0) 
          return "unable to unlink a file";
    }
  return NULL;
}

/********************************************************************************
 * Description: Checks the number of virtual domains defined for a virtual 
 *              server.
 *
 * Returns: NULL on success, and a string describing error otherwise
 *
 *******************************************************************************/
static char *
vd_check_domain_limit (struct vsd_vs_map *map,
                      struct vsd_vs *virtual_server,
                      const char *root_path)
{
  DIR *dfd;
  struct dirent *dir_ptr;
  int x = 0;
  int registered_domain_number = 0;
  int directory_domain_number = 0;
  
  /* Cycle through vd list, incrementing counter each time one is registered to the vs*/
  for (x = 0; x < map->domains; x++)
    {
      if (strcmp (map->domain[x].hs, virtual_server->name) == 0)
        registered_domain_number++;
    }

  
  /* Count the number of directories in /etc/procmailrcs/virtual */
  if (registered_domain_number != 0)
    {
      if ((dfd = opendir (vd_procmailrc_path (root_path, ""))) == NULL)
        return "unable to open the procmail recipe directory";
    
      while ((dir_ptr = readdir (dfd)) != NULL)
        {
          /* skip self and parent */
          if ((strcmp (dir_ptr->d_name, ".") == 0) || (strcmp (dir_ptr->d_name, "..") == 0))
            {
              continue;
            }
          else if (vd_valid_domainname (dir_ptr->d_name) == NULL)
            {
              /* Presume it is a domain directory */
              directory_domain_number++;
            }
        }
    }
    
  if (directory_domain_number != registered_domain_number)
    return "there appears to be unregistered virtual domains";
  else if (registered_domain_number >= virtual_server->maxvd)
    return "maximum number of virtual domains already reached";
  else
    return NULL;
}
