/* vuser.c - vuser module for virtual user manipulation.
 */

/* 
 * VSD - Virtual Server Daemon
 * Copyright (c) 2000 - 2001 Idaya Ltd.
 *
 * VSD is free software; you can redistribute it and/or 
 * modify it under the terms of the GNU General Public License 
 * as published by the Free Software Foundation; either 
 * version 2 of the License, or (at your option) any later version.
 *
 * VSD is distributed in the hope that it will be useful, 
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU 
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License 
 * along with VSD - see the COPYING file;  if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330, 
 * Boston, MA 02111-1307, USA.  
 */

/*
 * See the AUTHORS file for a list of contributors to VSD. 
 * See the ChangeLog files for a list of changes. 
 */

#include "config.h"

#include <sys/stat.h>
#include <sys/types.h>
#include <fcntl.h>
#include <unistd.h>
#include <errno.h>
#include <sys/dir.h>

#include "vsd.h"
#include "libvsd.h"
#include "vuser.h"

#define VUSER_NAME_LENGTH_MAX   32
#define VUSER_NAME_LENGTH_MIN   1
#define VUSER_DOMAIN_LENGTH_MAX 64
#define VUSER_DOMAIN_LENGTH_MIN 5

#define VSD_OK  0
#define VSD_ERR -1

/******************************************************************************
 * Description: Structures and functions for vuser singly linked list.
 *****************************************************************************/
struct vuser_cnf
{
  char *name;

  struct vsd_file_line *line;

  struct vuser_cnf *rsp_cnf;
  struct vuser_cnf *fwd_cnf;

  struct vuser_cnf *prev;
  struct vuser_cnf *next;
};

static struct vuser_cnf * 
vuser_list_free (struct vuser_cnf  *list)
{ 
  struct vuser_cnf *tmp;

  while (list)
    {
      tmp = list->next;

      free (list->name);
      free (list);

      list = tmp;
    }
  
  return list;
}

static struct vuser_cnf * 
vuser_list_last (struct vuser_cnf  *list)
{ 
  if (list)
    {
      while (list->next)
	list = list->next;
    }

  return list;
}

static struct vuser_cnf * 
vuser_list_append (struct vuser_cnf     *list,
		   struct vsd_file_line *line,
		   char                 *name)
{ 
  struct vuser_cnf *new;
  struct vuser_cnf *last;

  new = (struct vuser_cnf*) malloc (sizeof (struct vuser_cnf)); 
  memset (new, 0, sizeof (struct vuser_cnf));

  new->line = line;
  new->name = (char *) strdup (name);

  if (list)
    {
      last = vuser_list_last (list);

      last->next = new;
      new->prev = last;

      return list;
    }
  else
    return new;
}

static struct vuser_cnf * 
vuser_list_find (struct vuser_cnf  *list,
		 char              *name)
{ 
  struct vuser_cnf *tmp;

  for (tmp = list; tmp; tmp = tmp->next)
    if (!strcmp (name, tmp->name))
      break;

  return tmp;
}

/******************************************************************************
 * Description: Structure and functions for placeholder replacement.
 *****************************************************************************/

typedef enum
{
  VUSER_PH_VSD_VD_NAME,
  VUSER_PH_VSD_VUSER_NAME,
  VUSER_PH_VSD_VUSER_FORWARDER_ADDRESS,
  VUSER_PH_VSD_VUSER_RESPONDER_MESSAGE,
  VUSER_PH_UNKNOWN

} vuser_enum_placeholders;

struct vuser_placeholder 
{
  vuser_enum_placeholders id;
  char *placeholder;
  char *replacement;
};

static struct vuser_placeholder g_placeholders[] = 
{ 
  {
    VUSER_PH_VSD_VD_NAME,
    "$VSD_VD_NAME",
    NULL
  },
  {
    VUSER_PH_VSD_VUSER_NAME,
    "$VSD_VUSER_NAME",
    NULL
  },
  {
    VUSER_PH_VSD_VUSER_FORWARDER_ADDRESS,
    "$VSD_VUSER_FORWARDER_ADDRESS",
    NULL
  },
  {
    VUSER_PH_VSD_VUSER_RESPONDER_MESSAGE,
    "$VSD_VUSER_RESPONDER_MESSAGE",
    NULL
  },
  { VUSER_PH_UNKNOWN,
    NULL,
    NULL
  }
};

void vuser_g_placeholders_init( void )
{
  int ph;
  for (ph = 0; ph < VUSER_PH_UNKNOWN; ph++)
    g_placeholders[ph].replacement = NULL;
}

/******************************************************************************
 * Description: validate a vuser name.
 *
 *     Returns: NULL on success.
 *              char * pointer to static string describing error.
 *****************************************************************************/
static char *
vuser_name_validate (char *name)
{ 
  char error[256];
  int rc = VSD_OK;
  
  if (strlen (name) > VUSER_NAME_LENGTH_MAX
      || strlen (name) < VUSER_NAME_LENGTH_MIN)
      rc = VSD_ERR;
  
  return (rc ? error : NULL);
}

/******************************************************************************
 * Description: validate a vuser domain.
 *
 *     Returns: NULL on success.
 *              char * pointer to static string describing error.
 *****************************************************************************/
static char *
vuser_domain_validate (char *domain)
{ 
  char error[256];
  int rc = VSD_OK;
  
  if (strlen (domain) > VUSER_DOMAIN_LENGTH_MAX
      || strlen (domain) < VUSER_DOMAIN_LENGTH_MIN)
      rc = VSD_ERR;
  
  return (rc ? error : NULL);
}

/******************************************************************************
 * Description: vuser's domain recipe file path.
 *
 *     Returns: char * pointer to domain recipe file path.
 *              NULL on error.
 *****************************************************************************/
static char * 
vuser_domain_rc (char *root, 
		  char *domain)
{ 
  static char path[256];
  
  sprintf (path, "%setc/procmailrcs/virtual/%s/pm.rc", 
	   (root ? root : "/"),
	   domain);
  
  return path;
}

/******************************************************************************
 * Description: vuser's passwd file path.
 *
 *     Returns: char * pointer to passwd file path.
 *              NULL on error.
 *****************************************************************************/
static char * 
vuser_passwd (char *root, 
	      char *domain)
{ 
  static char path[256];
  
  sprintf (path, "%s/etc/virtual/%s/passwd", 
	   (root ? root : ""),
	   domain);
  
  return path;
}

/******************************************************************************
 * Description: vuser's mbx file path.
 *
 *     Returns: char * pointer to mbx file path.
 *              NULL on error.
 *****************************************************************************/
static char * 
vuser_mbx (char *root,
	   char *domain,
	   char *name)
{ 
  static char path[256];
  
  sprintf (path, "%s/var/spool/virtual/%s/%s", 
	   (root ? root : ""),
	   domain, name);
  
  return path;
}

/******************************************************************************
 * Description: vuser's domain directory path.
 *
 *     Returns: char * pointer to vuser domain directory path.
 *              NULL on error.
 *****************************************************************************/
static char * 
vuser_domain_dir (char *root,
		  char *domain)
{ 
  static char path[256];
  
  sprintf (path, "%s/etc/procmailrcs/virtual/%s", 
	   (root ? root : ""),
	   domain);
  
  return path;
}

/******************************************************************************
 * Description: vuser's template vuser recipe path.
 *
 *     Returns: char * pointer to template vuser recipe path.
 *              NULL on error.
 *****************************************************************************/
static char * 
vuser_template_vuser_rc (char *root,
			 char *domain)
{ 
  static char path[256];
  
  sprintf (path, "%s/pm-vsd-vuser.rc", 
	   vuser_domain_dir (root, domain));
  
  return path;
}

/******************************************************************************
 * Description: vuser's template dflt-int recipe path.
 *
 *     Returns: char * pointer to template dflt-int recipe path.
 *              NULL on error.
 *****************************************************************************/
static char * 
vuser_template_dflt_int_rc (char *root,
			    char *domain)
{ 
  static char path[256];
  
  sprintf (path, "%s/pm-vsd-dflt-int.rc", 
	   vuser_domain_dir (root, domain));
  
  return path;
}

/******************************************************************************
 * Description: vuser's template dflt-ext recipe path.
 *
 *     Returns: char * pointer to template dflt-ext recipe path.
 *              NULL on error.
 *****************************************************************************/
static char * 
vuser_template_dflt_ext_rc (char *root,
			    char *domain)
{ 
  static char path[256];
  
  sprintf (path, "%s/pm-vsd-dflt-ext.rc", 
	   vuser_domain_dir (root, domain));
  
  return path;
}

/******************************************************************************
 * Description: vuser's template fwd-int recipe path.
 *
 *     Returns: char * pointer to template fwd-int recipe path.
 *              NULL on error.
 *****************************************************************************/
static char * 
vuser_template_fwd_int_rc (char *root,
			   char *domain)
{ 
  static char path[256];
  
  sprintf (path, "%s/pm-vsd-fwd-int.rc", 
	   vuser_domain_dir (root, domain));
  
  return path;
}

/******************************************************************************
 * Description: vuser's template fwd-ext recipe path.
 *
 *     Returns: char * pointer to template fwd-ext recipe path.
 *              NULL on error.
 *****************************************************************************/
static char * 
vuser_template_fwd_ext_rc (char *root,
			   char *domain)
{ 
  static char path[256];
  
  sprintf (path, "%s/pm-vsd-fwd-ext.rc", 
	   vuser_domain_dir (root, domain));
  
  return path;
}

/******************************************************************************
 * Description: vuser's template rsp recipe path.
 *
 *     Returns: char * pointer to template rsp recipe path.
 *              NULL on error.
 *****************************************************************************/
static char * 
vuser_template_rsp_rc (char *root,
		       char *domain)
{ 
  static char path[256];
  
  sprintf (path, "%s/pm-vsd-rsp.rc", 
	   vuser_domain_dir (root, domain));
  
  return path;
}

/******************************************************************************
 * Description: vuser's recipe path.
 *
 *     Returns: char * pointer to vuser recipe path.
 *              NULL on error.
 *****************************************************************************/
static char * 
vuser_vuser_rc (char *root,
		char *domain,
		char *name)
{ 
  static char path[256];
  
  sprintf (path, "%s/%s.rc", 
	   vuser_domain_dir (root, domain),
	   name);
  
  return path;
}

/******************************************************************************
 * Description: vuser's directory path.
 *
 *     Returns: char * pointer to vuser directory path.
 *              NULL on error.
 *****************************************************************************/
static char * 
vuser_vuser_dir (char *root,
		 char *domain,
		 char *name)
{ 
  static char path[256];
  
  sprintf (path, "%s/%s", 
	   vuser_domain_dir (root, domain),
	   name);
  
  return path;
}

/******************************************************************************
 * Description: vuser's msg directory path.
 *
 *     Returns: char * pointer to vuser msg directory path.
 *              NULL on error.
 *****************************************************************************/
static char * 
vuser_msg_dir (char *root,
	       char *domain,
	       char *name)
{ 
  static char path[256];
  
  sprintf (path, "%s/msg", 
	   vuser_vuser_dir (root, domain, name));
  
  return path;
}

/******************************************************************************
 * Description: vuser's msg index file path.
 *
 *     Returns: char * pointer to msg index file path.
 *              NULL on error.
 *****************************************************************************/
static char * 
vuser_msgs (char *root,
	    char *domain,
	    char *name)
{ 
  static char path[256];
  
  sprintf (path, "%s/.index", 
	   vuser_msg_dir (root, domain, name));
  
  return path;
}

/******************************************************************************
 * Description: vuser's msg file path.
 *
 *     Returns: char * pointer to msg file path.
 *              NULL on error.
 *****************************************************************************/
static char * 
vuser_msg (char *root,
	   char *domain,
	   char *name,
	   char *msg)
{ 
  static char path[256];
  
  sprintf (path, "%s/%s", 
	   vuser_msg_dir (root, domain, name), msg);
  
  return path;
}

/******************************************************************************
 * Description: vuser's msg cache file path.
 *
 *     Returns: char * pointer to msg cache  path.
 *              NULL on error.
 *****************************************************************************/
static char * 
vuser_msg_cache (char *root,
		 char *domain,
		 char *name,
		 char *msg)
{ 
  static char path[256];
  
  sprintf (path, "%s/%s.cache", 
	   vuser_msg_dir (root, domain, name), msg);
  
  return path;
}

/******************************************************************************
 * Description: find passwd entry.
 *
 *     Returns: struct vsd_file_line * pointer to found entry.
 *              NULL on error.
 *****************************************************************************/
static struct vsd_file_line * 
vuser_passwd_find (struct vsd_file_line *fpasswd, 
		   char                 *name)
{ 
  char *pch, *pbuf;
  char *buf = NULL;
  struct vsd_file_line *line;

  for (line = fpasswd; line; line=line->next)
    {
      /* Match on name. */
      if ((buf = (char *) strdup (line->line)) 
	  && (pch = (char *) strchr (buf, ':')))	
	{
	  *pch = '\0';
	  
	  /* Handle disabled vusers. */
	  pbuf = buf;
	  if (buf[0] == '!')
	    pbuf++;

	  if (!strcmp (pbuf, name))
	    {
	      free (buf);
	      break;
	    }
	}
      free (buf);
    }
  
  return line;
}

/******************************************************************************
 * Description: find recipe file entry.
 *
 *     Returns: struct vsd_file_line * pointer to found entry.
 *              NULL on error.
 *****************************************************************************/
static struct vsd_file_line * 
vuser_rc_find (struct vsd_file_line *frc, 
       	       char                 *entry)
{ 
  struct vsd_file_line *line;

  for (line = frc; line; line=line->next)
    if (!strcmp (line->line, entry))
	break;

  return line;
}

/******************************************************************************
 * Description: find msgs entry.
 *
 *     Returns: struct vsd_file_line * pointer to found entry.
 *              NULL on error.
 *****************************************************************************/
static struct vsd_file_line * 
vuser_msgs_find (struct vsd_file_line *fmsgs, 
		 const char           *msg)
{ 
  struct vsd_file_line *line;
  
  for (line = fmsgs; line; line=line->next)
    if (!strcmp (line->line, msg))
      break;
  
  return line;
}

/******************************************************************************
 * Description: parse recipe file.
 *
 *     Returns: char * pointer to linked list of recipe configuration.
 *              NULL on error.
 *****************************************************************************/
static int 
vuser_rc_parse (struct vsd_file_line *frc,
		struct vuser_cnf     **cnf)
{ 
  char *pch, type;
  char buf[256];
  struct vsd_file_line *line;
  struct vuser_cnf *last;
  int rc = VSD_OK;
  int level, i;
  
  /****************************************************************************
   * Format of recipe file assumed to be as follows:
   * 

   # Comments...
   
   :0: # _vuser usrone
   {
     :0c: # _rsp ???
     | vacaction $LOGNAME
   
     :0c: # _fwd-ext someone@somewhere
     ! somebody@somewhere
   
     :0c: # _fwd-int localvuser
     | formail -ds >> $MAILDIR/localvuser
   }
      
  ****************************************************************************/
  
  for (line = frc; !rc && line; line=line->next)
    {
      pch = line->line;

      /* Trip on underscore character and nesting. */
      while ( *pch 
	      && *pch != '_'
	      && *pch != '{'
	      && *pch != '}')
	pch++;

      /* Handle nesting. */
      if (*pch == '{')
	level++;
      if (*pch == '}')
	level--;
      
      /* Trap out underscores within names */
      if (isspace (*((char *) pch + 1)))
	continue;
      if (! isspace(*((char *) pch - 1)))
	continue;

      /* Establish type */
      pch++;
      type = *pch;

      /* Skip over token */
      while (*pch && ! isspace (*pch))
	pch++;
      while (*pch && isspace (*pch))
	pch++;
      
      /* Extract identifier */
      i = 0;
      while (*pch 
	     && *pch != '\n' 
	     && *pch != '\r' 
	     && ! isspace (*pch)
	     && i < sizeof (buf))
	buf[i++] = *pch++;
      buf[i] = '\0';

      switch (type)
	{
	  /* _vuser */
	case 'v': 
	  *cnf = vuser_list_append (*cnf, line, buf);

	  break;

	  /* _fwd_cnf */
	case 'f': 
	  last = vuser_list_last (*cnf);
	  last->fwd_cnf = vuser_list_append (last->fwd_cnf, line, buf);

	  break;

	  /* _rsp_cnf */
	case 'r': 
	  last = vuser_list_last (*cnf);
	  last->rsp_cnf = vuser_list_append (last->rsp_cnf, line, buf);
	  
	  break;
	}      
    }

  return rc;
}

/******************************************************************************
 * Description: parse recipe template and replace placeholders
 *
 *     Returns: 0 on success
 *             -1 on error.
 *****************************************************************************/
static int 
vuser_template_parse (struct vsd_file_line      *frc)
{ 
  char *pch, *buf;
  struct vsd_file_line *line;
  int rc = VSD_OK;
  int ph;

  for (line = frc; !rc && line; line=line->next)
    {
      /* Trip on recognised place-holders */
      for (ph = 0; ph < VUSER_PH_UNKNOWN; ph++)
	{
	  if (g_placeholders[ph].replacement 
	      && (pch = (char *) strstr (line->line, g_placeholders[ph].placeholder)))
	    {
	      if (buf = (char *) malloc (strlen (line->line) 
					 - strlen (g_placeholders[ph].placeholder)
					 + strlen (g_placeholders[ph].replacement)
					 + 1))
		{
		  *pch = '\0';
		  pch = pch + strlen (g_placeholders[ph].placeholder);
		  
		  strcpy (buf, line->line);
    		  strcat (buf, g_placeholders[ph].replacement);
		  strcat (buf, pch); 

		  free (line->line);
		  line->line = (char *) buf;
		}
	      else
		rc = VSD_ERR;

	    }
	}
    }

  return rc;
}

/******************************************************************************
 * Description: Sets up the appropriate args for chown, and passes them to
 *              the run_command executable.
 *
 * Sorry Tim for putting this in here, quickest way to do it? Damion.
 *
 *     Returns: 0 on success
 *              -1 on error
 *****************************************************************************/
int
vuser_chown_run_command (struct connection *vc,
                         char *user,
                         char *file_to_change)
{
  char args[256];
  int run_command_ret = 0;

  sprintf (args, "%s %s", user, file_to_change);
  run_command_ret = run_command (vc, vc->server_root, "root", "/bin/chown", args, 
                                 NULL, NULL);

  return run_command_ret;
}

/******************************************************************************
 * Description: Check that the number of vusers hasn't been exceeded.
 *
 *     Returns: Null on success, string describing error otherwise.
 *****************************************************************************/
static char *
vuser_check_vuser_limit (struct vsd_vs *domain,
                         char *server_root)
{

  DIR *dir_fd;
  struct dirent *dir_ptr;
  struct stat statbuf;
  char *full_name;
  int num_vuser = 0;
  int rc = VSD_OK;
  char *err = NULL;

  if (domain == NULL)
    {
      rc = VSD_ERR;
      err = (char *) strdup ("invalid domain");
    }

  if (server_root == NULL)
    {
      rc = VSD_ERR;
      err = (char *) strdup ("invalid server root");
    }

  if (!rc && (dir_fd = opendir (vuser_domain_dir (server_root, 
                                                  domain->name))) == NULL)
    {
      rc = VSD_ERR;
      err = (char *) strdup ("unable to open recipe directory");
    }

  while (!rc && (dir_ptr = readdir (dir_fd)) != NULL)
    {
      if ((strcmp (dir_ptr->d_name, ".") == 0) || 
          (strcmp (dir_ptr->d_name, "..") == 0))
        continue;
      
      /* prepend the server_root to the filename */
      if ((full_name = (char *) malloc (strlen (server_root) + 
                                        strlen (vuser_domain_dir (server_root,
                                                  domain->name)) +
                                        strlen (dir_ptr->d_name) + 2)) == NULL)
        {
          rc = VSD_ERR;
          err = (char *) strdup ("memory allocation error");
        }

      sprintf (full_name, "%s/%s", vuser_domain_dir (server_root, domain->name),
                                      dir_ptr->d_name);
      
      if ((stat (full_name, &statbuf) == 0) && 
          ((statbuf.st_mode & S_IFMT) == S_IFDIR))
        num_vuser++;
    }

  if (!rc && (num_vuser >= domain->max_vuser))
    {
      err = (char *) strdup ("maximum number of virtual users reached");
    }

  return err;
}

/******************************************************************************
 * Description: Checks that the number of mailboxes hasn't been exceeded.
 *
 *     Returns: Null on success, string describing error otherwise.
 *****************************************************************************/
static char * 
vuser_check_mbx_limit (struct vsd_vs *domain,
                       char *server_root)
{

  struct vsd_file_line *line, *passwd_file_line;
  int num_mbx = 0;
  int rc = VSD_OK;
  char *err = NULL;

  if (vsd_load_file (&passwd_file_line, server_root, 
                     vuser_passwd (NULL, domain->name)))
    {
      rc = VSD_ERR;
      err = (char *) strdup ("unable to open domain's passwd file");
    }
  
  /* Count the number of lines in the passwd file. One line = One mailbox */
  for (line = passwd_file_line; !rc && line; line = line->next)
    num_mbx++;

  if (!rc && (num_mbx >= domain->max_vmbx))
    {
      rc = VSD_ERR;
      err = (char *) strdup ("maximum number of virtual mailboxes reached");
    }

  return err;
}

/******************************************************************************
 * Description: Checks that the number of messages hasn't been exceeded.
 *
 *     Returns: NULL on success, string describing error otherwise.
 *****************************************************************************/
static char *
vuser_check_msg_limit (struct vsd_vs *domain,
                       char *server_root,
                       char *user)
{

  struct vsd_file_line *line, *index_file_line;
  int num_msg = 0;
  int rc = VSD_OK;
  char *err = NULL;
  char *p;

  if (vsd_load_file (&index_file_line, server_root, 
                      vuser_msgs (NULL, domain->name, user)))
    {
      rc = VSD_ERR;
      err = (char *) strdup ("unable to open virtual user's message index file");
    }

  for (line = index_file_line; !rc && line; line = line->next)
    {
      /* Skip white space.  */
      for (p = line->line; *p && isspace (*p); p++)
        ;

      /* Skip blank lines.  */
      if (*p == '\0')
        continue;

      /* cna be sure that we now have a message entry */
      num_msg++;
    }

  if (!rc && (num_msg >= domain->max_vmsg))
    {
      rc = VSD_ERR;
      err = (char *) strdup ("maximum number of messages reached");
    }

  return err;
}

/******************************************************************************
* Description: Searches a domain's recipe files for references to the given
* 	       mailbox. Returns on the first occurence of the mailbox.
*
* Returns: 0 if no mailboxes are found, 1 otherwise, and -1 on error.
******************************************************************************/
int
vuser_check_mbx_ref (char *domain,
		     char *server_root,
		     char *mailbox)
{
  int rc = VSD_OK;
  struct vsd_file_line *fdomain_rc = NULL;
  struct vsd_file_line *fvuser_rc = NULL;
  struct vuser_cnf *domain_cnf = NULL;
  struct vuser_cnf *vuser_cnf = NULL;
  struct vuser_cnf *fwd_cnf = NULL;
  struct vuser_cnf *cnf = NULL;

  /* Load domain recipe file. */
  if (rc = vsd_load_file (&fdomain_rc, server_root, 
			  vuser_domain_rc (NULL, domain)))
    rc = VSD_ERR;

  /* Parse domain recipe file. */
  if (!rc && (rc = vuser_rc_parse (fdomain_rc, &domain_cnf)))
    rc = VSD_ERR;

  /* Obtain list of vusers for domain, check default forwarder for
   * then user and then parse each file */
  for (cnf =  domain_cnf; !rc && cnf; cnf = cnf->next)
    {
      
      /* Check for reference to mailbox in domain recipe file */
      if (!rc && (fwd_cnf = vuser_list_find (cnf->fwd_cnf, mailbox)))
        return 1;

      /* Load vuser recipe file. */
      if (!rc && (rc = vsd_load_file (&fvuser_rc, server_root, 
				  vuser_vuser_rc (NULL, domain, cnf->name))))
        rc = VSD_ERR;
  
      /* Parse vuser recipe file. */
      if (!rc && (rc = vuser_rc_parse (fvuser_rc, &vuser_cnf)))
        rc = VSD_ERR;

      /* Test for existing forwarder entry. */
      if (!rc && (fwd_cnf = vuser_list_find (vuser_cnf->fwd_cnf, mailbox)))
        return 1;

      /* Free existing vuser_cnf to save the next one being added to the end
       * each time round */
      free (vuser_cnf);
      vuser_cnf = NULL;
    }

  if (rc)
    return -1;
  else
    return 0;
}

/******************************************************************************
 * Description: add virtual user.
 *
 *     Returns: void.
 *****************************************************************************/
void 
vuser_add (struct connection *vc, 
	   const int         argc, 
	   const char        *argv[])
{ 
  char *pch, *buf;
  char *domain, *name, *fwd, *vuser;
  char entry[1024];
  uid_t admin_uid = 0;
  gid_t admin_gid = 0;
  int rc = VSD_OK;
  struct vsd_file_line *fdomain_rc = NULL;
  struct vsd_file_line *fvuser_rc = NULL;
  struct vsd_file_line *ffwd_rc = NULL;
  struct vsd_file_line *domain_line = NULL;
  struct vsd_file_line *vuser_line = NULL;
  struct vsd_file_line *fwd_line = NULL;
  struct vuser_cnf *domain_cnf = NULL;
  struct vuser_cnf *fwd_cnf = NULL;
  struct vuser_cnf *rsp_cnf = NULL;
  struct vuser_cnf *cnf = NULL;
  struct vsd_vs *domain_map;

  if (argc != 1)
    {
      io_buffer_store (vc, "-ERR [vuser] Incorrect arguments\n");
      return;
    }
  
  /* Useful strings. */
  domain = (char *) strdup (vc->virtual_server);
  name = (char *) strdup (argv[0]);

  /* Damion: Removed the attempt to set fwd, because on occasion, argv[1]
   * was not NULL, and was causing VUSER_ADD to fall over, fwd is not used 
   * anywhere here, and more than one arg is an error.
   * fwd = (argv[1] != NULL ? 
   * (char *) strdup (argv[1]) : (char *) strdup (argv[0])); */

  domain_map = vsd_getvdent (vc->map, domain);

  if ((pch = 
       vuser_check_vuser_limit (domain_map,
                                vsd_map_server_root (vc->map, domain_map->hs))) != NULL)
    {
      rc = VSD_ERR;
      io_buffer_store (vc, "-ERR [vuser][limit check] %s", pch);
    } 

  if (vuser = (char *)malloc (strlen (domain) + strlen (name) + 2))
    sprintf (vuser, "%s@%s", name, domain);  

 /* Validate input. */
  if (!rc && (pch = vuser_name_validate (name)))
    {
      rc = VSD_ERR;
      io_buffer_store (vc, "-ERR [vuser][%s] %s", vuser, pch);
    }
  
  if (!rc && (pch = vuser_domain_validate (domain)))
    {
      rc = VSD_ERR;
      io_buffer_store (vc, "-ERR [vuser][%s] %s", vuser, pch);
    }

  /* Load domain recipe file. */
  if (!rc && (rc = vsd_load_file (&fdomain_rc, vc->server_root,
			  vuser_domain_rc (NULL, domain))))
    {
      rc = VSD_ERR;
      io_buffer_store (vc,
		     "-ERR [vuser][%s] error loading domain recipe file: %m",
		     vuser);  
    }
  
  /* Parse domain recipe file. */
  if (!rc && (rc = vuser_rc_parse (fdomain_rc, &domain_cnf)))
    io_buffer_store (vc, 
		     "-ERR [vuser][%s] error parsing domain recipe file\n",
		     vuser);		  
  
  /* Test for existing vuser entry. */
  if (!rc && (vuser_list_find (domain_cnf, name)))
    {
      rc = VSD_ERR;
      io_buffer_store (vc, 
		       "-ERR [vuser][%s] vuser already exists\n",
		       vuser);		  
    }
  
  /* Load vuser recipe template. */
  if (!rc && (rc = vsd_load_file (&fvuser_rc, vc->server_root, 
				  vuser_template_vuser_rc (NULL, domain))))
    io_buffer_store (vc, 
		     "-ERR [vuser][%s] error loading vuser recipe template: %m\n", 
		     vuser);  

  /* Parse vuser recipe template. */
  if (!rc)
    {
      /* FIXME - globals, there must be a better way... */
      vuser_g_placeholders_init ();
      g_placeholders[VUSER_PH_VSD_VUSER_NAME].replacement = name;
      if (rc = vuser_template_parse (fvuser_rc))
	io_buffer_store (vc, 
			 "-ERR [vuser][%s] error parsing vuser recipe template\n",
			 vuser);		  
    }
  
  /* Add vuser entry. */
  if (!rc)
    {
      domain_line = domain_cnf->line->prev;
      
      for (vuser_line = fvuser_rc; !rc && vuser_line; 
	   vuser_line = vuser_line->next, domain_line = domain_line->next)
	  rc = vsd_file_insert_line (domain_line, vuser_line->line);
      
      if (rc)
	io_buffer_store (vc, 
			 "-ERR [vuser][%s] error adding vuser\n",
			 vuser);
    }

  /* Save domain recipe file. */
  if (!rc && (rc = vsd_save_file (fdomain_rc, vc->server_root, vuser_domain_rc (NULL, domain))))
    io_buffer_store (vc, 
		     "-ERR [vuser][%s] error saving domain recipe file: %m\n", 
		     vuser);
  
  /* Create vuser recipe file. */
  if (!rc)  /* FIXME - Another Hack... */
      for (vuser_line = fvuser_rc; !rc && vuser_line; vuser_line = vuser_line->next)
	if (strstr (vuser_line->line, "??") || strstr (vuser_line->line, "INCLUDERC"))
	  {
	    vsd_file_delete_line (fvuser_rc, vuser_line);
	    vuser_line = fvuser_rc;
	  }

  if (!rc && (rc = vsd_save_file (fvuser_rc, vc->server_root, vuser_vuser_rc (NULL, domain, name))))
  io_buffer_store (vc, 
		   "-ERR [vuser][%s] error saving vuser recipe file: %m\n", 
		   vuser);  

  if (!rc)
    {
      /* Create vuser directory. */
      if (!(rc = vsd_get_uidgid (vc->server_root, &admin_uid, NULL, &admin_gid, NULL)))
	if (!(rc = mkdir (vuser_vuser_dir (vc->server_root, domain, name), (mode_t) 0755)))
	  rc = chown (vuser_vuser_dir (vc->server_root, domain, name), admin_uid, admin_gid);
      
      if (rc)
	io_buffer_store (vc, 
			 "-ERR [vuser][%s] error creating vuser directory: %m\n", 
			 vuser);
      
      /* Create vuser msg directory. */
      if (!(rc = mkdir (vuser_msg_dir (vc->server_root, domain, name), (mode_t) 0755)))
	rc = chown (vuser_msg_dir (vc->server_root, domain, name), admin_uid, admin_gid);
      
      if (rc)
	io_buffer_store (vc, 
			 "-ERR [vuser][%s] error creating vuser msg directory: %m\n", 
			 vuser);
      
      /* Create vuser msgs index. */
      {
	struct vsd_file_line fmsgs;
	
	fmsgs.prev = fmsgs.next = NULL;
	fmsgs.line = (char *) strdup ("");
	
	if (!rc && (rc = vsd_save_file (&fmsgs, vc->server_root, vuser_msgs (NULL, domain, name))))
	  io_buffer_store (vc, 
			   "-ERR [vuser][%s] error saving msg index: %m\n", 
			   vuser);
	free (fmsgs.line);
      }
    }

  /* Really sorry.... */
  vuser_chown_run_command (vc, "mail.admin", vuser_vuser_rc (NULL, domain, name));
  vuser_chown_run_command (vc, "mail.admin",  vuser_msgs (NULL, domain, name));

  free (domain);
  free (name);
  /* free (fwd); Damion: fwd is not actually used in this sub */
  free (vuser);

  for (cnf = domain_cnf; cnf; cnf = cnf->next)
    {
      vuser_list_free (cnf->fwd_cnf);
      vuser_list_free (cnf->rsp_cnf);
    }
  vuser_list_free (domain_cnf);

  vsd_free_file (ffwd_rc);    
  vsd_free_file (fvuser_rc);    
  vsd_free_file (fdomain_rc);    
}

/******************************************************************************
 * Description: delete virtual user.
 *
 *     Returns: void.
 *****************************************************************************/
void 
vuser_del (struct connection *vc, 
	   const int         argc, 
	   const char        *argv[])
{
  char *pch, *domain, *name, *vuser;
  int rc = VSD_OK;
  struct vsd_file_line *fdomain_rc = NULL;
  struct vsd_file_line *fvuser_rc = NULL;
  struct vsd_file_line *fmsgs = NULL;
  struct vsd_file_line *msgs_line = NULL;
  struct vsd_file_line *domain_line = NULL;
  struct vsd_file_line *vuser_line = NULL;
  struct vsd_file_line *fwd_line = NULL;
  struct vsd_file_line *tmp_line = NULL;
  struct vuser_cnf *domain_cnf = NULL;
  struct vuser_cnf *fwd_cnf = NULL;
  struct vuser_cnf *rsp_cnf = NULL;
  struct vuser_cnf *cnf = NULL;

  if (argc != 1)
    {
      io_buffer_store (vc, "-ERR [vuser] Incorrect arguments\n");
      return;
    }

  /* Useful strings. */
  domain = (char *) strdup (vc->virtual_server);
  name = (char *) strdup (argv[0]);

  if (vuser = (char *)malloc (strlen (domain) + strlen (name) + 2))
    sprintf (vuser, "%s@%s", name, domain);  

  /* Load domain recipe file. */
  if (rc = vsd_load_file (&fdomain_rc, vc->server_root, vuser_domain_rc (NULL, domain)))
    io_buffer_store (vc, 
		     "-ERR [vuser][%s] error loading domain recipe file\n", 
		     vuser);  

  /* Parse domain recipe file. */
  if (!rc && (rc = vuser_rc_parse (fdomain_rc, &domain_cnf)))
     io_buffer_store (vc, 
       "-ERR [vuser][%s] error parsing domain recipe file\n",
       vuser);		  

  /* Test for default vuser */
  if (!rc && !(strcmp (name, "admin")))
    {
      rc = VSD_ERR;
      io_buffer_store (vc, 
		       "-ERR [vuser][%s] cannot delete default vuser\n",
		       vuser);		  	
    }

  /* Test for existing vuser entry. */
  if (!rc && !(cnf = vuser_list_find (domain_cnf, name)))
    {
      rc = VSD_ERR;
      io_buffer_store (vc, 
		       "-ERR [vuser][%s] vuser does not exist\n",
		       vuser);		  
    }

  /* Load vuser recipe template. */
  if (!rc && (rc = vsd_load_file (&fvuser_rc, vc->server_root, vuser_template_vuser_rc (NULL, domain))))
    io_buffer_store (vc, 
		     "-ERR [vuser][%s] error loading vuser recipe template: %m\n", 
		     vuser);  

  /* Parse vuser recipe template. */
  if (!rc)
    {
      /* FIXME - globals, there must be a better way... */
      vuser_g_placeholders_init ();
      g_placeholders[VUSER_PH_VSD_VUSER_NAME].replacement = name;
      if (rc = vuser_template_parse (fvuser_rc))
	io_buffer_store (vc, 
			 "-ERR [vuser][%s] error parsing vuser recipe template\n",
			 vuser);		  
    }
  
  /* Locate vuser entry. */
  if (!rc && !(domain_line = vuser_rc_find (fdomain_rc, fvuser_rc->line)))
  {
    rc = VSD_ERR;
    io_buffer_store (vc, 
		     "-ERR [vuser][%s] error locating vuser entry\n",
		     vuser);		  
  }
      
  /* Delete vuser entry. */
  if (!rc)
    {
      while (tmp_line = domain_line->next)
	{
	  vsd_file_delete_line (fdomain_rc, domain_line);
	  domain_line = tmp_line;

	  pch = domain_line->line;
	  if ( *pch == '}' )
	    break;
	}
      vsd_file_delete_line (fdomain_rc, domain_line->next);
      vsd_file_delete_line (fdomain_rc, domain_line);
    }
  
  /* Save domain recipe file. */
  if (!rc && (rc = vsd_save_file (fdomain_rc, vc->server_root, vuser_domain_rc (NULL, domain))))
    io_buffer_store (vc, 
		     "-ERR [vuser][%s] error saving domain recipe file: %m\n", 
		     vuser);

  /* Load msg index. */
  if (!rc && (rc = vsd_load_file (&fmsgs, vc->server_root, vuser_msgs (NULL, domain, name))))
    io_buffer_store (vc, 
		     "-ERR [vuser][%s] error loading msg index: %m\n", 
		     vuser);  
  if (!rc)
    {
      /* Delete vuser msgs. */
      for (msgs_line = fmsgs->next; !rc && msgs_line; msgs_line=msgs_line->next)
	rc = unlink (vuser_msg (vc->server_root, domain, name, msgs_line->line));
      if (rc)
	io_buffer_store (vc, 
			 "-ERR [vuser][%s][%s] error deleting vuser msg: %m\n", 
			 vuser, msgs_line->line);

      /* Delete vuser msgs caches. */
      for (msgs_line = fmsgs->next; !rc && msgs_line; msgs_line=msgs_line->next)
	unlink (vuser_msg_cache (vc->server_root, domain, name, msgs_line->line));
      
      /* Delete vuser msg index. */
      if (rc = unlink (vuser_msgs (vc->server_root, domain, name)))
	io_buffer_store (vc, 
			 "-ERR [vuser][%s] error deleting vuser msg index: %m\n", 
			 vuser);

      /* Delete vuser msg directory. */
      if (rc = rmdir (vuser_msg_dir (vc->server_root, domain, name)))
	io_buffer_store (vc, 
			 "-ERR [vuser][%s] error deleting vuser msg directory: %m\n", 
			 vuser, vuser_msg_dir (vc->server_root, domain, name));
      
      /* Delete vuser recipe file. */
      if (rc = unlink (vuser_vuser_rc (vc->server_root, domain, name)))
	io_buffer_store (vc, 
			 "-ERR [vuser][%s] error deleting vuser recipe file: %m\n", 
			 vuser);

      /* Delete vuser directory. */
      if (rc = rmdir (vuser_vuser_dir (vc->server_root, domain, name)))
	io_buffer_store (vc, 
			 "-ERR [vuser][%s] error deleting vuser directory: %m\n", 
			 vuser);
    }

  free (domain);
  free (name);
  free (vuser);

  for (cnf = domain_cnf; cnf; cnf = cnf->next)
    {
      vuser_list_free (cnf->fwd_cnf);
      vuser_list_free (cnf->rsp_cnf);
    }
  vuser_list_free (domain_cnf);

  vsd_free_file (fmsgs);
  vsd_free_file (fvuser_rc);      
  vsd_free_file (fdomain_rc);      
}

/******************************************************************************
 * Description: add default forwarder to virtual user.
 *
 *     Returns: void.
 *****************************************************************************/
void 
vuser_dfltadd (struct connection *vc, 
	       const int         argc, 
	       const char        *argv[])
{ 
  char *pch, *buf;
  char *domain, *name, *fwd, *vuser;
  char entry[1024];
  int rc = VSD_OK;
  struct vsd_file_line *fdomain_rc = NULL;
  struct vsd_file_line *fvuser_rc = NULL;
  struct vsd_file_line *ffwd_rc = NULL;
  struct vsd_file_line *fpasswd = NULL;
  struct vsd_file_line *domain_line = NULL;
  struct vsd_file_line *fwd_line = NULL;
  struct vsd_file_line *passwd_line = NULL;
  struct vuser_cnf *domain_cnf = NULL;
  struct vuser_cnf *vuser_cnf = NULL;
  struct vuser_cnf *fwd_cnf = NULL;
  struct vuser_cnf *dflt_cnf = NULL;
  struct vuser_cnf *rsp_cnf = NULL;
  struct vuser_cnf *cnf = NULL;

  if (argc != 2)
    {
      io_buffer_store (vc, "-ERR [vuser] Incorrect arguments\n");
      return;
    }

  /* Useful strings. */
  domain = (char *) strdup (vc->virtual_server);
  name = (char *) strdup (argv[0]);
  fwd = (char *) strdup (argv[1]);

  if (vuser = (char *)malloc (strlen (domain) + strlen (name) + 2))
    sprintf (vuser, "%s@%s", name, domain);  

  /* Load domain recipe file. */
  if (rc = vsd_load_file (&fdomain_rc, vc->server_root, 
			  vuser_domain_rc (NULL, domain)))
    io_buffer_store (vc, 
		     "-ERR [vuser][%s] error loading domain recipe file: %m\n",
		     vuser);  

  /* Parse domain recipe file. */
  if (!rc && (rc = vuser_rc_parse (fdomain_rc, &domain_cnf)))
    io_buffer_store (vc, 
		     "-ERR [vuser][%s] error parsing domain recipe file\n",
		     vuser);		  

  /* Test for existing vuser entry. */
  if (!rc && !(cnf = vuser_list_find (domain_cnf, name)))
    {
      rc = VSD_ERR;
      io_buffer_store (vc, 
		       "-ERR [vuser][%s] vuser does not exist\n",
		       vuser);		  
    }

  /* Load passwd. */
  if (!rc && (rc = vsd_load_file (&fpasswd, vc->server_root, 
				  vuser_passwd (NULL, domain))))
    io_buffer_store (vc, 
		     "-ERR [vuser][%s] error loading passwd file: %m\n", 
		     vuser);  

  /* Test for valid destination. */
  if (!rc 
      && !(strchr (fwd, '@')) 
      && !(passwd_line = vuser_passwd_find (fpasswd, fwd)))
    {
      rc = VSD_ERR;
      io_buffer_store (vc, 
		       "-ERR [vuser][%s][%s] forwarder mailbox does not exist\n",
		       vuser, fwd);		  
    }

  /* Load vuser recipe file. */
  if (!rc && (rc = vsd_load_file (&fvuser_rc, vc->server_root, 
				  vuser_vuser_rc (NULL, domain, name))))
    io_buffer_store (vc, 
		     "-ERR [vuser][%s] error loading vuser recipe file: %m\n", 
		     vuser);  
  
  /* Parse vuser recipe file. */
  if (!rc && (rc = vuser_rc_parse (fvuser_rc, &vuser_cnf)))
    io_buffer_store (vc, 
		     "-ERR [vuser][%s] error parsing vuser recipe file\n",
		     vuser);		  

  /* Check for any existing default forwarder */
  if (!rc && (cnf->fwd_cnf != NULL))
    {
      rc = VSD_ERR;
      io_buffer_store (vc, 
		       "-ERR [vuser][%s][%s] default forwarder already exists: %s\n",
		       vuser, fwd, cnf->fwd_cnf->name);		  	
    }

  /* Test for existing forwarder entry. */
  if (!rc && (fwd_cnf = vuser_list_find (vuser_cnf->fwd_cnf, fwd)))
    {
      rc = VSD_ERR;
      io_buffer_store (vc, 
		       "-ERR [vuser][%s][%s] forwarder already exists\n",
		       vuser, fwd);		  	
    }

  /* Load default forwarder recipe template. */
  if (!rc)
    {
      if (strchr (fwd, '@'))
	{
	  if (rc = vsd_load_file (&ffwd_rc, vc->server_root, 
				  vuser_template_dflt_ext_rc (NULL, domain)))
	    io_buffer_store (vc, 
			     "-ERR [vuser][%s] error loading vuser default forwarder recipe template: %m\n", 
			     vuser);  
	}
      else
	{
	  if (rc = vsd_load_file (&ffwd_rc, vc->server_root, 
				  vuser_template_dflt_int_rc (NULL, domain)))
	    io_buffer_store (vc, 
			     "-ERR [vuser][%s] error loading vuser default forwarder recipe template: %m\n", 
			     vuser);  
	}
    }

  /* Parse default forwarder recipe template. */
  if (!rc)
    {
      /* FIXME - globals, there must be a better way... */
      vuser_g_placeholders_init ();
      g_placeholders[VUSER_PH_VSD_VUSER_NAME].replacement = fwd;
      g_placeholders[VUSER_PH_VSD_VUSER_FORWARDER_ADDRESS].replacement = fwd;
      if (rc = vuser_template_parse (ffwd_rc))
	io_buffer_store (vc, 
			 "-ERR [vuser][%s] error parsing default forwarder recipe template\n",
			 vuser);		  
    }

  /* Add default forwarder entry. */
  if (!rc)
    {
      /* Step to closing brace (then back a line). */
      for (domain_line = cnf->line; domain_line;
	   domain_line = domain_line->next)
	if (domain_line->line[0] == '}')
	  break;
      domain_line = domain_line->prev;

      for (fwd_line = ffwd_rc; !rc && fwd_line; 
	   fwd_line = fwd_line->next, domain_line = domain_line->next)
	  rc = vsd_file_insert_line (domain_line, fwd_line->line);
      
      if (rc)
	io_buffer_store (vc, 
			 "-ERR [vuser][%s] error adding default forwarder\n",
			 vuser);
    }

  /* Save domain recipe file. */
  if (!rc && (rc = vsd_save_file (fdomain_rc, vc->server_root, 
				  vuser_domain_rc (NULL, domain))))
    io_buffer_store (vc, 
		     "-ERR [vuser][%s] error saving domain recipe file: %m\n", 
		     vuser);
  
  free (vuser);
  free (domain);
  free (name);
  free (fwd);

  for (cnf = domain_cnf; cnf; cnf = cnf->next)
    {
      vuser_list_free (cnf->fwd_cnf);
      vuser_list_free (cnf->rsp_cnf);
    }
  vuser_list_free (domain_cnf);

  for (cnf = vuser_cnf; cnf; cnf = cnf->next)
    {
      vuser_list_free (cnf->fwd_cnf);
      vuser_list_free (cnf->rsp_cnf);
    }
  vuser_list_free (vuser_cnf);

  vsd_free_file (fpasswd);    
  vsd_free_file (ffwd_rc);    
  vsd_free_file (fvuser_rc);    
  vsd_free_file (fdomain_rc);    
}

/******************************************************************************
 * Description: delete default forwarder from virtual user.
 *
 *     Returns: void.
 *****************************************************************************/
void 
vuser_dfltdel (struct connection *vc, 
	       const int         argc, 
	       const char        *argv[])
{ 
  char *pch, *buf;
  char *domain, *name, *fwd, *vuser;
  char entry[1024];
  int rc = VSD_OK;
  struct vsd_file_line *fdomain_rc = NULL;
  struct vsd_file_line *fvuser_rc = NULL;
  struct vsd_file_line *ffwd_rc = NULL;
  struct vsd_file_line *domain_line = NULL;
  struct vsd_file_line *fwd_line = NULL;
  struct vsd_file_line *tmp_line = NULL;
  struct vuser_cnf *domain_cnf = NULL;
  struct vuser_cnf *vuser_cnf = NULL;
  struct vuser_cnf *fwd_cnf = NULL;
  struct vuser_cnf *rsp_cnf = NULL;
  struct vuser_cnf *cnf = NULL;

  if (argc != 2)
    {
      io_buffer_store (vc, "-ERR [vuser] Incorrect arguments\n");
      return;
    }

  /* Useful strings. */
  domain = (char *) strdup (vc->virtual_server);
  name = (char *) strdup (argv[0]);
  fwd = (char *) strdup (argv[1]);

  if (vuser = (char *)malloc (strlen (domain) + strlen (name) + 2))
    sprintf (vuser, "%s@%s", name, domain);  

  /* Load domain recipe file. */
  if (rc = vsd_load_file (&fdomain_rc, vc->server_root, 
			  vuser_domain_rc (NULL, domain)))
    io_buffer_store (vc, 
		     "-ERR [vuser][%s] error loading domain recipe file: %m\n",
		     vuser);  

  /* Parse domain recipe file. */
  if (!rc && (rc = vuser_rc_parse (fdomain_rc, &domain_cnf)))
    io_buffer_store (vc, 
		     "-ERR [vuser][%s] error parsing domain recipe file\n",
		     vuser);		  

  /* Test for existing vuser entry. */
  if (!rc && !(cnf = vuser_list_find (domain_cnf, name)))
    {
      rc = VSD_ERR;
      io_buffer_store (vc, 
		       "-ERR [vuser][%s] vuser does not exist\n",
		       vuser);		  
    }

  /* Test that fwd matches the actual forwarder */
  if (!rc && !(fwd_cnf = vuser_list_find (cnf->fwd_cnf, fwd)))
    {
      rc = VSD_ERR;
      io_buffer_store (vc,
		       "-ERR [vuser][%s][%s] forwarder does not match that of the vuser\n", vuser, fwd);
    }

  /* Load default forwarder recipe template. */
  if (!rc)
    {
      if (strchr (fwd, '@'))
	{
	  if (rc = vsd_load_file (&ffwd_rc, vc->server_root, 
				  vuser_template_dflt_ext_rc (NULL, domain)))
	    io_buffer_store (vc, 
			     "-ERR [vuser][%s] error loading vuser default forwarder recipe template: %m\n", 
			     vuser);  
	}
      else
	{
	  if (rc = vsd_load_file (&ffwd_rc, vc->server_root, 
				  vuser_template_dflt_int_rc (NULL, domain)))
	    io_buffer_store (vc, 
			     "-ERR [vuser][%s] error loading vuser default forwarder recipe template: %m\n", 
			     vuser);  
	}
    }
  
  /* Parse default forwarder recipe template. */
  if (!rc)
    {
      /* FIXME - globals, there must be a better way... */
      vuser_g_placeholders_init ();
      g_placeholders[VUSER_PH_VSD_VUSER_NAME].replacement = fwd;
      g_placeholders[VUSER_PH_VSD_VUSER_FORWARDER_ADDRESS].replacement = fwd;
      if (rc = vuser_template_parse (ffwd_rc))
	io_buffer_store (vc, 
			 "-ERR [vuser][%s] error parsing default forwarder recipe template\n",
			 vuser);		  
    }

#if 0
  /* Locate default forwarder entry. */
  if (!rc && !(domain_line = vuser_rc_find (fdomain_rc, ffwd_rc->line)))
  {
    rc = VSD_ERR;
    io_buffer_store (vc, 
		     "-ERR [vuser][%s] error locating default forwarder entry\n",
		     vuser);		  
  }
#endif
  /* Delete default forwarder entry. */
  if (!rc)
    {
      /* Step to closing brace (then back a line). */
      for (domain_line = cnf->line; domain_line;
	   domain_line = domain_line->next)
	if (domain_line->line[0] == '}')
	  break;
      domain_line = domain_line->prev;

      if (strlen(domain_line->line) != 0)
      {
	rc = VSD_ERR;
	io_buffer_store (vc, 
			 "-ERR [vuser][%s] default forwarder entry does not exist\n",
			 vuser);
      }

      for (fwd_line = ffwd_rc; !rc && fwd_line; fwd_line = fwd_line->next)
	{
	  tmp_line = domain_line->prev;
	  vsd_file_delete_line (fdomain_rc, domain_line);
	  domain_line = tmp_line;
	}
    }

  /* Save domain recipe file. */
  if (!rc && (rc = vsd_save_file (fdomain_rc, vc->server_root, 
				  vuser_domain_rc (NULL, domain))))
    io_buffer_store (vc, 
		     "-ERR [vuser][%s] error saving domain recipe file: %m\n", 
		     vuser);
  
  free (vuser);
  free (domain);
  free (name);
  free (fwd);

  for (cnf = domain_cnf; cnf; cnf = cnf->next)
    {
      vuser_list_free (cnf->fwd_cnf);
      vuser_list_free (cnf->rsp_cnf);
    }
  vuser_list_free (domain_cnf);

  for (cnf = vuser_cnf; cnf; cnf = cnf->next)
    {
      vuser_list_free (cnf->fwd_cnf);
      vuser_list_free (cnf->rsp_cnf);
    }
  vuser_list_free (vuser_cnf);

  vsd_free_file (ffwd_rc);    
  vsd_free_file (fvuser_rc);    
  vsd_free_file (fdomain_rc);    
}

/******************************************************************************
 * Description: add forwarder to virtual user.
 *
 *     Returns: void.
 *****************************************************************************/
void 
vuser_fwdadd (struct connection *vc, 
	      const int         argc, 
	      const char        *argv[])
{ 
  char *pch, *buf;
  char *domain, *name, *fwd, *vuser;
  char entry[1024];
  int rc = VSD_OK;
  struct vsd_file_line *fdomain_rc = NULL;
  struct vsd_file_line *fvuser_rc = NULL;
  struct vsd_file_line *ffwd_rc = NULL;
  struct vsd_file_line *fpasswd = NULL;
  struct vsd_file_line *vuser_line = NULL;
  struct vsd_file_line *fwd_line = NULL;
  struct vsd_file_line *passwd_line = NULL;
  struct vuser_cnf *domain_cnf = NULL;
  struct vuser_cnf *vuser_cnf = NULL;
  struct vuser_cnf *fwd_cnf = NULL;
  struct vuser_cnf *rsp_cnf = NULL;
  struct vuser_cnf *cnf = NULL;

  if (argc != 2)
    {
      io_buffer_store (vc, "-ERR [vuser] Incorrect arguments\n");
      return;
    }

  /* Useful strings. */
  domain = (char *) strdup (vc->virtual_server);
  name = (char *) strdup (argv[0]);
  fwd = (char *) strdup (argv[1]);

  if (vuser = (char *)malloc (strlen (domain) + strlen (name) + 2))
    sprintf (vuser, "%s@%s", name, domain);  

  /* Load domain recipe file. */
  if (rc = vsd_load_file (&fdomain_rc, vc->server_root, vuser_domain_rc (NULL, domain)))
    io_buffer_store (vc, 
		     "-ERR [vuser][%s] error loading domain recipe file: %m\n", 
		     vuser);  

  /* Parse domain recipe file. */
  if (!rc && (rc = vuser_rc_parse (fdomain_rc, &domain_cnf)))
    io_buffer_store (vc, 
		     "-ERR [vuser][%s] error parsing domain recipe file\n",
		     vuser);		  

  /* Test for existing vuser entry. */
  if (!rc && !(cnf = vuser_list_find (domain_cnf, name)))
    {
      rc = VSD_ERR;
      io_buffer_store (vc, 
		       "-ERR [vuser][%s] vuser does not exist\n",
		       vuser);		  
    }

  /* Test for default forwarder. */
  if (!rc && !(strcmp (cnf->name, fwd)))
    {
      rc = VSD_ERR;
      io_buffer_store (vc, 
		       "-ERR [vuser][%s][%s] forwarder already exists\n",
		       vuser, fwd);		  	
    }

  /* Load passwd. */
  if (!rc && (rc = vsd_load_file (&fpasswd, vc->server_root, vuser_passwd (NULL, domain))))
    io_buffer_store (vc, 
		     "-ERR [vuser][%s] error loading passwd file: %m\n", 
		     vuser);  

  /* Test for valid destination. */
  if (!rc && !(strchr (fwd, '@')) && !(passwd_line = vuser_passwd_find (fpasswd, fwd)))
    {
      rc = VSD_ERR;
      io_buffer_store (vc, 
		       "-ERR [vuser][%s][$s] forwarder mailbox does not exist\n",
		       vuser, fwd);		  
    }

  /* Load vuser recipe file. */
  if (!rc && (rc = vsd_load_file (&fvuser_rc, vc->server_root, vuser_vuser_rc (NULL, domain, name))))
    io_buffer_store (vc, 
		     "-ERR [vuser][%s] error loading vuser recipe file: %m\n", 
		     vuser);  
  
  /* Parse vuser recipe file. */
  if (!rc && (rc = vuser_rc_parse (fvuser_rc, &vuser_cnf)))
    io_buffer_store (vc, 
		     "-ERR [vuser][%s] error parsing vuser recipe file\n",
		     vuser);		  

  /* Test for existing forwarder entry. */
  if (!rc && (fwd_cnf = vuser_list_find (vuser_cnf->fwd_cnf, fwd)))
    {
      rc = VSD_ERR;
      io_buffer_store (vc, 
		       "-ERR [vuser][%s][%s] forwarder already exists\n",
		       vuser, fwd);		  	
    }

  /* Load forwarder recipe template. */
  if (!rc)
    {
      if (strchr (fwd, '@'))
	{
	  if (rc = vsd_load_file (&ffwd_rc, vc->server_root, vuser_template_fwd_ext_rc (NULL, domain)))
	    io_buffer_store (vc, 
			     "-ERR [vuser][%s] error loading vuser forwarder recipe template: %m\n", 
			     vuser);  
	}
      else
	{
	  if (rc = vsd_load_file (&ffwd_rc, vc->server_root, vuser_template_fwd_int_rc (NULL, domain)))
	    io_buffer_store (vc, 
			     "-ERR [vuser][%s] error loading vuser forwarder recipe template: %m\n", 
			     vuser);  
	}
    }

  /* Parse forwarder recipe template. */
  if (!rc)
    {
      /* FIXME - globals, there must be a better way... */
      vuser_g_placeholders_init ();
      g_placeholders[VUSER_PH_VSD_VUSER_NAME].replacement = fwd;
      g_placeholders[VUSER_PH_VSD_VUSER_FORWARDER_ADDRESS].replacement = fwd;
      if (rc = vuser_template_parse (ffwd_rc))
	io_buffer_store (vc, 
			 "-ERR [vuser][%s] error parsing forwarder recipe template\n",
			 vuser);		  
    }

  /* Add forwarder entry. */
  if (!rc)
    {
	for (fwd_line = ffwd_rc, vuser_line = fvuser_rc->next; 
	     !rc && fwd_line; 
	     fwd_line = fwd_line->next, vuser_line = vuser_line->next)
	  rc = vsd_file_insert_line (vuser_line, fwd_line->line);

	if (rc)
	  io_buffer_store (vc, 
			   "-ERR [vuser][%s] error adding forwarder\n",
			   vuser);
    }
  
  /* Save vuser recipe file. */
  if (!rc && (rc = vsd_save_file (fvuser_rc, vc->server_root, vuser_vuser_rc (NULL, domain, name))))
    io_buffer_store (vc, 
		     "-ERR [vuser][%s] error saving vuser recipe file: %m\n", 
		     vuser);
  
  free (vuser);
  free (domain);
  free (name);
  free (fwd);

  for (cnf = domain_cnf; cnf; cnf = cnf->next)
    {
      vuser_list_free (cnf->fwd_cnf);
      vuser_list_free (cnf->rsp_cnf);
    }
  vuser_list_free (domain_cnf);

  for (cnf = vuser_cnf; cnf; cnf = cnf->next)
    {
      vuser_list_free (cnf->fwd_cnf);
      vuser_list_free (cnf->rsp_cnf);
    }
  vuser_list_free (vuser_cnf);

  vsd_free_file (fpasswd);    
  vsd_free_file (ffwd_rc);    
  vsd_free_file (fvuser_rc);    
  vsd_free_file (fdomain_rc);    
}


/******************************************************************************
 * Description: delete forwarder from virtual user.
 *
 *     Returns: void.
 *****************************************************************************/
void 
vuser_fwddel (struct connection *vc, 
	      const int         argc, 
	      const char        *argv[])
{
  char *pch, *domain, *name, *fwd, *vuser;
  char entry[256];
  int rc = VSD_OK;
  struct vsd_file_line *fdomain_rc = NULL;
  struct vsd_file_line *fvuser_rc = NULL;
  struct vsd_file_line *ffwd_rc = NULL;
  struct vsd_file_line *vuser_line = NULL;
  struct vsd_file_line *fwd_line = NULL;
  struct vsd_file_line *tmp_line = NULL;
  struct vuser_cnf *domain_cnf = NULL;
  struct vuser_cnf *vuser_cnf = NULL;
  struct vuser_cnf *fwd_cnf = NULL;
  struct vuser_cnf *rsp_cnf = NULL;
  struct vuser_cnf *cnf = NULL;

  if (argc != 2)
    {
      io_buffer_store (vc, "-ERR [vuser] Incorrect arguments\n");
      return;
    }

  /* Useful strings. */
  domain = (char *) strdup (vc->virtual_server);
  name = (char *) strdup (argv[0]);
  fwd = (char *) strdup (argv[1]);

  if (vuser = (char *)malloc (strlen (domain) + strlen (name) + 2))
    sprintf (vuser, "%s@%s", name, domain);  

  /* Load domain recipe file. */
  if (rc = vsd_load_file (&fdomain_rc, vc->server_root, vuser_domain_rc (NULL, domain)))
    io_buffer_store (vc, 
		     "-ERR [vuser][%s] error loading domain recipe file\n", 
		     vuser);  

  /* Parse domain recipe file. */
  if (!rc && (rc = vuser_rc_parse (fdomain_rc, &domain_cnf)))
    io_buffer_store (vc, 
		     "-ERR [vuser][%s] error parsing domain recipe file\n",
		     vuser);		  

  /* Test for existing vuser entry. */
  if (!rc && !(cnf = vuser_list_find (domain_cnf, name)))
    {
      rc = VSD_ERR;
      io_buffer_store (vc, 
		       "-ERR [vuser][%s] vuser does not exist\n",
		       vuser);		  
    }
  /* Test for default forwarder. */
  if (!rc && !(strcmp (cnf->name, fwd)))
    {
      rc = VSD_ERR;
      io_buffer_store (vc, 
		       "-ERR [vuser][%s][%s] cannot delete default forwarder\n",
		       vuser, fwd);		  	
    }
    
  /* Load vuser recipe file. */
  if (!rc && (rc = vsd_load_file (&fvuser_rc, vc->server_root, vuser_vuser_rc (NULL, domain, name))))
    io_buffer_store (vc, 
		     "-ERR [vuser][%s] error loading vuser recipe file: %m\n", 
		     vuser);  

  /* Parse vuser recipe file. */
  if (!rc && (rc = vuser_rc_parse (fvuser_rc, &vuser_cnf)))
    io_buffer_store (vc, 
		     "-ERR [vuser][%s] error parsing vuser recipe file\n",
		     vuser);		  

  /* Test for existing forwarder entry. */
  if (!rc && !(fwd_cnf = vuser_list_find (vuser_cnf->fwd_cnf, fwd)))
    {
      rc = VSD_ERR;
      io_buffer_store (vc, 
		       "-ERR [vuser][%s][%s] forwarder does not exist\n",
		       vuser, fwd);		  	
    }

  /* Load forwarder recipe template. */
  if (!rc)
    {
      if (strchr (fwd, '@'))
	{
	  if (rc = vsd_load_file (&ffwd_rc, vc->server_root, vuser_template_fwd_ext_rc (NULL, domain)))
	    io_buffer_store (vc, 
			     "-ERR [vuser][%s] error loading vuser forwarder recipe template: %m\n", 
			     vuser);  
	}
      else
	{
	  if (rc = vsd_load_file (&ffwd_rc, vc->server_root, vuser_template_fwd_int_rc (NULL, domain)))
	    io_buffer_store (vc, 
			     "-ERR [vuser][%s] error loading vuser forwarder recipe template: %m\n", 
			     vuser);  
	}
    }

  /* Parse forwarder recipe template. */
  if (!rc)
    {
      /* FIXME - globals, there must be a better way... */
      vuser_g_placeholders_init ();
      g_placeholders[VUSER_PH_VSD_VUSER_NAME].replacement = fwd;
      g_placeholders[VUSER_PH_VSD_VUSER_FORWARDER_ADDRESS].replacement = fwd;
      if (rc = vuser_template_parse (ffwd_rc))
	io_buffer_store (vc, 
			 "-ERR [vuser][%s] error parsing forwarder recipe template\n",
			 vuser);		  
    }
  
  /* Locate forwarder entry. */
  if (!rc && !(vuser_line = vuser_rc_find (fvuser_rc, ffwd_rc->line)))
    {
      rc = VSD_ERR;
      io_buffer_store (vc, 
		       "-ERR [vuser][%s][%s] error locating forwarder entry\n",
		       vuser, fwd);		  
    }

  /* Delete forwarder entry. */
  if (!rc)
    {
      for (fwd_line = ffwd_rc; !rc && fwd_line; fwd_line = fwd_line->next)
	{
	  tmp_line = vuser_line->next;
	  vsd_file_delete_line (fvuser_rc, vuser_line);
	  vuser_line = tmp_line;
	}
    }
  
  /* Save vuser recipe file. */
  if (!rc && (rc = vsd_save_file (fvuser_rc, vc->server_root, vuser_vuser_rc (NULL, domain, name))))
    io_buffer_store (vc, 
		     "-ERR [vuser][%s] error saving vuser recipe file: %m\n", 
		     vuser);

  free (domain);
  free (name);
  free (fwd);
  free (vuser);

  for (cnf = domain_cnf; cnf; cnf = cnf->next)
    {
      vuser_list_free (cnf->fwd_cnf);
      vuser_list_free (cnf->rsp_cnf);
    }
  vuser_list_free (domain_cnf);

  for (cnf = vuser_cnf; cnf; cnf = cnf->next)
    {
      vuser_list_free (cnf->fwd_cnf);
      vuser_list_free (cnf->rsp_cnf);
    }
  vuser_list_free (vuser_cnf);

  vsd_free_file (ffwd_rc);      
  vsd_free_file (fvuser_rc);      
  vsd_free_file (fdomain_rc);      
}

/******************************************************************************
 * Description: add responder to virtual user.
 *
 *     Returns: void.
 *****************************************************************************/
void 
vuser_rspadd (struct connection *vc, 
	      const int         argc, 
	      const char        *argv[])
{ 
  char *pch, *buf;
  char *domain, *name, *rsp, *msg, *vuser;
  char entry[1024];
  int rc = VSD_OK;
  struct vsd_file_line *fdomain_rc = NULL;
  struct vsd_file_line *fvuser_rc = NULL;
  struct vsd_file_line *frsp_rc = NULL;
  struct vsd_file_line *fmsgs = NULL;
  struct vsd_file_line *vuser_line = NULL;
  struct vsd_file_line *rsp_line = NULL;
  struct vsd_file_line *msg_line = NULL;
  struct vuser_cnf *domain_cnf = NULL;
  struct vuser_cnf *vuser_cnf = NULL;
  struct vuser_cnf *rsp_cnf = NULL;
  struct vuser_cnf *cnf = NULL;

  if (argc != 2)
    {
      io_buffer_store (vc, "-ERR [vuser] Incorrect arguments\n");
      return;
    }

  /* Useful strings. */
  domain = (char *) strdup (vc->virtual_server);
  name = (char *) strdup (argv[0]);
  rsp = (char *) strdup (argv[1]);
  msg = rsp;

  if (vuser = (char *)malloc (strlen (domain) + strlen (name) + 2))
    sprintf (vuser, "%s@%s", name, domain);  

  /* Load domain recipe file. */
  if (rc = vsd_load_file (&fdomain_rc, vc->server_root, vuser_domain_rc (NULL, domain)))
    io_buffer_store (vc, 
		     "-ERR [vuser][%s] error loading domain recipe file: %m\n", 
		     vuser);  

  /* Parse domain recipe file. */
  if (!rc && (rc = vuser_rc_parse (fdomain_rc, &domain_cnf)))
    io_buffer_store (vc, 
		     "-ERR [vuser][%s] error parsing domain recipe file\n",
		     vuser);		  
  
  /* Test for existing vuser entry. */
  if (!rc && !(cnf = vuser_list_find (domain_cnf, name)))
    {
      rc = VSD_ERR;
      io_buffer_store (vc, 
		       "-ERR [vuser][%s] vuser does not exist\n",
		       vuser);		  
    }
  
  /* Test for existing responder entry. */
  if (!rc && (rsp_cnf = vuser_list_find (domain_cnf->rsp_cnf, rsp)))
    {
      rc = VSD_ERR;
      io_buffer_store (vc, 
		       "-ERR [vuser][%s][%s] responder already exists\n",
		       vuser, rsp);		  	
    }
  
  /* Load msg index. */
  if (!rc && (rc = vsd_load_file (&fmsgs, vc->server_root, vuser_msgs (NULL, domain, name))))
    io_buffer_store (vc, 
		     "-ERR [vuser][%s] error loading msg index: %m\n", 
		     vuser);  
  
  /* Test for existing msg entry. */
  if (!rc && !(msg_line = vuser_msgs_find (fmsgs, msg)))
    {
      rc = VSD_ERR;
      io_buffer_store (vc, 
		       "-ERR [vuser][%s][%s] msg does not exist\n",
		       vuser, msg);		  
    }
    
  /* Load vuser recipe file. */
  if (!rc && (rc = vsd_load_file (&fvuser_rc, vc->server_root, vuser_vuser_rc (NULL, domain, name))))
    io_buffer_store (vc, 
		     "-ERR [vuser][%s] error loading vuser recipe file: %m\n", 
		     vuser);  
  
  /* Parse vuser recipe file. */
  if (!rc && (rc = vuser_rc_parse (fvuser_rc, &vuser_cnf)))
    io_buffer_store (vc, 
		     "-ERR [vuser][%s] error parsing vuser recipe file\n",
		     vuser);		  

  /* Test for existing responder entry. */
  if (!rc && (rsp_cnf = vuser_list_find (vuser_cnf->rsp_cnf, rsp)))
    {
      rc = VSD_ERR;
      io_buffer_store (vc, 
		       "-ERR [vuser][%s][%s] responder already exists\n",
		       vuser, rsp);		  	
    }

  /* Load responder recipe template. */
  if (!rc && (rc = vsd_load_file (&frsp_rc, vc->server_root, vuser_template_rsp_rc (NULL, domain))))
	    io_buffer_store (vc, 
			     "-ERR [vuser][%s] error loading vuser responder recipe template: %m\n", 
			     vuser);  

  /* Parse responder recipe template. */
  if (!rc)
    {
      /* FIXME - globals, there must be a better way... */
      vuser_g_placeholders_init ();
      g_placeholders[VUSER_PH_VSD_VUSER_NAME].replacement = name;
      g_placeholders[VUSER_PH_VSD_VUSER_RESPONDER_MESSAGE].replacement = msg;
      if (rc = vuser_template_parse (frsp_rc))
	io_buffer_store (vc, 
			 "-ERR [vuser][%s] error parsing responder recipe template\n",
			 vuser);		  
    }

  /* Add responder entry. */
  if (!rc)
    {
	for (rsp_line = frsp_rc, vuser_line = fvuser_rc->next; 
	     !rc && rsp_line; 
	     rsp_line = rsp_line->next, vuser_line = vuser_line->next)
	  rc = vsd_file_insert_line (vuser_line, rsp_line->line);

	if (rc)
	  io_buffer_store (vc, 
			   "-ERR [vuser][%s] error adding responder\n",
			   vuser);
    }
  
  /* Save vuser recipe file. */
  if (!rc && (rc = vsd_save_file (fvuser_rc, vc->server_root, vuser_vuser_rc (NULL, domain, name))))
    io_buffer_store (vc, 
		     "-ERR [vuser][%s] error saving vuser recipe file: %m\n", 
		     vuser);
  
  free (vuser);
  free (domain);
  free (name);
  free (rsp);

  for (cnf = domain_cnf; cnf; cnf = cnf->next)
    {
      vuser_list_free (cnf->fwd_cnf);
      vuser_list_free (cnf->rsp_cnf);
    }
  vuser_list_free (domain_cnf);

  for (cnf = vuser_cnf; cnf; cnf = cnf->next)
    {
      vuser_list_free (cnf->fwd_cnf);
      vuser_list_free (cnf->rsp_cnf);
    }
  vuser_list_free (vuser_cnf);

  vsd_free_file (fmsgs);    
  vsd_free_file (frsp_rc);    
  vsd_free_file (fvuser_rc);    
  vsd_free_file (fdomain_rc);    
}

/******************************************************************************
 * Description: delete responder from virtual user.
 *
 *     Returns: void.
 *****************************************************************************/
void 
vuser_rspdel (struct connection *vc, 
	      const int         argc, 
	      const char        *argv[])
{
  char *pch, *domain, *name, *msg, *rsp, *vuser;
  char entry[256];
  int rc = VSD_OK;
  struct vsd_file_line *fdomain_rc = NULL;
  struct vsd_file_line *fvuser_rc = NULL;
  struct vsd_file_line *frsp_rc = NULL;
  struct vsd_file_line *vuser_line = NULL;
  struct vsd_file_line *rsp_line = NULL;
  struct vsd_file_line *tmp_line = NULL;
  struct vuser_cnf *domain_cnf = NULL;
  struct vuser_cnf *vuser_cnf = NULL;
  struct vuser_cnf *rsp_cnf = NULL;
  struct vuser_cnf *cnf = NULL;

  if (argc != 2)
    {
      io_buffer_store (vc, "-ERR [vuser] Incorrect arguments\n");
      return;
    }

  /* Useful strings. */
  domain = (char *) strdup (vc->virtual_server);
  name = (char *) strdup (argv[0]);
  rsp = (char *) strdup (argv[1]);
  msg = rsp;

  if (vuser = (char *)malloc (strlen (domain) + strlen (name) + 2))
    sprintf (vuser, "%s@%s", name, domain);  

  /* Load domain recipe file. */
  if (rc = vsd_load_file (&fdomain_rc, vc->server_root, vuser_domain_rc (NULL, domain)))
    io_buffer_store (vc, 
		     "-ERR [vuser][%s] error loading procmailrc file\n", 
		     vuser);  

  /* Parse domain recipe file. */
  if (!rc && (rc = vuser_rc_parse (fdomain_rc, &domain_cnf)))
    io_buffer_store (vc, 
		     "-ERR [vuser][%s] error parsing domain recipe file\n",
		     vuser);		  

  /* Test for existing vuser entry. */
  if (!rc && !(cnf = vuser_list_find (domain_cnf, name)))
    {
      rc = VSD_ERR;
      io_buffer_store (vc, 
		       "-ERR [vuser][%s] vuser does not exist\n",
		       vuser);		  
    }

  /* Load vuser recipe file. */
  if (!rc && (rc = vsd_load_file (&fvuser_rc, vc->server_root, vuser_vuser_rc (NULL, domain, name))))
    io_buffer_store (vc, 
		     "-ERR [vuser][%s] error loading vuser recipe file: %m\n", 
		     vuser);  
  
  /* Parse vuser recipe file. */
  if (!rc && (rc = vuser_rc_parse (fvuser_rc, &vuser_cnf)))
    io_buffer_store (vc, 
		     "-ERR [vuser][%s] error parsing vuser recipe file\n",
		     vuser);		  

  /* Test for existing responder entry. */
  if (!rc && !(rsp_cnf = vuser_list_find (vuser_cnf->rsp_cnf, rsp)))
    {
      rc = VSD_ERR;
      io_buffer_store (vc, 
		       "-ERR [vuser][%s][%s] responder does not exist\n",
		       vuser, rsp);		  	
    }

  /* Load responder recipe template. */
  if (!rc && (rc = vsd_load_file (&frsp_rc, vc->server_root, vuser_template_rsp_rc (NULL, domain))))
    io_buffer_store (vc, 
		     "-ERR [vuser][%s] error loading vuser responder recipe template: %m\n", 
		     vuser);  
  
  /* Parse responder recipe template. */
  if (!rc)
    {
      /* FIXME - globals, there must be a better way... */
      vuser_g_placeholders_init ();
      g_placeholders[VUSER_PH_VSD_VUSER_NAME].replacement = name;
      g_placeholders[VUSER_PH_VSD_VUSER_RESPONDER_MESSAGE].replacement = msg;
      if (rc = vuser_template_parse (frsp_rc))
	io_buffer_store (vc, 
			 "-ERR [vuser][%s] error parsing responder recipe template\n",
			 vuser);		  
    }

  /* Locate responder entry. */
  if (!rc && !(vuser_line = vuser_rc_find (fvuser_rc, frsp_rc->line)))
    {
      rc = VSD_ERR;
      io_buffer_store (vc, 
		       "-ERR [vuser][%s][%s] error locating responder entry\n",
		       vuser, rsp);		  
    }

  /* Delete entry. */
  if (!rc)
    {
      for (rsp_line = frsp_rc; !rc && rsp_line; rsp_line = rsp_line->next)
	{
	  tmp_line = vuser_line->next;
	  vsd_file_delete_line (fvuser_rc, vuser_line);
	  vuser_line = tmp_line;
	}
    }
  
  /* Save vuser recipe file. */
  if (!rc && (rc = vsd_save_file (fvuser_rc, vc->server_root, vuser_vuser_rc (NULL, domain, name))))
    io_buffer_store (vc, 
		     "-ERR [vuser][%s] error saving vuser recipe file: %m\n", 
		     vuser);
  
  free (domain);
  free (name);
  free (rsp);
  free (vuser);

  for (cnf = domain_cnf; cnf; cnf = cnf->next)
    {
      vuser_list_free (cnf->fwd_cnf);
      vuser_list_free (cnf->rsp_cnf);
    }
  vuser_list_free (domain_cnf);

  for (cnf = vuser_cnf; cnf; cnf = cnf->next)
    {
      vuser_list_free (cnf->fwd_cnf);
      vuser_list_free (cnf->rsp_cnf);
    }
  vuser_list_free (vuser_cnf);

  vsd_free_file (frsp_rc);      
  vsd_free_file (fvuser_rc);      
  vsd_free_file (fdomain_rc);      
}

/******************************************************************************
 * Description: list virtual users.
 *
 *     Returns: void.
 *****************************************************************************/
void 
vuser_list (struct connection *vc, 
	    const int         argc, 
	    const char        *argv[])
{
  char *pch, *domain, *name, *vuser;
  int rc = VSD_OK;
  struct vsd_file_line *fdomain_rc = NULL;
  struct vsd_file_line *fvuser_rc = NULL;
  struct vsd_file_line *line = NULL;
  struct vuser_cnf *domain_cnf = NULL;
  struct vuser_cnf *vuser_cnf = NULL;
  struct vuser_cnf *fwd_cnf = NULL;
  struct vuser_cnf *rsp_cnf = NULL;
  struct vuser_cnf *cnf = NULL;

  if (argc < 0 || argc > 1)
    {
      io_buffer_store (vc, "-ERR [vuser] Incorrect arguments\n");
      return;
    }

  /* Useful strings. */
  domain = (char *) strdup (vc->virtual_server);
  name = (argc == 1 ? (char *) strdup (argv[0]) : NULL);

  if (argc == 1)
    {
      if (vuser = (char *)malloc (strlen (domain) + strlen (name) + 2))
	sprintf (vuser, "%s@%s", name, domain);  

      /* Load domain recipe file. */
      if (rc = vsd_load_file (&fdomain_rc, vc->server_root, vuser_domain_rc (NULL, domain)))
	io_buffer_store (vc, 
			 "-ERR [vuser][%s] error loading domain recipe file\n", 
			 vuser);  
      
      /* Parse domain recipe file. */
      if (!rc && (rc = vuser_rc_parse (fdomain_rc, &domain_cnf)))
	io_buffer_store (vc, 
			 "-ERR [vuser][%s] error parsing domain recipe file\n", 
			 vuser);  

      /* Test for existing vuser entry. */
      if (!rc && !(cnf = vuser_list_find (domain_cnf, name)))
	{
	  rc = VSD_ERR;
	  io_buffer_store (vc, 
			   "-ERR [vuser][%s] vuser does not exist\n",
			   vuser);		  
	}

      /* Load vuser recipe file. */
      if (!rc && (rc = vsd_load_file (&fvuser_rc, vc->server_root, vuser_vuser_rc (NULL, domain, name))))
	io_buffer_store (vc, 
			 "-ERR [vuser][%s] error loading vuser recipe file\n", 
			 domain);  
      
      /* Parse vuser recipe file. */
      if (!rc && (rc = vuser_rc_parse (fvuser_rc, &vuser_cnf)))
	io_buffer_store (vc, 
			 "-ERR [vuser][%s] error parsing vuser recipe file\n", 
			 vuser);  

      /* Return entries. */
      for (cnf = vuser_cnf; !rc && cnf; cnf = cnf->next)
	{
	  io_buffer_store (vc, "vuser=%s", name);
	  
	  io_buffer_store (vc, " fwd=");
	  for (fwd_cnf = cnf->fwd_cnf; !rc && fwd_cnf; fwd_cnf = fwd_cnf->next)
	    io_buffer_store (vc, "%s,", fwd_cnf->name);

	  io_buffer_store (vc, " rsp=");
	  for (rsp_cnf = cnf->rsp_cnf; !rc && rsp_cnf; rsp_cnf = rsp_cnf->next)
	    io_buffer_store (vc, "%s,", rsp_cnf->name);
	  
	  io_buffer_store (vc, " \n");
	}
    }
  else
    {
      if (vuser = (char *)malloc (strlen (domain) + 1))
	sprintf (vuser, "%s", domain);  

      /* Load domain recipe file. */
      if (rc = vsd_load_file (&fdomain_rc, vc->server_root, vuser_domain_rc (NULL, domain)))
	io_buffer_store (vc, 
			 "-ERR [vuser][%s] error loading domain recipe file\n", 
			 vuser);  
      
      /* Parse domain recipe file. */
      if (!rc && (rc = vuser_rc_parse (fdomain_rc, &domain_cnf)))
	io_buffer_store (vc, 
			 "-ERR [vuser][%s] error parsing domain recipe file\n", 
			 vuser);  

      /* Return entries. */
      for (cnf = domain_cnf; !rc && cnf; cnf = cnf->next)
	{
	  io_buffer_store (vc, "vuser=%s", cnf->name);
	  
	  io_buffer_store (vc, " default=");
	  if (cnf->fwd_cnf)
	    io_buffer_store (vc, "%s", cnf->fwd_cnf->name);

	  io_buffer_store (vc, " \n");
	}
    }

  free (domain);
  free (name);
  free (vuser);

  for (cnf = domain_cnf; cnf; cnf = cnf->next)
    {
      vuser_list_free (cnf->fwd_cnf);
      vuser_list_free (cnf->rsp_cnf);
    }
  vuser_list_free (domain_cnf);

  for (cnf = vuser_cnf; cnf; cnf = cnf->next)
    {
      vuser_list_free (cnf->fwd_cnf);
      vuser_list_free (cnf->rsp_cnf);
    }
  vuser_list_free (vuser_cnf);

  vsd_free_file (fvuser_rc);
  vsd_free_file (fdomain_rc);
}

/******************************************************************************
 * Description: add virtual user mailbox.
 *
 *     Returns: void.
 *****************************************************************************/
void 
vuser_mbxadd (struct connection *vc, 
	      const int         argc, 
	      const char        *argv[])
{ 
  char *pch, *buf;
  char *domain, *name, *vuser;
  char *crypt, *pass;
  char entry[256];
  uid_t admin_uid = 0;
  gid_t admin_gid = 0;
  uid_t mail_uid = 0;
  gid_t mail_gid = 0;
  int rc = VSD_OK;
  struct vsd_file_line *fpasswd = NULL;
  struct vsd_file_line *line;
  struct vsd_vs *domain_map;

  if (argc < 1 || argc > 2)
    {
      io_buffer_store (vc, "-ERR [vuser] Incorrect arguments\n");
      return;
    }

  /* Useful strings. */
  domain = (char *) strdup (vc->virtual_server);
  name = (char *) strdup (argv[0]);
  pass = (argv[1] ? (char *) strdup (argv[1]) : NULL);
  sprintf (entry, "%s:!PleaseChangeMe", name);

  if (vuser = (char *)malloc (strlen (domain) + strlen (name) + 2))
    sprintf (vuser, "%s@%s", name, domain);  

  /* Validate input. */
  if (pch = vuser_name_validate (name))
    io_buffer_store (vc, "-ERR [vuser][%s] %s\n", vuser, pch);
  
  if (pch = vuser_domain_validate (domain))
    io_buffer_store (vc, "-ERR [vuser][%s] %s\n", vuser, pch);

  domain_map = vsd_getvdent (vc->map, domain);

  if ((pch = 
       vuser_check_mbx_limit (domain_map,
                              vsd_map_server_root (vc->map, domain_map->hs))) != NULL)
    {
      rc = VSD_ERR;
      io_buffer_store (vc, "-ERR [vuser][limit check] %s\n", pch);
    }

  /* Load passwd. */
  if (!rc && (rc = vsd_load_file (&fpasswd, vc->server_root, 
                                  vuser_passwd (NULL, domain))))
    io_buffer_store (vc, 
		     "-ERR [vuser][%s] error loading passwd file: %m\n", 
		     vuser);  

  /* Test for existing entry. */
  if (!rc)
    {
      if (line = vuser_passwd_find (fpasswd, name))
	{
	  rc = VSD_ERR;
	  io_buffer_store (vc, 
			   "-ERR [vuser][%s] mailbox already exists\n",
			   vuser);		  
	}
    }

  /* Set password if specified. */
  if (!rc && pass)
    {
      if (!(crypt = (char *) vsd_passwd_crypt (pass)))
	{
	  rc = VSD_ERR;
	  io_buffer_store (vc, 
			   "-ERR [vuser][%s] error encrypting mailbox passwd\n",
			   vuser);
	}
      else
	sprintf (entry, "%s:%s", name, crypt);    
    }
  
  /* Add entry. */
  if (!rc && (rc = vsd_file_insert_line (vsd_file_tail (fpasswd), entry)))
    io_buffer_store (vc, 
		     "-ERR [vuser][%s] error adding mailbox to passwd file\n",
		     vuser);

  /* Save passwd. */
  if (!rc && (rc = vsd_save_file (fpasswd, vc->server_root, vuser_passwd (NULL, domain))))
    io_buffer_store (vc, 
		     "-ERR [vuser][%s] error saving passwd file: %m\n", 
		     vuser);

  free (vuser);
  free (name);
  free (domain);
  free (pass);

  vsd_free_file (fpasswd);    
}

/******************************************************************************
 * Description: delete virtual user mailbox.
 *
 *     Returns: void.
 *****************************************************************************/
void 
vuser_mbxdel (struct connection *vc, 
	      const int         argc, 
	      const char        *argv[])
{
  char *pch, *domain, *name, *vuser;
  int rc = VSD_OK;
  struct vsd_file_line *fpasswd = NULL;
  struct vsd_file_line *line;
  
  if (argc != 1)
    {
      io_buffer_store (vc, "-ERR [vuser] Incorrect arguments\n");
      return;
    }

  /* Useful strings. */
  domain = (char *) strdup (vc->virtual_server);
  name = (char *) strdup (argv[0]);

  if (vuser = (char *)malloc (strlen (domain) + strlen (name) + 2))
    sprintf (vuser, "%s@%s", name, domain);  

  /* Load passwd. */
  if (rc = vsd_load_file (&fpasswd, vc->server_root, vuser_passwd (NULL, domain)))
    io_buffer_store (vc, 
		     "-ERR [vuser][%s] error loading passwd file: %m\n", 
		     vuser);  

  /* Test for admin. */
  if (!rc && !strcmp (name, "admin"))
    {
      rc = VSD_ERR;
      io_buffer_store (vc, 
		       "-ERR [vuser][%s] cannot delete admin mailbox\n", 
		       vuser);
      return;
    }
    
  /* Test for existing entry. */
  if (!rc && !(line = vuser_passwd_find (fpasswd, name)))
    {
      rc = VSD_ERR;
      io_buffer_store (vc, 
		       "-ERR [vuser][%s] mailbox does not exist\n",
		       vuser);		  
    }

  /* Test for mailbox still referenced */
  if (!rc && (vuser_check_mbx_ref (domain, vc->server_root, name)))
    {
      rc = VSD_ERR;
      io_buffer_store (vc,
		       "-ERR [vuser][%s] mailbox still referenced\n",
		       vuser);
    }

  /* Delete entry. */
  if (!rc)
    fpasswd = vsd_file_delete_line (fpasswd, line);

  /* Save passwd. */
  if (!rc && (rc = vsd_save_file (fpasswd, vc->server_root, vuser_passwd (NULL, domain))))
    io_buffer_store (vc, 
		     "-ERR [vuser][%s] error saving passwd file: %m\n", 
		     vuser);

  /* Delete mbx. */
  if (!rc && (rc = unlink (vuser_mbx (vc->server_root, domain, name))) && (errno != ENOENT))
      io_buffer_store (vc, 
		       "-ERR [vuser][%s] error deleting mailbox: %m\n", 
		       vuser);
    
  free (domain);
  free (name);
  free (vuser);
  
  vsd_free_file (fpasswd);    
}

/******************************************************************************
 * Description: list virtual user mailboxes.
 *
 *     Returns: void.
 *****************************************************************************/
void 
vuser_mbxlist (struct connection *vc, 
	       const int         argc, 
	       const char        *argv[])
{
  char *pch, *domain;
  int rc = VSD_OK;
  struct vsd_file_line *fpasswd = NULL;
  struct vsd_file_line *line;

  if (argc != 0)
    {
      io_buffer_store (vc, "-ERR [vuser] Incorrect arguments\n");
      return;
    }

  /* Useful strings. */
  domain = (char *) strdup (vc->virtual_server);
  
  if (rc = vsd_load_file (&fpasswd, vc->server_root, vuser_passwd (NULL, domain)))
    io_buffer_store (vc, 
		     "-ERR [vuser][%s] error loading passwd file: %m\n", 
		     domain);  
  
  /* Return entries. */
  for (line = fpasswd; !rc && line; line=line->next)
    {
      if (pch = (char *) strchr (line->line, ':'))	
	{
	  *pch = '\0';
	  io_buffer_store (vc, "%s\n", line->line);
	}
    }
  
  free (domain);
  
  vsd_free_file (fpasswd);
}

/******************************************************************************
 * Description: change virtual user mailbox password.
 *
 *     Returns: void.
 *****************************************************************************/
void 
vuser_mbxpass (struct connection *vc, 
	       const int         argc, 
	       const char        *argv[])
{

  char *pch;
  char *domain, *name, *vuser;
  char *old, *new;
  char *crypt = NULL;
  char entry[256];
  int rc = VSD_OK;
  struct vsd_file_line *fpasswd = NULL;
  struct vsd_file_line *line;
  
  if (argc != 3)
    {
      io_buffer_store (vc, "-ERR [vuser] Incorrect arguments\n");
      return;
    }

  /* Useful strings. */
  domain = (char *) strdup (vc->virtual_server);
  name = (char *) strdup (argv[0]);
  old = (char *) strdup (argv[1]);
  new = (char *) strdup (argv[2]);

  if (vuser = (char *)malloc (strlen (domain) + strlen (name) + 2))
    sprintf (vuser, "%s@%s", name, domain);  

  if (rc = vsd_load_file (&fpasswd, vc->server_root, vuser_passwd (NULL, domain)))
    io_buffer_store (vc, 
		     "-ERR [vuser][%s] error loading passwd file: %m\n", 
		     vuser); 
 
  /* Test for existing entry. */
  if (!rc && !(line = vuser_passwd_find (fpasswd, name)))
    {
      rc = VSD_ERR;
      io_buffer_store (vc, 
		       "-ERR [vuser][%s] mailbox does not exist\n",
		       vuser);		  
    }

  /* Confirm password. */
  if (!rc)
    {
      pch = (char *) strchr (line->line, ':');
      pch++;
      if (rc = vsd_passwd_validate (pch, old))
	io_buffer_store (vc, 
			 "-ERR [vuser][%s] incorrect mailbox passwd\n",
			 vuser);
    }

  /* Encrypt password. */
  if (!rc && !(crypt = (char *) vsd_passwd_crypt (new)))
    {
      rc = VSD_ERR;
      io_buffer_store (vc, 
		       "-ERR [vuser][%s] error encrypting mailbox passwd\n",
		       vuser);
    }

  /* Set password. */
  if (!rc)
    {
      sprintf (entry, "%s:%s", name, crypt);
      if (!(rc = vsd_file_insert_line (line, entry)))
	fpasswd = vsd_file_delete_line (fpasswd, line);
      else
	io_buffer_store (vc, 
			 "-ERR [vuser][%s] error setting mailbox passwd\n",
			 vuser);
    }

  if (!rc && (rc = vsd_save_file (fpasswd, vc->server_root, vuser_passwd (NULL, domain))))
    io_buffer_store (vc, 
		     "-ERR [vuser][%s] error saving passwd file: %m\n", 
		     vuser);
  
  free (domain);
  free (name);
  free (old);
  free (new);
  free (vuser);
  free (crypt);

  vsd_free_file (fpasswd);    
}

/******************************************************************************
 * Description: enable virtual user mailbox.
 *
 *     Returns: void.
 *****************************************************************************/
void 
vuser_mbxenable (struct connection *vc, 
		 const int         argc, 
		 const char        *argv[])
{
  char *pch;
  char *domain, *name, *vuser;
  char entry[256];
  int rc = VSD_OK;
  struct vsd_file_line *fpasswd = NULL;
  struct vsd_file_line *line;

  if (argc != 1)
    {
      io_buffer_store (vc, "-ERR [vuser] Incorrect arguments\n");
      return;
    }

  /* Useful strings. */
  domain = (char *) strdup (vc->virtual_server);
  name = (char *) strdup (argv[0]);

  if (vuser = (char *)malloc (strlen (domain) + strlen (name) + 2))
    sprintf (vuser, "%s@%s", name, domain);  

  if (rc = vsd_load_file (&fpasswd, vc->server_root, vuser_passwd (NULL, domain)))
    io_buffer_store (vc, 
		     "-ERR [vuser][%s] error loading passwd file: %m\n", 
		     vuser);  

  if (!rc && !(line = vuser_passwd_find (fpasswd, name)))
    {
      rc = VSD_ERR;
      io_buffer_store (vc, 
		       "-ERR [vuser][%s] mailbox does not exist\n",
		       vuser);		  
    }

  if (!rc)
    {
      /* Handle disabled vusers. */
      if (line->line[0] != '!')
	{
	  strncpy (entry, line->line, sizeof (entry) - 1);
	  entry [sizeof (entry) - 1] = '\0';
	}
      else 
	{
	  pch = line->line;
	  strncpy (entry, ++pch, sizeof (entry) - 1);
	  entry [sizeof (entry) - 1] = '\0';
	}	
    }

  if (!rc)
    {
      if (!(rc = vsd_file_insert_line (line, entry)))
	fpasswd = vsd_file_delete_line (fpasswd, line);
      else
	io_buffer_store (vc, 
			 "-ERR [vuser][%s] error enabling mailbox\n",
			 vuser);
    }

  if (!rc && (rc = vsd_save_file (fpasswd, vc->server_root, vuser_passwd (NULL, domain))))
    io_buffer_store (vc, 
		     "-ERR [vuser][%s] error saving passwd file: %m\n", 
		     vuser);

  free (vuser);
  free (name);
  free (domain);

  vsd_free_file (fpasswd);    
}

/******************************************************************************
 * Description: disable virtual user mailbox.
 *
 *     Returns: void.
 *****************************************************************************/
void 
vuser_mbxdisable (struct connection *vc, 
		  const int         argc, 
		  const char        *argv[])
{
  char *pch;
  char *domain, *name, *vuser;
  char entry[256];
  int rc = VSD_OK;
  struct vsd_file_line *fpasswd = NULL;
  struct vsd_file_line *line;

  if (argc != 1)
    {
      io_buffer_store (vc, "-ERR [vuser] Incorrect arguments\n");
      return;
    }

  /* Useful strings. */
  domain = (char *) strdup (vc->virtual_server);
  name = (char *) strdup (argv[0]);

  if (vuser = (char *)malloc (strlen (domain) + strlen (name) + 2))
    sprintf (vuser, "%s@%s", name, domain);  

  if (rc = vsd_load_file (&fpasswd, vc->server_root, vuser_passwd (NULL, domain)))
    io_buffer_store (vc, 
		     "-ERR [vuser][%s] error loading passwd file: %m\n", 
		     vuser);  

  if (!rc && !(line = vuser_passwd_find (fpasswd, name)))
    {
      rc = VSD_ERR;
      io_buffer_store (vc, 
		       "-ERR [vuser][%s] mailbox does not exist\n",
		       vuser);		  
    }

  if (!rc)
    {
      /* Handle disabled vusers. */
      if (line->line[0] != '!')
	{
	  strcpy (entry, "!");
	  strncat (entry, line->line, sizeof (entry) - 2);
	  entry [sizeof (entry) - 1] = '\0';
	}
      else 
	{
	  strncpy (entry, line->line, sizeof (entry) - 1);
	  entry [sizeof (entry) - 1] = '\0';
	}
    }

  if (!rc)
    {
      if (!(rc = vsd_file_insert_line (line, entry)))
	fpasswd = vsd_file_delete_line (fpasswd, line);
      else
	io_buffer_store (vc, 
			 "-ERR [vuser][%s] error disabling mailbox\n",
			 vuser);
    }

  if (!rc && (rc = vsd_save_file (fpasswd, vc->server_root, vuser_passwd (NULL, domain))))
    io_buffer_store (vc, 
		     "-ERR [vuser][%s] error saving passwd file: %m\n", 
		     vuser);

  free (vuser);
  free (name);
  free (domain);

  vsd_free_file (fpasswd);    
}

/******************************************************************************
 * Description: add virtual user response message.
 *
 *     Returns: void.
 *****************************************************************************/
void 
vuser_msgadd (struct connection *vc, 
	      const int         argc, 
	      const char        *argv[])
{
  char *pch, *buf;
  char *domain, *name, *msg, *vuser;
  int rc = VSD_OK;
  struct vsd_file_line *fdomain_rc = NULL;
  struct vsd_file_line *fmsgs = NULL;
  struct vsd_file_line *msgs_line = NULL;
  struct vuser_cnf *domain_cnf = NULL;
  struct vuser_cnf *cnf = NULL;
  struct vsd_vs *domain_map;
  
  if (argc != 2)
    {
      io_buffer_store (vc, "-ERR [vuser] Incorrect arguments\n");
      return;
    }

  /* Useful strings. */
  domain = (char *) strdup (vc->virtual_server);
  name = (char *) strdup (argv[0]);
  msg = (char *) strdup (argv[1]);

  domain_map = vsd_getvdent (vc->map, domain);

  if ((pch = 
       vuser_check_msg_limit (domain_map, 
                              vsd_map_server_root (vc->map, domain_map->hs),
                              name)) != NULL)
    {
      rc = VSD_ERR;
      io_buffer_store (vc, "-ERR [vuser][limit check] %s\n", pch);
    }

  if (vuser = (char *)malloc (strlen (domain) + strlen (name) + 2))
    sprintf (vuser, "%s@%s", name, domain);  

  /* Load domain recipe file. */
  if (!rc && (rc = vsd_load_file (&fdomain_rc, vc->server_root, 
			  vuser_domain_rc (NULL, domain))))
    io_buffer_store (vc, 
		     "-ERR [vuser][%s] error loading domain recipe file: %m\n",
		     vuser);  
  
  /* Parse domain recipe file. */
  if (!rc && (rc = vuser_rc_parse (fdomain_rc, &domain_cnf)))
    io_buffer_store (vc, 
		     "-ERR [vuser][%s] error parsing domain recipe file\n",
		     vuser);		  
  
  /* Test for existing vuser entry. */
  if (!rc && !(vuser_list_find (domain_cnf, name)))
    {
      rc = VSD_ERR;
      io_buffer_store (vc, 
		       "-ERR [vuser][%s] vuser does not exist\n",
		       vuser);		  
    }
  
  /* Load msg index. */
  if (!rc && (rc = vsd_load_file (&fmsgs, vc->server_root, 
				  vuser_msgs (NULL, domain, name))))
    io_buffer_store (vc, 
		     "-ERR [vuser][%s] error loading msg index: %m\n", 
		     vuser);  
  
  /* Test for existing msg entry. */
  if (!rc && (msgs_line = vuser_msgs_find (fmsgs, msg)))
    {
      rc = VSD_ERR;
      io_buffer_store (vc, 
		       "-ERR [vuser][%s][%s] msg already exists\n",
		       vuser, msg);		  
    }

  /* Create vuser msg. */
  if (!rc)
    {
      struct vsd_file_line fmsg;
      
      fmsg.prev = fmsg.next = NULL;
      fmsg.line = (char *) strdup ("");
      
      if (!rc && (rc = vsd_save_file (&fmsg, vc->server_root, 
				      vuser_msg (NULL, domain, name, msg))))
	io_buffer_store (vc, 
			 "-ERR [vuser][%s][%s] error saving vuser msg: %m\n", 
			 vuser, msg);
      free (fmsg.line);
    }

  /* Add entry. */
  if (!rc && (rc = vsd_file_insert_line (vsd_file_tail (fmsgs), msg)))
    io_buffer_store (vc, 
		     "-ERR [vuser][%s][%s] error adding msg to msg index\n",
		     vuser, msg);

  /* Save msg index. */
  if (!rc && (rc = vsd_save_file (fmsgs, vc->server_root, 
				  vuser_msgs (NULL, domain, name))))
    io_buffer_store (vc, 
		     "-ERR [vuser][%s] error saving msg index: %m\n", 
		     vuser);
  
  free (vuser);
  free (name);
  free (domain);
  free (msg);
  
  for (cnf = domain_cnf; cnf; cnf = cnf->next)
    {
      vuser_list_free (cnf->fwd_cnf);
      vuser_list_free (cnf->rsp_cnf);
    }
  vuser_list_free (domain_cnf);
  
  vsd_free_file (fdomain_rc);    
  vsd_free_file (fmsgs);
}

/******************************************************************************
 * Description: delete virtual user response message.
 *
 *     Returns: void.
 *****************************************************************************/
void 
vuser_msgdel (struct connection *vc, 
	      const int         argc, 
	      const char        *argv[])
{
  char *pch, *buf;
  char *domain, *name, *msg, *vuser;
  int rc = VSD_OK;
  struct vsd_file_line *fdomain_rc = NULL;
  struct vsd_file_line *fmsgs = NULL;
  struct vsd_file_line *msgs_line = NULL;
  struct vuser_cnf *domain_cnf = NULL;
  struct vuser_cnf *cnf = NULL;

  if (argc != 2)
    {
      io_buffer_store (vc, "-ERR [vuser] Incorrect arguments\n");
      return;
    }

  /* Useful strings. */
  domain = (char *) strdup (vc->virtual_server);
  name = (char *) strdup (argv[0]);
  msg = (char *) strdup (argv[1]);

  if (vuser = (char *)malloc (strlen (domain) + strlen (name) + 2))
    sprintf (vuser, "%s@%s", name, domain);  

  /* Load domain recipe file. */
  if (rc = vsd_load_file (&fdomain_rc, vc->server_root, vuser_domain_rc (NULL, domain)))
    io_buffer_store (vc, 
		     "-ERR [vuser][%s] error loading domain recipe file: %m\n", 
		     vuser);  

  /* Parse domain recipe file. */
  if (!rc && (rc = vuser_rc_parse (fdomain_rc, &domain_cnf)))
    io_buffer_store (vc, 
		     "-ERR [vuser][%s] error parsing domain recipe file\n",
		     vuser);		  

  /* Test for existing vuser entry. */
  if (!rc && !(vuser_list_find (domain_cnf, name)))
    {
      rc = VSD_ERR;
      io_buffer_store (vc, 
		       "-ERR [vuser][%s] vuser does not exist\n",
		       vuser);		  
    }
  
  /* Load msg index. */
  if (!rc && (rc = vsd_load_file (&fmsgs, vc->server_root, vuser_msgs (NULL, domain, name))))
    io_buffer_store (vc, 
		     "-ERR [vuser][%s] error loading msg index: %m\n", 
		     vuser);  
  
  /* Test for existing msg entry. */
  if (!rc && !(msgs_line = vuser_msgs_find (fmsgs, msg)))
    {
      rc = VSD_ERR;
      io_buffer_store (vc, 
		       "-ERR [vuser][%s][%s] msg does not exist\n",
		       vuser, msg);		  
    }

  /* Delete msg entry. */
  if (!rc)
    fmsgs = vsd_file_delete_line (fmsgs, msgs_line);

  /* Save msg index. */
  if (!rc && (rc = vsd_save_file (fmsgs, vc->server_root, vuser_msgs (NULL, domain, name))))
    io_buffer_store (vc, 
		     "-ERR [vuser][%s] error saving msg index: %m\n", 
		     vuser);
  
  /* Delete msg. */
  if (!rc && (rc = unlink (vuser_msg (vc->server_root, domain, name, msg))))
    io_buffer_store (vc, 
		     "-ERR [vuser][%s][%s] error deleting msg: %m\n", 
		     vuser, msg);
  
  free (vuser);
  free (name);
  free (domain);
  free (msg);

  for (cnf = domain_cnf; cnf; cnf = cnf->next)
    {
      vuser_list_free (cnf->fwd_cnf);
      vuser_list_free (cnf->rsp_cnf);
    }
  vuser_list_free (domain_cnf);

  vsd_free_file (fdomain_rc);    
  vsd_free_file (fmsgs);
}

/******************************************************************************
 * Description: get virtual user response message.
 *
 *     Returns: void.
 *****************************************************************************/
void 
vuser_msgget (struct connection *vc, 
	      const int         argc, 
	      const char        *argv[])
{
  char *pch, *buf;
  char *domain, *name, *msg, *vuser;
  int rc = VSD_OK;
  struct vsd_file_line *fdomain_rc = NULL;
  struct vsd_file_line *fmsgs = NULL;
  struct vsd_file_line *fmsg = NULL;
  struct vsd_file_line *msgs_line = NULL;
  struct vsd_file_line *msg_line = NULL;
  struct vuser_cnf *domain_cnf = NULL;
  struct vuser_cnf *cnf = NULL;
  
  if (argc != 2)
    {
      io_buffer_store (vc, "-ERR [vuser] Incorrect arguments\n");
      return;
    }

  /* Useful strings. */
  domain = (char *) strdup (vc->virtual_server);
  name = (char *) strdup (argv[0]);
  msg = (char *) strdup (argv[1]);

  if (vuser = (char *)malloc (strlen (domain) + strlen (name) + 2))
    sprintf (vuser, "%s@%s", name, domain);  

  /* Load domain recipe file. */
  if (rc = vsd_load_file (&fdomain_rc, vc->server_root, vuser_domain_rc (NULL, domain)))
    io_buffer_store (vc, 
		     "-ERR [vuser][%s] error loading domain recipe file: %m\n", 
		     vuser);  

  /* Parse domain recipe file. */
  if (!rc && (rc = vuser_rc_parse (fdomain_rc, &domain_cnf)))
    io_buffer_store (vc, 
		     "-ERR [vuser][%s] error parsing domain recipe file\n",
		     vuser);		  

  /* Test for existing vuser entry. */
  if (!rc && !(vuser_list_find (domain_cnf, name)))
    {
      rc = VSD_ERR;
      io_buffer_store (vc, 
		       "-ERR [vuser][%s] vuser does not exist\n",
		       vuser);		  
    }
  
  /* Load msg index. */
  if (!rc && (rc = vsd_load_file (&fmsgs, vc->server_root, vuser_msgs (NULL, domain, name))))
    io_buffer_store (vc, 
		     "-ERR [vuser][%s] error loading msg index: %m\n", 
		     vuser);  
  
  /* Test for existing msg entry. */
  if (!rc && !(msgs_line = vuser_msgs_find (fmsgs, msg)))
    {
      rc = VSD_ERR;
      io_buffer_store (vc, 
		       "-ERR [vuser][%s][%s] msg does not exist\n",
		       vuser, msg);		  
    }

  /* Load msg. */
  if (!rc && (rc = vsd_load_file (&fmsg, vc->server_root, vuser_msg (NULL, domain, name, msg))))
    io_buffer_store (vc, 
		     "-ERR [vuser][%s][%s] error loading msg: %m\n", 
		     vuser, msg);  
  
  /* Return msg. */
  if (!rc)
    {
      for (msg_line = fmsg; !rc && msg_line; msg_line=msg_line->next)
	io_buffer_store (vc, "%s\n", msg_line->line);
    }

  free (vuser);
  free (name);
  free (domain);
  free (msg);

  for (cnf = domain_cnf; cnf; cnf = cnf->next)
    {
      vuser_list_free (cnf->fwd_cnf);
      vuser_list_free (cnf->rsp_cnf);
    }
  vuser_list_free (domain_cnf);

  vsd_free_file (fdomain_rc);    
  vsd_free_file (fmsgs);
  vsd_free_file (fmsg);
}

/******************************************************************************
 * Description: set virtual user response message.
 *
 *     Returns: void.
 *****************************************************************************/
void 
vuser_msgset (struct connection *vc, 
	      const int         argc, 
	      const char        *argv[],
	      const char        *data)
{
  char *pch, *buf;
  char *domain, *name, *msg, *vuser;
  int rc = VSD_OK;
  struct vsd_file_line *fdomain_rc = NULL;
  struct vsd_file_line *fmsgs = NULL;
  struct vsd_file_line *fmsg = NULL;
  struct vsd_file_line *msgs_line = NULL;
  struct vuser_cnf *domain_cnf = NULL;
  struct vuser_cnf *cnf = NULL;
  FILE *pfile;
  
  if (argc != 2)
    {
      io_buffer_store (vc, "-ERR [vuser] Incorrect arguments\n");
      return;
    }

  /* Useful strings. */
  domain = (char *) strdup (vc->virtual_server);
  name = (char *) strdup (argv[0]);
  msg = (char *) strdup (argv[1]);

  if (vuser = (char *)malloc (strlen (domain) + strlen (name) + 2))
    sprintf (vuser, "%s@%s", name, domain);  

  /* Load domain recipe file. */
  if (rc = vsd_load_file (&fdomain_rc, vc->server_root, vuser_domain_rc (NULL, domain)))
    io_buffer_store (vc, 
		     "-ERR [vuser][%s] error loading domain recipe file: %m\n", 
		     vuser);  

  /* Parse domain recipe file. */
  if (!rc && (rc = vuser_rc_parse (fdomain_rc, &domain_cnf)))
    io_buffer_store (vc, 
		     "-ERR [vuser][%s] error parsing domain recipe file\n",
		     vuser);		  

  /* Test for existing vuser entry. */
  if (!rc && !(vuser_list_find (domain_cnf, name)))
    {
      rc = VSD_ERR;
      io_buffer_store (vc, 
		       "-ERR [vuser][%s] vuser does not exist\n",
		       vuser);		  
    }
  
  /* Load msg index. */
  if (!rc && (rc = vsd_load_file (&fmsgs, vc->server_root, vuser_msgs (NULL, domain, name))))
    io_buffer_store (vc, 
		     "-ERR [vuser][%s] error loading msg index: %m\n", 
		     vuser);  

  /* Test for existing msg entry. */
  if (!rc && !(msgs_line = vuser_msgs_find (fmsgs, msg)))
    {
      rc = VSD_ERR;
      io_buffer_store (vc, 
		       "-ERR [vuser][%s][%s] msg does not exist\n",
		       vuser, msg);		  
    }

  /* Open new msg. */
  if (!rc)
    {
      if ((pfile = fopen (vuser_msg (vc->server_root, domain, name, msg), "w+")) == NULL)
	{
	  rc = VSD_ERR;
	  io_buffer_store (vc, 
			   "-ERR [vuser][%s][%s] error saving msg: %m\n",
			   vuser, msg);
	}
    }

  /* Write new msg. */
  if (!rc)
    {
      while ((buf = io_getline ('\n', &data)))
	{
	  fprintf (pfile, "%s\n", buf);
	  free (buf);
	}
      fclose (pfile);
    }

  return;

  free (vuser);
  free (name);
  free (domain);
  free (msg);

  for (cnf = domain_cnf; cnf; cnf = cnf->next)
    {
      vuser_list_free (cnf->fwd_cnf);
      vuser_list_free (cnf->rsp_cnf);
    }
  vuser_list_free (domain_cnf);

  vsd_free_file (fdomain_rc);    
  vsd_free_file (fmsgs);
}

/******************************************************************************
 * Description: list virtual user response messages.
 *
 *     Returns: void.
 *****************************************************************************/
void 
vuser_msglist (struct connection *vc, 
	       const int         argc, 
	       const char        *argv[])
{
  char *pch, *buf;
  char *domain, *name, *vuser;
  char entry[256];
  int rc = VSD_OK;
  struct vsd_file_line *fdomain_rc = NULL;
  struct vsd_file_line *fmsgs = NULL;
  struct vsd_file_line *msgs_line = NULL;
  struct vuser_cnf *domain_cnf = NULL;
  struct vuser_cnf *cnf = NULL;
  
  if (argc != 1)
    {
      io_buffer_store (vc, "-ERR [vuser] Incorrect arguments\n");
      return;
    }

  /* Useful strings. */
  domain = (char *) strdup (vc->virtual_server);
  name = (char *) strdup (argv[0]);

  if (vuser = (char *)malloc (strlen (domain) + strlen (name) + 2))
    sprintf (vuser, "%s@%s", name, domain);  

  /* Load domain recipe file. */
  if (rc = vsd_load_file (&fdomain_rc, vc->server_root, vuser_domain_rc (NULL, domain)))
    io_buffer_store (vc, 
		     "-ERR [vuser][%s] error loading domain recipe file: %m\n", 
		     vuser);  

  /* Parse procmailrc. */
  if (!rc && (rc = vuser_rc_parse (fdomain_rc, &domain_cnf)))
    io_buffer_store (vc, 
		     "-ERR [vuser][%s] error parsing domain recipe file\n",
		     vuser);		  

  /* Test for existing vuser entry. */
  if (!rc && !(vuser_list_find (domain_cnf, name)))
    {
      rc = VSD_ERR;
      io_buffer_store (vc, 
		       "-ERR [vuser][%s] vuser does not exist\n",
		       vuser);		  
    }

  /* Load msg index. */
  if (!rc && (rc = vsd_load_file (&fmsgs, vc->server_root, vuser_msgs (NULL, domain, name))))
    io_buffer_store (vc, 
		     "-ERR [vuser][%s] error loading msg index: %m\n", 
		     vuser);  
  
  /* Return entries. */
  for (msgs_line = fmsgs; !rc && msgs_line; msgs_line=msgs_line->next)
    io_buffer_store (vc, "%s\n", msgs_line->line);
  
  free (vuser);
  free (name);
  free (domain);

  for (cnf = domain_cnf; cnf; cnf = cnf->next)
    {
      vuser_list_free (cnf->fwd_cnf);
      vuser_list_free (cnf->rsp_cnf);
    }
  vuser_list_free (domain_cnf);

  vsd_free_file (fdomain_rc);    
  vsd_free_file (fmsgs);
}





