/* Client process.
   Copyright (c) 2001 Designer Servers Ltd.
   Written by Nick Burrett <nick@dsvr.net>

   This file is part of the Virtual Server Administrator (FreeVSD)

   FreeVSD is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2, or (at your option)
   any later version.

   FreeVSD is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with FreeVSD; see the file COPYING.  If not, write to
   the Free Software Foundation, Inc., 59 Temple Place - Suite 330,
   Boston, MA 02111-1307, USA.  */

#include <errno.h>
#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <unistd.h>
#include <signal.h>
#include <sys/time.h>
#include <sys/wait.h>
#include <netdb.h>
#include <netinet/in.h>
#include <netinet/tcp.h>
#include <sys/socket.h>

#include "libvsd.h"
#include "vsd.h"
#include "client.h"
#include "cryptssl.h"

void Client::do_redirection (int in_fd, int out_fd, CryptSSL *in_ssl)
{
  /* client_fd is the SSL-client side.  server_fd is the file descriptor
     for our connection to the hosting server.  */
  fd_set masterfds, readfds;
  struct timeval tv;
  int maxfd;
  char buff[256];

  FD_ZERO (&masterfds);
  FD_SET (in_fd, &masterfds);
  FD_SET (out_fd, &masterfds);

  maxfd = (in_fd > out_fd) ? in_fd : out_fd;
  /* Timeout after 60 seconds.  */
  tv.tv_sec = 60;
  tv.tv_usec = 0;
  while (1)
    {
      memcpy (&readfds, &masterfds, sizeof (fd_set));

      if (select (maxfd + 1, &readfds, NULL, NULL, &tv) < 0)
	{
	  printf ("select: %s\n", strerror (errno));
	  break;
	}

      if (FD_ISSET (in_fd, &readfds))
	{
	  int num_read, num_wrote, ret;
	  if ((num_read = in_ssl->Read (buff, sizeof (buff) - 1)) <= 0)
	    break;

	  num_wrote = 0;
	  while (num_wrote < num_read)
	    {
	      ret = write (out_fd, buff + num_wrote, num_read - num_wrote);
	      if (ret < 0)
		break;
	      num_wrote += ret;
	    }
	}

      if (FD_ISSET (out_fd, &readfds))
	{
	  int num_read, num_wrote, ret;
	  if ((num_read = read (out_fd, buff, sizeof (buff) - 1)) <= 0)
	    break;

	  num_wrote = 0;
	  while (num_wrote < num_read)
	    {
	      ret = in_ssl->Write (buff + num_wrote, num_read - num_wrote);
	      if (ret < 0)
		break;
	      num_wrote += ret;
	    }
	}
    }
  printf ("... terminating\n");
}

/* Called as soon as a connection comes in, negotiate SSL with the
   client, then setup a non-SSL connection to the hosting server.  Once
   that's done, simply forward communications between the two connections.  */
void Client::serve (int in_fd, int server_port, Lookup *vsmap)
{
  // in_fd is client to proxy file descriptor
  // out_fd is proxy to vsd hosting server file descriptor
  int out_fd;
  VSD hostsvr;
  CryptSSL in_ssl;

  // Setup the SSL comms layer between ourselves and the client.
  // This will set `certificate' to the name of the virtual server that
  // the client wishes to connect to.
  if (in_ssl.Init (in_fd))
    printf ("client connection failed: %s\n", strerror (errno));
  else
    {
      // Find the hosting server that the virtual server belongs to
      const char *fqdn = vsmap->Find (in_ssl.GetCertificate ());

      printf ("-- certificate for vs %s maps to svr %s\n",
	      in_ssl.GetCertificate (), fqdn);

      // Establish a non-SSL connection to the hosting server
      out_fd = hostsvr.Connect (fqdn, server_port);
      
      if (out_fd != -1)
	{
	  // Perform port redirection between the SSL-client and the non-SSL
	  // hosting server.
	  do_redirection (in_fd, out_fd, &in_ssl);
	  
	  // Close connection to the hosting server
	  hostsvr.Disconnect ();
	}

      // Close SSL connection to the client
      in_ssl.Destroy ();
      close (in_fd);
    }
}
