/* Client-side SSL layer.
   Copyright (c) 2001 Designer Servers Ltd.
   Written by Nick Burrett <nick@dsvr.net>

   This file is part of the Virtual Server Administrator (FreeVSD)

   FreeVSD is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2, or (at your option)
   any later version.

   FreeVSD is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with FreeVSD; see the file COPYING.  If not, write to
   the Free Software Foundation, Inc., 59 Temple Place - Suite 330,
   Boston, MA 02111-1307, USA.  */

#include <string.h>
#include "libvsd.h"
#include "cryptssl.h"

#define DEBUG 1
#define SSL_BASE "/usr/local/etc/vsd"

static int verify_callback (int ok, X509_STORE_CTX *ctx)
{
  X509 *cert;
  X509_NAME *subject, *issuer;
  int errnum, errdepth;
  char *sname, *iname;

  cert = X509_STORE_CTX_get_current_cert (ctx);
  errnum = X509_STORE_CTX_get_error (ctx);
  errdepth = X509_STORE_CTX_get_error_depth (ctx);

  subject = X509_get_subject_name (cert);
  issuer = X509_get_issuer_name (cert);
  sname = X509_NAME_oneline (subject, NULL, 0);
  iname = X509_NAME_oneline (issuer, NULL, 0);

  vsd_log_error (VSD_LOG_DEBUG,
		 "SSL certificate verification: ok: %d, depth: %d, subject: %s, issuer: %s",
		 ok, errdepth,
		 sname ? sname : "-unknown-",
		 iname ? iname : "-unknown-");
  if (sname)
    CRYPTO_free (sname);
  if (iname)
    CRYPTO_free (iname);

  return ok;
}

int CryptSSL::Init (int fd)
{
  int err;
  X509 *client_cert;
  char buf[256];

  /* Setting this alarm will stop our program hanging indefinately, waiting
     for an SSL session to be established.  */
  //alarm (command_timeout);
  SSL_load_error_strings ();
  SSLeay_add_ssl_algorithms ();

  /* Create a context.  */
  ctx = SSL_CTX_new (SSLv23_method ());

  /* In order to create new session contexts, we need the certificate
     and RSA key.  */
  if (ctx == NULL
      || SSL_CTX_use_PrivateKey_file (ctx, SSL_BASE "/server/host.key",
				      SSL_FILETYPE_PEM) <= 0
      || SSL_CTX_use_certificate_file (ctx, SSL_BASE "/server/host.crt",
				       SSL_FILETYPE_PEM) <= 0)
    {
      vsd_log_error (VSD_LOG_ERROR, "%s",
		     ERR_error_string (ERR_get_error (), buf));
      return 1;
    }

  /* Check the private key is valid.  */
  if (! SSL_CTX_check_private_key (ctx))
    {
      vsd_log_error (VSD_LOG_ERROR, "Private key does not match the certificate public key");
      return 1;
    }

  if (! SSL_CTX_load_verify_locations (ctx, SSL_BASE "/ca/cacert.pem",
				       SSL_BASE "/ca")
      || ! SSL_CTX_set_default_verify_paths (ctx))
    {
      vsd_log_error (VSD_LOG_ERROR, "load_verify_locations: %s",
		     ERR_error_string (ERR_get_error (), buf));
    }
  /* Add a Certificate Authority.  */
  SSL_CTX_set_client_CA_list (ctx,
			      SSL_load_client_CA_file (SSL_BASE "/server/ca-cert.pem"));

  /**/
  SSL_CTX_set_verify (ctx, SSL_VERIFY_PEER
		      | SSL_VERIFY_FAIL_IF_NO_PEER_CERT, &verify_callback);

  /*SSL_CTX_set_tmp_rsa_callback (io->ctx, rsa_cb);*/

  /* Create a new SSL structure.  */
  ssl = SSL_new (ctx);
  if (ssl == NULL)
    {
      vsd_log_error (VSD_LOG_ERROR, "cannot initialise SSL");
      return 1;
    }

  /* Associate the input file descriptor with the SSL structure.  */
  SSL_set_fd (ssl, fd);

  /* Do SSL using our certificate and RSA key.  */
  err = SSL_accept (ssl);
  if (err == -1)
    {
      vsd_log_error (VSD_LOG_ERROR, "%s",
		     ERR_error_string (ERR_get_error (), buf));
      return 1;
    }
  vsd_log_error (VSD_LOG_DEBUG, "SSL connection using %s",
		 SSL_get_cipher (ssl));

  client_cert = SSL_get_peer_certificate (ssl);
  if (client_cert != NULL)
    {
      char *s, buf[256], cn[128];
      int x;

      /* Extract CN field from certificate subject.
	 FIXME: There is a proper way to do this but the documentation
	 for SSL leaves a lot to guesswork.  */
      vsd_log_error (VSD_LOG_DEBUG, "client certificate:");
      s = X509_NAME_oneline (X509_get_subject_name (client_cert),
			     buf, sizeof (buf));
      if (s)
	vsd_log_error (VSD_LOG_DEBUG, "  subject: %s", s);

      s = strstr (buf, "/CN=");
      if (s == NULL)
	vsd_log_error (VSD_LOG_ERROR,
		       "  expected a CN field in certificate subject: %s", buf);
      s += 4;
      x = 0;
      while (*s && *s != '/' && x < sizeof (cn))
	cn[x++] = *s++;
      if (*s == '/' || *s == '\0')
	{
	  cn[x] = '\0';
	  vsd_log_error (VSD_LOG_DEBUG, "  found cn: %s", cn);
	  certificate = strdup (cn);
	}
      else
	{
	  cn[0] = '\0';
	  vsd_log_error (VSD_LOG_ERROR,
			 "  unexpected certificate subject format or buffer overflow: %s",
			 buf);
	  X509_free (client_cert);
	  return 1;
	}

      s = X509_NAME_oneline (X509_get_issuer_name (client_cert),
			     buf, sizeof (buf));
      if (s)
	vsd_log_error (VSD_LOG_DEBUG, "  issuer: %s", s);

      X509_free (client_cert);
    }
  else
    vsd_log_error (VSD_LOG_DEBUG, "client does not have a certificate");

  return 0;
}

int CryptSSL::Destroy (void)
{
  SSL_free (ssl);
  SSL_CTX_free (ctx);
  return 0;
}
