/* Class implementation for VSD server conversions.
   Copyright (c) 2001 Designer Servers Ltd.
   Written by Nick Burrett <nick@dsvr.net>

   This file is part of the Virtual Server Administrator (FreeVSD)

   FreeVSD is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2, or (at your option)
   any later version.

   FreeVSD is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with FreeVSD; see the file COPYING.  If not, write to
   the Free Software Foundation, Inc., 59 Temple Place - Suite 330,
   Boston, MA 02111-1307, USA.  */

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <errno.h>
#include <unistd.h>
#include <netdb.h>
#include <netinet/in.h>
#include <netinet/tcp.h>
#include <sys/socket.h>
#include "vsd.h"

int VSD::Connect (const char *fqdn, int port, int response=0)
{
  struct hostent *hostent;
  struct sockaddr_in addr;

  /* Establish a connection to the remote host.  */
  hostent = gethostbyname (fqdn);
  if (hostent == NULL)
    {
      printf ("could not resolve host %s\n", fqdn);
      return -1;
    }

  memset (&addr, 0, sizeof (addr));
  addr.sin_family = AF_INET;
  addr.sin_port = htons (port);
  addr.sin_addr = *(struct in_addr *) hostent->h_addr;

  fd = socket (AF_INET, SOCK_STREAM, 0);
  if (fd == -1)
    {
      printf ("socket create failed: %s\n", strerror (errno));
      return -1;
    }

  if (connect (fd, (struct sockaddr *) &addr, sizeof (addr)) < 0)
    {
      printf ("connection to %s could not be established: %s\n",
              fqdn, strerror (errno));
      return -1;
    }

  if (response)
    {
      int buflen;
      char *buff = Recv (buflen, '\n');
      buff[buflen] = '\0';
      printf ("received: %s\n", buff);
      free (buff);
    }

  return fd;
}

int VSD::Disconnect (void)
{
  return close (fd);
}

int VSD::Send (const char *text)
{
  write (fd, text, strlen (text));
}

/* Receive data from a socket.  Create a memory buffer, using malloc, to
   store the data, then set buffer_length to the length of data received.
   If term_char is non-zero, then scan the buffer for that character.  If
   a match is found, then we can assume that there is no more data to be
   read.  */
char *VSD::Recv (int &buffer_length, char term_char)
{
  fd_set m_fds, r_fds;
  int num_read;
  char *buff = NULL, *buffptr;
  int buflen, bufspace;

  buflen = 1024;  // Total size of our buffer
  buff = (char *) malloc (buflen);  // The buffer
  buffptr = buff; // Pointer to next byte to write to

  FD_ZERO (&m_fds);
  FD_SET (fd, &m_fds);
  while (1)
    {
      r_fds = m_fds;
      select (fd + 1, &r_fds, NULL, NULL, NULL);
      if (FD_ISSET (fd, &r_fds))
	{
	  if ((num_read = read (fd, buffptr, buflen - (buffptr - buff))) <= 0)
	    break;

	  buffptr += num_read;
	  if (term_char)
	    {
	      char *p;
	      for (p = buff; *p != term_char && p < buffptr; p++)
		;
	      if (*p == term_char)
		break;
	    }
	  if ((buflen - (buffptr - buff)) <= 32)
	    {
	      int temp = buffptr - buff;
	      buflen += 256;
	      buff = (char *) realloc (buff, buflen);
	      buffptr = buff + temp;
	    }
	}
    }
  buffer_length = buffptr - buff;
  return buff;
}

char *VSD::Recv (int &buffer_length)
{
  return Recv (buffer_length, 0);
}

char *VSD::GetLine (char *buff, char terminator)
{
  while (*buff && *buff != terminator)
    buff++;

  if (*buff == terminator)
    {
      *buff = '\0';
      return buff + 1;
    }

  return NULL;
}


/* Parse lines of the form:
     name=ivwd ip=212.69.194.220 uid=18900 users=200 quota=300 path="/virtual/disk1" prins=212.69.192.10 secns=212.69.192.11

   The only information we need is the virtual server name.  */
char *VSD::SendCommand (const char *command)
{
  char *buff, *line, *nextline;
  int buflen;

  Send ("<root> {\n");
  Send (command);
  Send ("\n}\nEOF\n");
  buff = Recv (buflen);
  buff[buflen] = '\0';
  return buff;
}

int VSD::VS_Details (char **ptr, char *buff)
{
  char *line;

  line = GetLine (*ptr, '\n');
  if (*ptr && strncmp (*ptr, "name=", 5) == 0)
    {
      char *p = *ptr + 5;
      while (*p != ' ')
	*buff++ = *p++;
      *buff = '\0';
    }
  else
    {
      *buff = '\0';
    }

  *ptr = line;
  return (line) ? 0 : 1;
}

VSD::~VSD (void)
{
  if (fd != -1)
    close (fd);
}
