/* VSD SSL protocol proxy for centralised certificate management.
   Copyright (c) 2001 Designer Servers Ltd.
   Written by Nick Burrett <nick@dsvr.net>

   This file is part of the Virtual Server Administrator (FreeVSD)

   FreeVSD is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2, or (at your option)
   any later version.

   FreeVSD is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with FreeVSD; see the file COPYING.  If not, write to
   the Free Software Foundation, Inc., 59 Temple Place - Suite 330,
   Boston, MA 02111-1307, USA.  */

#include <errno.h>
#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <unistd.h>
#include <signal.h>
#include <sys/time.h>
#include <sys/wait.h>
#include <netdb.h>
#include <netinet/in.h>
#include <netinet/tcp.h>
#include <sys/socket.h>

#include "libvsd.h"
#include "client.h"
//#include "mysql.h"
#include "lookup.h"

#define DEBUG 1

class Config
{
public:
  int daemonise;
  int listen_port;
  int listen_backlog;
  int max_clients;
  int server_port;
  const char *server_name;

  Config ()
  {
    daemonise = 1;
    // the port we listed for connections on
    listen_port = 1725;
    listen_backlog = 10;
    // maximum number of simultaneous connections we can handle
    // only works if debugging is off
    max_clients = 30;
    // the hostname we proxy connections to (probably unused)
    server_name = "localhost";
    server_port = 1726;  // the port number we proxy connections to
  }
};


static void sigchld_handler (int sig)
{
  pid_t pid;
  int status;

  pid = waitpid (-1, &status, 0);
  printf ("-- reaping child %d\n", pid);
}

static int bind_socket (int port, int backlog)
{
  struct sockaddr_in server;
  int sock = 0;

  server.sin_family = AF_INET;
  server.sin_port = htons (port);
  server.sin_addr.s_addr = htonl (INADDR_ANY);
  sock = socket (AF_INET, SOCK_STREAM, 0);
  if (sock < 0)
    return sock;

  for (int tries = 0; tries < 10; tries ++)
    {
      if (! bind (sock, (struct sockaddr *) &server,
		  sizeof (struct sockaddr_in)))
	break;
      printf ("can't bind: %s\n", strerror (errno));
      // Wait five seconds before attempting to bind again.
      sleep (5);
    }
  
  if (listen (sock, backlog) < 0)
    printf ("can't listen on socket: %s\n", strerror (errno));

  return sock;
}

static void serve_connections (Client *clients, Config *config, Lookup *vsmap)
{
  int sock_server, sock_client;
  struct sockaddr_in client;
  socklen_t client_len;

  signal (SIGCHLD, sigchld_handler);
  sock_server = bind_socket (config->listen_port, config->listen_backlog);
  client_len = sizeof (struct sockaddr_in);
  for (;;)
    {
      int x;
      sock_client = accept (sock_server, (struct sockaddr *) &client,
			    &client_len);
      // Find free client
      for (x = 0; x < config->max_clients; x++)
	{
	  if (! clients[x].pid)
	    break;
	}
      if (x == config->max_clients)
	fprintf (stderr, "Too many clients. Consider increasing maxclients\n");
      else
	{
	  struct vsd_vs_map *map = vsd_map_read ();
	  if (map == NULL)
	    fprintf (stderr, "-ERR Fatal error: proxy map read failure: %m\n");
	  else
	    {
	      int y;
	      for (y = 0; y < map->servers; y++ )
		{
		  if (map->server[y].name[0] == '\0')
		    continue; /* Skip deleted entries.  */
		  
#ifdef DEBUG
		  printf ("vs: `%s'\n", map->server[y].name);
#endif
		  // FIXME. Broken.
		  vsmap->Add (map->server[y].name, config->server_name);
		  //vsmap->Add (map->server[y].name, map->server[y].hs);
		}

	  /* Don't fork if debugging, since it makes it more difficult
	     to debug the code.  */
#ifndef DEBUG
	  clients[x].pid = fork ();
	  if (clients[x].pid == -1)
	    fprintf (stderr, "Error forking: %m");
	  else if (clients[x].pid == 0)
#endif
	    clients[x].serve (sock_client, config->server_port, vsmap);
	    }
	}
    }
}


int main (void)
{
  Config config;
  Client *clients;
  //  MySQL db;
  VSD vc;
  Lookup vsmap;
  char hosting_server[32], ip[20], fqdn[64];
  char prins[64], secns[64];

  clients = (Client *) malloc (config.max_clients * sizeof (Client));

#if 0
  printf ("-- querying host servers\n");
  /* Query database for hosting servers and retrieve a list of virtual
     servers.  */
  strcpy (fqdn, "localhost");
  strcpy (prins, "localhost");
  strcpy (secns, "localhost");

  //while (! db.gethostsvrs (hosting_server, ip, fqdn))
    {
      VSD vc;

      printf ("-- connecting to %s\n", fqdn);
      if (vc.Connect (fqdn, 1725, 1) != -1)
	{
	  char vs_buff[64];
	  char *ptr, *recvbuff = vc.SendCommand ("VS_DETAILS");

	  ptr = recvbuff;
	  while (! vc.VS_Details (&ptr, vs_buff))
	    {
	      if (*vs_buff)
		{
		  //printf ("vs: `%s'\n", vs_buff);
		  vsmap.Add (vs_buff, fqdn);
		  //if (! db.getnamesvrs (vs_buff, prins, secns))
		    {
		      printf ("-- vs '%s', pri '%s', sec '%s'\n",
			      vs_buff, prins, secns);
		      vsmap.Add (vs_buff, prins, secns);
		    }
		}
	    }
	  free (recvbuff);
	  vc.Disconnect ();
	}
    }
  printf ("-- testing map\n");
  printf ("blobby: %s\n", vsmap.Find ("blobby"));
  printf ("bart: %s\n", vsmap.Find ("bart"));
  printf ("foobar: %s\n", vsmap.Find ("foobar"));
#endif
  printf ("-- waiting to serve\n");

  serve_connections (clients, &config, &vsmap);
  return 0;
}

